(* Author: Vincent Cremet *)
(* Date: December 2006 *) 
(* File: Binders_proofs.v *)
(* Description: Representation of FGJ-omega binders using DeBruijn's indices (Proofs) *)

(** * Representation of FGJ-omega binders using DeBruijn's indices (Proofs) *)

Require Import List.
Require Import Arith.
Require Import Omega.

Require Import MatchCompNat.

Require Syntax.
Module SetProgram(My_Program: Syntax.Program).
Import My_Program.

Require Semantics.
Module My_Semantics := Semantics.SetProgram(My_Program).
Import My_Semantics.

Import My_WellFormedness.
Import My_Typing.
Import My_Auxiliary.
Import My_Substitutions.
Import My_Binders.

(*****************************************************************)

(** Successive variable liftings *)
Lemma Successive_variable_liftings:
  forall (n1 n2 m: nat) (X: TypeVar) (k: nat),
    (lift_at_depth_in_typeVar (lift_at_depth_in_typeVar X (n1 + n2) k) m (k + n2)) =
    (lift_at_depth_in_typeVar X (n1 + n2 + m) k)
.

Proof.
intros n1 n2 m X k.
unfold lift_at_depth_in_typeVar.
induction (match_comp_nat2_aux_0 X k).
(* X < k *)
pattern (match_comp_nat2 X k TypeVar X (X + (n1 + n2))).
rewrite match_comp_nat2_lt.
rewrite match_comp_nat2_lt.
rewrite match_comp_nat2_lt. trivial. trivial. auto with arith. trivial.
(* X >= k *)
pattern (match_comp_nat2 X k TypeVar X (X + (n1 + n2))).
rewrite match_comp_nat2_ge.
pattern (match_comp_nat2 X k TypeVar X (X + (n1 + n2 + m))).
rewrite match_comp_nat2_ge.
rewrite match_comp_nat2_ge. auto with arith. omega. trivial. trivial.
Qed.

(** Successive type liftings (auxiliary) *)
Lemma Successive_type_liftings_aux:
  forall (n1 n2 m: nat) (t: TypeS) (k: nat),
    (lift_at_depth_in_type (lift_at_depth_in_type t (n1 + n2) k) m (k + n2)) =
    (lift_at_depth_in_type t (n1 + n2 + m) k)
.

Proof.
intros n1 n2 m t. induction t.
(* proj *)
intro. simpl. rewrite IHt. trivial.
(* class *)
intro. simpl. trivial.
(* fun *)
intro. simpl. rewrite IHt1.
replace (S (k + n2)) with ((S k) + n2). rewrite IHt2. trivial. trivial.
(* apply *)
intro. simpl. rewrite IHt1. rewrite IHt2. trivial.
(* var *)
intro. simpl. rewrite Successive_variable_liftings. trivial.
(* tuple *)
intro. simpl. rewrite IHt. trivial.
(* param *)
intro. simpl. rewrite IHt1.
replace (S (k + n2)) with ((S k) + n2). rewrite IHt2. trivial. trivial.
(* section *)
intro. simpl. replace (S (k + n2)) with ((S k) + n2). rewrite IHt. trivial. trivial.
(* none *)
intro. simpl. trivial.
(* some *)
intro. simpl. rewrite IHt. trivial.
(* nil *)
intro. simpl. trivial.
(* cons *)
intro. simpl. rewrite IHt1. rewrite IHt2. trivial.
Qed.

Lemma Successive_type_liftings_cor1:
  forall (n m k: nat) (t : TypeS),
    n > k ->
    (lift_at_depth_in_type (lift_in_type t n) m k) =
    (lift_in_type t (n + m))
.

Proof.
intros.
cut (
    (lift_at_depth_in_type (lift_at_depth_in_type t ((n - k) + k) 0) m (0 + k)) =
    (lift_at_depth_in_type t ((n - k) + k + m) 0)
).
unfold lift_in_type.
replace ((n - k) + k) with n. trivial. omega.
apply Successive_type_liftings_aux.
Qed.

(** Successive type liftings *)
Lemma Successive_type_liftings:
  forall (t: TypeS) (n m: nat),
    (lift_in_type (lift_in_type t n) m) =
    (lift_in_type t (n + m))
.

Proof.
intros t n m.
unfold lift_in_type. 
replace (
  lift_at_depth_in_type (lift_at_depth_in_type t n 0) m 0 =
  lift_at_depth_in_type t (n + m) 0)
with (
 lift_at_depth_in_type (lift_at_depth_in_type t (n + 0) 0) m (0 + 0) =
 lift_at_depth_in_type t (n + 0 + m) 0).
apply Successive_type_liftings_aux.
simpl.
rewrite plus_comm. simpl. trivial.
Qed.

Lemma Lifting_then_dropping_in_typeVar:
  forall (X: nat) (n k: nat),
    (drop_at_depth_in_typeVar (lift_at_depth_in_typeVar X (S n) k) k) =
    (lift_at_depth_in_typeVar X n k)  
.

Proof.
intros.
destruct (match_comp_nat2_aux_0 X k).
unfold lift_at_depth_in_typeVar. rewrite match_comp_nat2_lt. rewrite match_comp_nat2_lt.
unfold drop_at_depth_in_typeVar. rewrite match_comp_nat2_lt. trivial. trivial. trivial.
trivial.
unfold lift_at_depth_in_typeVar. rewrite match_comp_nat2_ge. rewrite match_comp_nat2_ge.
unfold drop_at_depth_in_typeVar. rewrite match_comp_nat2_ge. 
replace (X + S n) with ((S X) + n). simpl. trivial. apply plus_Snm_nSm. omega. trivial.
trivial.
Qed.

Lemma Lifting_then_dropping:
  forall (t: TypeS) (n k: nat),
    (drop_at_depth_in_type (lift_at_depth_in_type t (S n) k) k) =
    (lift_at_depth_in_type t n k)  
.

Proof.
intros. generalize k. clear k. induction t.
(* proj *)
intro. simpl. rewrite IHt. trivial.
(* class *)
intro. simpl. trivial.
(* fun *)
intro. simpl. rewrite IHt1. rewrite IHt2. trivial.
(* apply *)
intro. simpl. rewrite IHt1. rewrite IHt2. trivial.
(* var *)
intro. simpl. rewrite Lifting_then_dropping_in_typeVar. trivial.
(* tuple *)
intro. simpl. rewrite IHt. trivial.
(* param *)
intro. simpl. rewrite IHt1. rewrite IHt2. trivial.
(* section *)
intro. simpl. rewrite IHt. trivial.
(* none *)
intro. simpl. trivial.
(* some *)
intro. simpl. rewrite IHt. trivial.
(* nil *)
intro. simpl. trivial.
(* cons *)
intro. simpl. rewrite IHt1. rewrite IHt2. trivial.
Qed.

(********************************************************************)

Lemma lift_in_typeVar_by_zero:
  forall (i k: nat),
    (lift_at_depth_in_typeVar i 0 k) = i
.

Proof.
intros. unfold lift_at_depth_in_typeVar.
destruct (match_comp_nat2_aux_0 i k).
rewrite match_comp_nat2_lt. trivial. trivial.
rewrite match_comp_nat2_ge. rewrite plus_comm. trivial. trivial.
Qed.

Lemma lift_in_type_by_zero:
  forall (t: TypeS) (k: nat),
    (lift_at_depth_in_type t 0 k) = t
.

Proof.
intro t. induction t.
intros. simpl. rewrite IHt; trivial.
intros. simpl. trivial.
intros. simpl. rewrite IHt1. rewrite IHt2. trivial.
intros. simpl. rewrite IHt1. rewrite IHt2. trivial.
intros. simpl. rewrite lift_in_typeVar_by_zero. trivial.
intros. simpl. rewrite IHt; trivial.
intros. simpl. rewrite IHt1. rewrite IHt2. trivial.
intros. simpl. rewrite IHt; trivial.
intros. simpl. trivial.
intros. simpl. rewrite IHt; trivial.
intros. simpl. trivial.
intros. simpl. rewrite IHt1. rewrite IHt2. trivial.
Qed.

End SetProgram.
