(* Author: Vincent Cremet *)
(* Date: June 2007 *) 
(* File: Confluence.v *)
(* Description: Confluence of type reduction in FGJ-omega *)

(** * Confluence of type reduction in FGJ-omega *)

(* Coq library *)
Require Import List.
Require Import Arith.
Require Import Wf_nat.
Require Import Relations.

(* My library *)
Require Import MatchCompNat.
Require Import Util.

(********************************************************************)

Require Syntax.
Module SetProgram(My_Program: Syntax.Program).
Import My_Program.

(* Import definitions of Auxiliary_proofs.v *)
Require Auxiliary_proofs.
Module My_Auxiliary_proofs := Auxiliary_proofs.SetProgram(My_Program).
Import My_Auxiliary_proofs.

Import My_Substitutions_proofs.
Import My_Binders_proofs.
Import My_Semantics.
Import My_WellFormedness.
Import My_Typing.
Import My_Auxiliary.
Import My_Substitutions.
Import My_Binders.

(********************************************************************)
(********************************************************************)
(********************************************************************)

(** Relation: parallel type reduction *)
Inductive RedTypePar: TypeS -> TypeS -> Prop :=

  (* Reflexivity *)

  | RTP_class:
    forall (C: ClassSym),
      (RedTypePar (Type_class C) (Type_class C))

  | RTP_var:
    forall (X: TypeVar),
      (RedTypePar (Type_var X) (Type_var X))

  | RTP_none:
    (RedTypePar Type_none Type_none)

  | RTP_nil:
    (RedTypePar Type_nil Type_nil)

  (* Constructors *)

  | RTP_proj_sub:
    forall (R R': TypeS) (i: nat),
      (RedTypePar R R') ->
      (RedTypePar (Type_proj R i) (Type_proj R' i))

  | RTP_proj_redex:
    forall (Ks Ks' Ki: TypeS) (i: nat),
      (RedTypePar Ks Ks') ->
      (TypeS_at Ks' i Ki ) ->
      (RedTypePar (Type_proj (Type_tuple Ks) i) Ki)

  | RTP_fun_left_right:
    forall (Pi Pi' T T': TypeS),
      (RedTypePar Pi Pi') ->
      (RedTypePar T T') ->
      (RedTypePar (Type_fun Pi T) (Type_fun Pi' T'))

  (* Types *)

  | RTP_apply_redex:
    forall (Pi Pi' T T' R R': TypeS),
      (RedTypePar Pi Pi') ->
      (RedTypePar T T') ->
      (RedTypePar R R') ->
      (RedTypePar (Type_apply (Type_fun Pi T) R) (typeSubst_in_type T' 0 R'))

  | RTP_apply_left_right:
    forall (K K': TypeS) (R R': TypeS),
      (RedTypePar K K') ->
      (RedTypePar R R') ->
      (RedTypePar (Type_apply K R) (Type_apply K' R'))

  (* Type tuples *)

  | RTP_tuple:
    forall (Ks Ks': TypeS),
    (RedTypePar Ks Ks') ->
    (RedTypePar (Type_tuple Ks) (Type_tuple Ks'))

  (* Type parameters *)

  | RTP_param_left_right:
    forall (Pi Pi' T T': TypeS),
      (RedTypePar Pi Pi') ->
      (RedTypePar T T') ->
      (RedTypePar (Type_param Pi T) (Type_param Pi' T'))

  (* Type sections *)

  | RTP_section:
    forall (Ps Ps': TypeS),
      (RedTypePar Ps Ps') ->
      (RedTypePar (Type_section Ps) (Type_section Ps'))

  (* Type options *)

  | RTP_some:
    forall (T T': TypeS),
    (RedTypePar T T') ->
    (RedTypePar (Type_some T) (Type_some T'))

  (* Type parameter and type constructor lists *)

  | RTP_cons_head_tail:
    forall (t t' ts ts': TypeS),
      (RedTypePar t t') ->
      (RedTypePar ts ts') ->
      (RedTypePar (Type_cons t ts) (Type_cons t' ts'))
.

(** Parallel type reduction is reflexive **)
Lemma RTP_refl:
  forall (t: TypeS),
    (RedTypePar t t)
.

Proof.
intros. induction t.
apply RTP_proj_sub; trivial.
apply RTP_class; trivial.
apply RTP_fun_left_right; trivial.
apply RTP_apply_left_right; trivial.
apply RTP_var; trivial.
apply RTP_tuple; trivial.
apply RTP_param_left_right; trivial.
apply RTP_section; trivial.
apply RTP_none; trivial.
apply RTP_some; trivial.
apply RTP_nil; trivial.
apply RTP_cons_head_tail; trivial.
Qed.

(** Reduction is included in parallel reduction **)
Lemma RedType_included_in_RedTypePar:
  forall (t t': TypeS),
    (RedType t t') ->
    (RedTypePar t t')
.

Proof.
intros. induction H.
(* RT-proj-sub *)
apply RTP_proj_sub. trivial.
(* RT-proj-redex *)
eapply RTP_proj_redex. apply RTP_refl. trivial.
(* RT-fun-left *)
apply RTP_fun_left_right. trivial. apply RTP_refl.
(* RT-fun-right *)
apply RTP_fun_left_right. apply RTP_refl. trivial.
(* RT-fun-redex *)
eapply RTP_apply_redex. apply RTP_refl. apply RTP_refl. apply RTP_refl.
(* RT-apply-left *)
apply RTP_apply_left_right. trivial. apply RTP_refl.
(* RT-apply-right *)
apply RTP_apply_left_right. apply RTP_refl. trivial.
(* RT-tuple *)
apply RTP_tuple. trivial.
(* RT-param-left *)
apply RTP_param_left_right. trivial. apply RTP_refl.
(* RT-param-right *)
apply RTP_param_left_right. apply RTP_refl. trivial.
(* RT-section *)
apply RTP_section. trivial.
(* RT-some *)
apply RTP_some. trivial.
(* RT-cons-head *)
eapply RTP_cons_head_tail. trivial. apply RTP_refl.
(* RT-cons-tail *)
eapply RTP_cons_head_tail. apply RTP_refl. trivial.
Qed.

(** Parallel reduction and type lifting **)
Lemma Parallel_reduction_in_lifting:
  forall (t t': TypeS) (n: nat),
    (RedTypePar t t') ->
    forall (k: nat),
      (RedTypePar (lift_at_depth_in_type t n k) (lift_at_depth_in_type t' n k))
.

Proof.
intros t t' n H. induction H.
(* RTP-class *)
intro. simpl. apply RTP_class.
(* RTP-var *)
intro. simpl. apply RTP_var.
(* RTP-none *)
intro. simpl. apply RTP_none.
(* RTP-nil *)
intro. simpl. apply RTP_nil.
(* RTP-proj-sub *)
intro. simpl. apply RTP_proj_sub. apply IHRedTypePar.
(* RTP-proj-redex *)
intro. simpl. eapply RTP_proj_redex. apply IHRedTypePar.
apply TypeS_at_of_lift. trivial.
(* RTP-fun-left-right *)
intro. simpl. apply RTP_fun_left_right. apply IHRedTypePar1. apply IHRedTypePar2.
(* RTP-apply-redex *)
intro. simpl.
replace  (lift_at_depth_in_type (typeSubst_in_type T' 0 R') n k) with
(typeSubst_in_type (lift_at_depth_in_type T' n (S k)) 0 (lift_at_depth_in_type R' n k)).
eapply RTP_apply_redex. apply IHRedTypePar1. apply IHRedTypePar2.
apply IHRedTypePar3.
pattern k at 1. replace (k) with (k + 0).
pattern k at 3. replace (k) with (k + 0).
rewrite Permuting_type_lifting_and_substitution_2_aux. trivial.
auto with arith. auto with arith.
(* RTP-apply-left-right *)
intro. simpl. apply RTP_apply_left_right. apply IHRedTypePar1. apply IHRedTypePar2.
(* RTP-tuple *)
intro. simpl. apply RTP_tuple. apply IHRedTypePar.
(* RTP-param-left-right *)
intro. simpl. apply RTP_param_left_right. apply IHRedTypePar1. apply IHRedTypePar2.
(* RTP-section *)
intro. simpl. apply RTP_section. apply IHRedTypePar.
(* RTP-some *)
intro. simpl. apply RTP_some. apply IHRedTypePar.
(* RTP-cons-head-tail *)
intro. simpl. apply RTP_cons_head_tail. apply IHRedTypePar1. apply IHRedTypePar2.
Qed.

(** Parallel reduction and type substitution **)
Lemma Reduction_and_substitution_par:
  forall (t t' R R': TypeS) (k: nat),
    (RedTypePar t t') ->
    (RedTypePar R R') ->
    (RedTypePar (typeSubst_in_type t k R) (typeSubst_in_type t' k R'))
.

Proof.
intros t t' R R' k H1 H2. generalize t' H1 k. clear t' H1 k. induction t.
(* proj *)
intros. simpl.
inversion H1.
(* - RTP-proj-sub *)
clear R0 i H H3. simpl. apply RTP_proj_sub. apply IHt. trivial.
(* - RTP-proj-redex *)
clear Ki i H0 H4. simpl.
assert (RedTypePar (typeSubst_in_type (Type_tuple Ks) k R)
                   (typeSubst_in_type (Type_tuple Ks') k R')).
rewrite <- H in IHt. apply IHt. apply RTP_tuple. trivial. clear IHt.
simpl in H0. inversion H0.
(* -- RTP-tuple *)
clear H4 H6. eapply RTP_proj_redex. apply H7. apply TypeS_at_of_typeSubst. trivial.
(* class *)
intros. inversion H1. simpl. apply RTP_refl.
(* fun *)
intros. inversion H1.
(* - RTP-fun-left-right *)
simpl. apply RTP_fun_left_right. apply IHt1. trivial. apply IHt2. trivial.
(* apply *)
intros. inversion H1.
(* - RTP-apply-redex *)
clear R0 H0.
simpl.
replace  (typeSubst_in_type (typeSubst_in_type T' 0 R'0) k R') with
  (typeSubst_in_type (typeSubst_in_type T' (S k) R') 0 (typeSubst_in_type R'0 k R')).
assert (RedTypePar (typeSubst_in_type (Type_fun Pi T) k R)
                   (typeSubst_in_type (Type_fun Pi' T') k R')).
rewrite <- H in IHt1. apply IHt1. apply RTP_fun_left_right; trivial.
simpl in H0. inversion H0.
(* -- RTP-fun-left-right *)
eapply RTP_apply_redex. apply H10. apply H12. apply IHt2. trivial.
pattern k at 3. replace (k) with (0 + k).
rewrite Permuting_type_substitutions. trivial. trivial.
(* - RTP-apply-left-right *)
simpl. apply RTP_apply_left_right. apply IHt1; trivial. apply IHt2; trivial.
(* var *)
intros.
rename t into X.
simpl.
inversion H1. simpl.
unfold typeSubst_in_typeVar.
induction (lt_eq_lt_dec X k). induction a.
(* X < k *)
rewrite match_comp_nat3_lt.
rewrite match_comp_nat3_lt. apply RTP_refl. trivial. trivial.
(* X = K *)
rewrite match_comp_nat3_eq.
rewrite match_comp_nat3_eq.
unfold lift_in_type.
apply Parallel_reduction_in_lifting; trivial.
trivial. trivial.
(* X > k *)
rewrite match_comp_nat3_gt.
rewrite match_comp_nat3_gt. apply RTP_refl. trivial. trivial.
(* tuple *)
intros. simpl. inversion H1.
(* - RTP-tuple *)
simpl. apply RTP_tuple. apply IHt. trivial.
(* param *)
intros. inversion H1.
(* - RTP-param-left-right *)
simpl. apply RTP_param_left_right. apply IHt1. trivial. apply IHt2. trivial.
(* section *)
intros. simpl. inversion H1.
(* - RTP-section *)
simpl. apply RTP_section. apply IHt. trivial.
(* none *)
intros. simpl. inversion H1. simpl. apply RTP_none.
(* some *)
intros. simpl. inversion H1.
(* - RTP-some *)
simpl. apply RTP_some. apply IHt. trivial.
(* nil *)
intros. simpl. inversion H1. simpl. apply RTP_nil.
(* cons *)
intros. inversion H1.
(* - RTP-cons-head-tail *)
simpl. apply RTP_cons_head_tail. apply IHt1. trivial. apply IHt2. trivial.
Qed.

(** Parallel reduction is included in multi-step reduction **)
Lemma RedTypePar_included_in_RedTypeStar:
  forall (t t': TypeS),
    (RedTypePar t t') ->
    (RedTypeStar t t')
.

Proof.
intros t t' H. induction H.
(* RTP-class *)
apply RedType_refl.
(* RTP-var *)
apply RedType_refl.
(* RTP-none *)
apply RedType_refl.
(* RTP-nil *)
apply RedType_refl.
(* (\* RTP-trans *\) *)
(* eapply RedType_trans. apply IHRedTypePar1. trivial. *)
(* RTP-proj-sub *)
apply RTS_proj_sub. trivial.
(* RTP-proj-redex *)
eapply RedType_trans. apply RTS_proj_sub. apply RTS_tuple. apply IHRedTypePar.
apply RedType_step. apply RT_proj_redex. trivial.
(* RTP-fun-left-right *)
eapply RedType_trans. apply RTS_fun_left. apply IHRedTypePar1.
apply RTS_fun_right. trivial.
(* RTP-apply-redex *)
eapply RedType_trans. apply RTS_apply_left. apply RTS_fun_right. apply IHRedTypePar2.
eapply RedType_trans. apply RTS_apply_right. apply IHRedTypePar3.
apply RedType_step. apply RT_apply_redex.
(* RTP-apply-left-right *)
eapply RedType_trans. apply RTS_apply_left. apply IHRedTypePar1.
apply RTS_apply_right. trivial.
(* RTP-tuple *)
apply RTS_tuple. trivial.
(* RTP-param-left-right *)
eapply RedType_trans. apply RTS_param_left. apply IHRedTypePar1.
apply RTS_param_right. trivial.
(* RTP-section *)
apply RTS_section. trivial.
(* RTP-some *)
apply RTS_some. trivial.
(* RTP-cons-head-tail *)
eapply RedType_trans. apply RTS_cons_head. apply IHRedTypePar1.
apply RTS_cons_tail. trivial.
Qed.

(** Parallel reduction and list lookup **)
Lemma Parallel_reduction_and_list_lookup:
  forall (ts: TypeS) (i: nat) (t: TypeS),
    (TypeS_at ts i t) ->
    forall (ts': TypeS),
      (RedTypePar ts ts') ->
      exists t': TypeS,
        (TypeS_at ts' i t') /\
        (RedTypePar t t')
.

Proof.
intros ts i t H. induction H.
(* TypeS_at_head *)
intros. inversion H.
(* - RTP-cons-head-tail *)
clear t0 ts0 H0 H1.
exists t'. split. apply TypeS_at_head. trivial.
(* TypeS_at_tail *)
intros. inversion H0.
(* - RTP-cons-head-tail *)
clear t0 ts0 H1 H2.
destruct (IHTypeS_at _ H5). destruct H1.
exists x. split. apply TypeS_at_tail. trivial. trivial.
Qed.

(** Structural size of a type **)
Fixpoint size (t: TypeS): nat :=
  match t with
    | (Type_proj R i)   => 1 + (size R)
    | (Type_class C)    => 1
    | (Type_fun Pi T)   => 1 + ((size Pi) + (size T))
    | (Type_apply K R)  => 1 + ((size K) + (size R))
    | (Type_var X)      => 1
    | (Type_tuple Ks)   => 1 + (size Ks)
    | (Type_param Pi T) => 1 + ((size Pi) + (size T))
    | (Type_section Ps) => 1 + (size Ps)
    | (Type_none)       => 1
    | (Type_some T)     => 1 + (size T)
    | (Type_nil)        => 1
    | (Type_cons u us)  => 1 + ((size u) + (size us))
  end
.

(** Parallel reduction is confluent *)
Lemma Parallel_reduction_is_confluent:
  forall (t u1 u2: TypeS),
    (RedTypePar t u1) ->
    (RedTypePar t u2) ->
    exists s: TypeS,
      (RedTypePar u1 s) /\
      (RedTypePar u2 s)
.

Proof.
cut (
  forall (n: nat) (t: TypeS),
    (n = (size t)) ->
    forall (u1 u2: TypeS),
      (RedTypePar t u1) ->
      (RedTypePar t u2) ->
      exists s: TypeS,
        (RedTypePar u1 s) /\
        (RedTypePar u2 s)
).
intros. eapply H with (t := t) (n := (size t)); trivial.
intro n. induction (lt_wf n). rename H0 into IH. clear H.
intros. rewrite H in IH. clear x H.
destruct H0.

(* RTP-class *)
exists u2. split. trivial. apply RTP_refl.
(* RTP-var *)
exists u2. split. trivial. apply RTP_refl.
(* RTP-none *)
exists u2. split. trivial. apply RTP_refl.
(* RTP-nil *)
exists u2. split. trivial. apply RTP_refl.

(* RTP-proj-sub *)
inversion H1.
(* - RTP-proj-sub *)
assert (A1 : exists R'': TypeS, (RedTypePar R' R'') /\ (RedTypePar R'0 R'')).
eapply IH with (y := (size R)) (t := R).
simpl. omega. trivial. trivial. trivial. destruct A1. destruct H5.
exists (Type_proj x i). split. apply RTP_proj_sub; trivial.
apply RTP_proj_sub; trivial.
(* - RTP-proj-redex *)
clear Ki i0 H2 H4.
rewrite <- H in H0.
inversion H0.
rename Ks'0 into Ks''.
assert (A1 : exists Ks1: TypeS, (RedTypePar Ks' Ks1) /\ (RedTypePar Ks'' Ks1)).
eapply IH with (y := (size Ks)) (t := Ks).
rewrite <- H.
simpl. omega. trivial. trivial. trivial. destruct A1. destruct H7.
destruct (Parallel_reduction_and_list_lookup _ _ _ H5 _ H7). destruct H9.
exists x0. split. eapply RTP_proj_redex. apply H8. trivial. trivial.

(* RTP-proj-redex *)
inversion H1.
(* - RTP-proj-sub *)
inversion H5.
(* exists Ks1 *)
assert (A1 : exists Ks1: TypeS, (RedTypePar Ks' Ks1) /\ (RedTypePar Ks'0 Ks1)).
eapply IH with (y := (size Ks)) (t := Ks).
simpl. omega. trivial. trivial. trivial. destruct A1. destruct H9.
rename x into Ks1.
destruct (Parallel_reduction_and_list_lookup _ _ _ H _ H9). destruct H11.
exists x. split. trivial. eapply RTP_proj_redex. apply H10. trivial.
(* - RTP-proj-redex *)
(* exists Ks1 *)
assert (A1 : exists Ks1: TypeS, (RedTypePar Ks' Ks1) /\ (RedTypePar Ks'0 Ks1)).
eapply IH with (y := (size Ks)) (t := Ks).
simpl. omega. trivial. trivial. trivial. destruct A1. destruct H7.
rename x into Ks1.
(* *)
destruct (Parallel_reduction_and_list_lookup _ _ _ H _ H7). destruct H9.
destruct (Parallel_reduction_and_list_lookup _ _ _ H6 _ H8). destruct H11.
exists x. split. trivial.
replace x with x0. trivial.
eapply TypeS_at_is_a_partial_function. apply H11. trivial.

(* RTP-fun-left-right *)
inversion H1.
rename Pi'0 into Pi''. rename T'0 into T''.
(* exists Pi1 *)
assert (A1 : exists Pi1: TypeS, (RedTypePar Pi' Pi1) /\ (RedTypePar Pi'' Pi1)).
eapply IH with (y := (size Pi)) (t := Pi).
simpl. omega. trivial. trivial. trivial. destruct A1. destruct H5.
rename x into Pi1.
(* exists T1 *)
assert (A1 : exists T1: TypeS, (RedTypePar T' T1) /\ (RedTypePar T'' T1)).
eapply IH with (y := (size T)) (t := T).
simpl. omega. trivial. trivial. trivial. destruct A1. destruct H7.
rename x into T1.
exists (Type_fun Pi1 T1). split. apply RTP_fun_left_right; trivial.
apply RTP_fun_left_right; trivial.

(* RTP-apply-redex *)
inversion H1.
(* - RTP-apply-redex *)
(* exists T1 *)
assert (A1 : exists T1: TypeS, (RedTypePar T' T1) /\ (RedTypePar T'0 T1)).
eapply IH with (y := (size T)) (t := T).
simpl. omega. trivial. trivial. trivial. destruct A1. destruct H7.
rename x into T1.
(* exists R1 *)
assert (A1 : exists R1: TypeS, (RedTypePar R' R1) /\ (RedTypePar R'0 R1)).
eapply IH with (y := (size R)) (t := R).
simpl. omega. trivial. trivial. trivial. destruct A1. destruct H9.
rename x into R1.
exists (typeSubst_in_type T1 0 R1). split.
apply Reduction_and_substitution_par; trivial.
apply Reduction_and_substitution_par; trivial.
(* - RTP-apply-left-right *)
inversion H2.
(* exists T1 *)
assert (A1 : exists T1: TypeS, (RedTypePar T' T1) /\ (RedTypePar T'0 T1)).
eapply IH with (y := (size T)) (t := T).
simpl. omega. trivial. trivial. trivial. destruct A1. destruct H10.
rename x into T1.
(* exists R1 *)
assert (A1 : exists R1: TypeS, (RedTypePar R' R1) /\ (RedTypePar R'0 R1)).
eapply IH with (y := (size R)) (t := R).
simpl. omega. trivial. trivial. trivial. destruct A1. destruct H12.
rename x into R1.
exists (typeSubst_in_type T1 0 R1). split.
apply Reduction_and_substitution_par; trivial.
eapply RTP_apply_redex. apply RTP_refl. trivial. trivial.

(* RTP-apply-left-right *)
inversion H1.
(* - RTP-apply-redex *)
clear R0 H0. rewrite <- H in H0_. inversion H0_. clear Pi0 T0 H0 H6.
(* exists T1 *)
assert (A1 : exists T1: TypeS, (RedTypePar T' T1) /\ (RedTypePar T'0 T1)).
eapply IH with (y := (size T)) (t := T).
rewrite <- H.
simpl. omega. trivial. trivial. trivial. destruct A1. destruct H0.
rename x into T1.
(* exists R1 *)
assert (A1 : exists R1: TypeS, (RedTypePar R' R1) /\ (RedTypePar R'0 R1)).
eapply IH with (y := (size R)) (t := R).
rewrite <- H.
simpl. omega. trivial. trivial. trivial. destruct A1. destruct H10.
rename x into R1.
exists (typeSubst_in_type T1 0 R1). split.
eapply RTP_apply_redex. apply RTP_refl. trivial. trivial.
apply Reduction_and_substitution_par; trivial.

(* - RTP-apply-left-right *)
(* exists K1 *)
assert (A1 : exists K1: TypeS, (RedTypePar K' K1) /\ (RedTypePar K'0 K1)).
eapply IH with (y := (size K)) (t := K).
simpl. omega. trivial. trivial. trivial. destruct A1. destruct H5.
rename x into K1.
(* exists R1 *)
assert (A1 : exists R1: TypeS, (RedTypePar R' R1) /\ (RedTypePar R'0 R1)).
eapply IH with (y := (size R)) (t := R).
simpl. omega. trivial. trivial. trivial. destruct A1. destruct H7.
rename x into R1.
exists (Type_apply K1 R1). split. apply RTP_apply_left_right; trivial.
apply RTP_apply_left_right; trivial.

(* RTP-tuple *)
inversion H1.
(* exists Ks *)
assert (A1 : exists Ks1: TypeS, (RedTypePar Ks' Ks1) /\ (RedTypePar Ks'0 Ks1)).
eapply IH with (y := (size Ks)) (t := Ks).
simpl. omega. trivial. trivial. trivial. destruct A1. destruct H4.
rename x into Ks1.
exists (Type_tuple Ks1). split.
apply RTP_tuple; trivial. apply RTP_tuple; trivial.

(* RTP-param-left-right *)
inversion H1.
rename Pi'0 into Pi''. rename T'0 into T''.
(* exists Pi1 *)
assert (A1 : exists Pi1: TypeS, (RedTypePar Pi' Pi1) /\ (RedTypePar Pi'' Pi1)).
eapply IH with (y := (size Pi)) (t := Pi).
simpl. omega. trivial. trivial. trivial. destruct A1. destruct H5.
rename x into Pi1.
(* exists T1 *)
assert (A1 : exists T1: TypeS, (RedTypePar T' T1) /\ (RedTypePar T'' T1)).
eapply IH with (y := (size T)) (t := T).
simpl. omega. trivial. trivial. trivial. destruct A1. destruct H7.
rename x into T1.
exists (Type_param Pi1 T1). split. apply RTP_param_left_right; trivial.
apply RTP_param_left_right; trivial.

(* RTP-section *)
inversion H1.
(* exists Ps1 *)
assert (A1 : exists Ps1: TypeS, (RedTypePar Ps' Ps1) /\ (RedTypePar Ps'0 Ps1)).
eapply IH with (y := (size Ps)) (t := Ps).
simpl. omega. trivial. trivial. trivial. destruct A1. destruct H4.
rename x into Ps1.
exists (Type_section Ps1). split.
apply RTP_section; trivial. apply RTP_section; trivial.

(* RTP-some *)
inversion H1.
(* exists T1 *)
assert (A1 : exists T1: TypeS, (RedTypePar T' T1) /\ (RedTypePar T'0 T1)).
eapply IH with (y := (size T)) (t := T).
simpl. omega. trivial. trivial. trivial. destruct A1. destruct H4.
rename x into T1.
exists (Type_some T1). split.
apply RTP_some; trivial. apply RTP_some; trivial.

(* RTP-cons-head-tail *)
inversion H1.
(* exists t1 *)
assert (A1 : exists t1: TypeS, (RedTypePar t' t1) /\ (RedTypePar t'0 t1)).
eapply IH with (y := (size t)) (t := t).
simpl. omega. trivial. trivial. trivial. destruct A1. destruct H5.
rename x into t1.
(* exists ts1 *)
assert (A1 : exists ts1: TypeS, (RedTypePar ts' ts1) /\ (RedTypePar ts'0 ts1)).
eapply IH with (y := (size ts)) (t := ts).
simpl. omega. trivial. trivial. trivial. destruct A1. destruct H7.
rename x into ts1.
exists (Type_cons t1 ts1). split. apply RTP_cons_head_tail; trivial.
apply RTP_cons_head_tail; trivial.
Qed.

(********************************************************************)
(* General lemmas about notions of reduction                        *)
(********************************************************************)

Definition SubRelation (A: Set) (R1 R2: A -> A -> Prop): Prop :=
  forall (x y: A), (R1 x y) -> (R2 x y)
.

Lemma Transitive_closure_preserves_inclusion:
  forall (A: Set) (R1 R2: A -> A -> Prop),
    (SubRelation _ R1 R2) ->
    (SubRelation _ (clos_trans _ R1) (clos_trans _ R2))
.

Proof.
intros. unfold SubRelation.
intros. induction H0.
apply t_step. apply H. trivial.
eapply t_trans. apply IHclos_trans1. trivial.
Qed.

Definition DiamondProperty (A: Set) (R: A -> A -> Prop): Prop :=
  forall (x y1 y2: A), (R x y1) -> (R x y2) ->
    exists z: A, (R y1 z) /\ (R y2 z)
.

Definition DiamondProperty_aux (A: Set) (R: A -> A -> Prop): Prop :=
  forall (x y1 y2: A), (clos_trans _ R x y1) -> (R x y2) ->
    exists z: A, (R y1 z) /\ (clos_trans _ R y2 z)
.

Lemma Transitive_closure_preserves_diamond_property_aux:
  forall (A: Set) (R: A -> A -> Prop),
    (DiamondProperty _ R) ->
    (DiamondProperty_aux _ R)
.

Proof.
intros. unfold DiamondProperty_aux. intros x y1 y2 H1. generalize y2. clear y2.
induction H1.
(* t_step *)
intros. destruct (H _ _ _ H0 H1). destruct H2.
exists x0. split. trivial. apply t_step; trivial.
(* t_trans *)
intros. destruct (IHclos_trans1 _ H0). destruct H1.
destruct (IHclos_trans2 _ H1). destruct H3.
exists x1. split. trivial. eapply t_trans. apply H2. trivial.
Qed.

Lemma Transitive_closure_preserves_diamond_property:
  forall (A: Set) (R: A -> A -> Prop),
    (DiamondProperty _ R) ->
    (DiamondProperty _ (clos_trans _ R))
.

Proof.
intros. unfold DiamondProperty. intros x y1 y2 H1. generalize y2. clear y2.
induction H1.
(* t_step *)
intros.
destruct (Transitive_closure_preserves_diamond_property_aux _ _ H _ _ _ H1 H0).
destruct H2.
exists x0. split. trivial. apply t_step; trivial.
(* t_trans *)
intros.
destruct (IHclos_trans1 _ H0). destruct H1.
destruct (IHclos_trans2 _ H1). destruct H3.
exists x1. split. trivial. eapply t_trans. apply H2. trivial.
Qed.

Lemma Equality_preserves_diamond_property:
  forall (A: Set) (R1 R2: A -> A -> Prop),
    (DiamondProperty _ R1) ->
    (SubRelation _ R1 R2) ->
    (SubRelation _ R2 R1) ->
    (DiamondProperty _ R2)
.

Proof.
intros. unfold DiamondProperty. intros.
assert (R1 x y1). apply H1; trivial. assert (R1 x y2). apply H1; trivial.
destruct (H _ _ _ H4 H5). destruct H6.
exists x0. split. apply H0; trivial. apply H0; trivial.
Qed.

(**************************************************************)

(** Transitive closure of parallel type reduction **)
Definition RedTypeParTrans: TypeS -> TypeS -> Prop :=
  (clos_trans _ RedTypePar).

Lemma RedTypeParTrans_has_diamond_property:
  (DiamondProperty _ RedTypeParTrans)
.

Proof.
unfold RedTypeParTrans.
apply Transitive_closure_preserves_diamond_property.
unfold DiamondProperty. apply Parallel_reduction_is_confluent.
Qed.

Lemma RedTypeStar_included_in_RedTypeParTrans:
  forall (t t': TypeS),
    (RedTypeStar t t') ->
    (RedTypeParTrans t t')
.

Proof.
intros. unfold RedTypeParTrans. induction H.
apply t_step. apply RedType_included_in_RedTypePar. trivial.
apply t_step. apply RTP_refl.
eapply t_trans. apply IHRedTypeStar1. trivial.
Qed.

Lemma RedTypeParTrans_included_in_RedTypeStar:
  forall (t t': TypeS),
    (RedTypeParTrans t t') ->
    (RedTypeStar t t')
.

Proof.
intros. unfold RedTypeParTrans in H. induction H.
apply RedTypePar_included_in_RedTypeStar. trivial.
eapply RedType_trans.  apply IHclos_trans1. trivial.
Qed.

(** Type reduction is confluent *)
Theorem Type_reduction_is_confluent:
  forall (t u1 u2: TypeS),
    (RedTypeStar t u1) ->
    (RedTypeStar t u2) ->
    exists s: TypeS,
      (RedTypeStar u1 s) /\
      (RedTypeStar u2 s)
.

Proof.
apply (Equality_preserves_diamond_property _ RedTypeParTrans).
apply RedTypeParTrans_has_diamond_property.
unfold SubRelation. apply RedTypeParTrans_included_in_RedTypeStar.
unfold SubRelation. apply RedTypeStar_included_in_RedTypeParTrans.
Qed.

End SetProgram.