(* Author: Vincent Cremet *)
(* Date: July 2007 *) 
(* File: SimplyTypedLambdaCalculus.v *)
(* Description: Simply typed lambda-calculus with pairs *)

(** * Simply typed lambda-calculus *)

Require Import List.
Require Import Arith.
Require Import Omega.

Require Import Util.
Require Import MatchCompNat.

(***********************************************************)

Module Type LambdaConstants.

Parameter LConst: Set.

(** Types - S, T, U *)
Inductive LType: Set :=
  | LType_base : LType                    (*   *    *)
  | LType_fun  : LType -> LType -> LType  (* T -> U *)
  | LType_prod : LType -> LType -> LType  (* (T, U) *)
.

Parameter type_of_const: LConst -> LType -> Prop.

End LambdaConstants.

(***********************************************************)

Module LambdaCalculus(My_LambdaConstants: LambdaConstants).

Import My_LambdaConstants.

(** Terms - s, t, u *)
Inductive LTerm: Set :=
  | LTerm_const : LConst -> LTerm          (* c *)
  | LTerm_var   : nat -> LTerm             (* x *)
  | LTerm_fun   : LType -> LTerm -> LTerm  (* fun x: T => t *)
  | LTerm_app   : LTerm -> LTerm -> LTerm  (* t u *)
  | LTerm_pair  : LTerm -> LTerm -> LTerm  (* (t, u) *)             
  | LTerm_fst   : LTerm -> LTerm           (* fst(t) *)
  | LTerm_snd   : LTerm -> LTerm           (* snd(t) *)
.

(** Function: lifts variable [i] with amplitude [n] at depth [k] *)
Definition lift_at_depth_in_var (i: nat) (k: nat) (n: nat): nat :=
  (match_comp_nat2 i k nat
    i
    (i + n)
  )
.

(** Function: lifts term [t] with amplitude [n] at depth [k] *)
Fixpoint lift_at_depth_in_term (t': LTerm) (k: nat) (n: nat) {struct t'}: LTerm :=
  match t' with
    | (LTerm_const c)  => (LTerm_const c)
    | (LTerm_var i)    => (LTerm_var (lift_at_depth_in_var i k n))
    | (LTerm_fun T t)  => (LTerm_fun T (lift_at_depth_in_term t (S k) n))
    | (LTerm_app t u)  => (LTerm_app (lift_at_depth_in_term t k n)
                                     (lift_at_depth_in_term u k n))
    | (LTerm_pair t u) => (LTerm_pair (lift_at_depth_in_term t k n)
                                      (lift_at_depth_in_term u k n))
    | (LTerm_fst t)    => (LTerm_fst (lift_at_depth_in_term t k n))
    | (LTerm_snd t)    => (LTerm_snd (lift_at_depth_in_term t k n))
  end
.

(** Function: lifts term [t] with amplitude [n] *)
Definition lift_in_term (t: LTerm) (n: nat): LTerm :=
  (lift_at_depth_in_term t 0 n)
.

(** Function: substitute [s] for [k] in [i], noted i[k\s] **)
Definition subst_in_var (i: nat) (k: nat) (s: LTerm): LTerm :=
  (match_comp_nat3 i k LTerm
    (LTerm_var i)
    (lift_in_term s k)
    (LTerm_var (pred i))
  )
.

(** Function: substitute [s] for [k] in [t'], noted t'[k\s] **)
Fixpoint subst_in_term (t': LTerm) (k: nat) (s: LTerm) {struct t'}: LTerm :=
  match t' with
    | (LTerm_const c)  => (LTerm_const c)
    | (LTerm_var i)    => (subst_in_var i k s)
    | (LTerm_fun T t)  => (LTerm_fun T (subst_in_term t (S k) s))
    | (LTerm_app t u)  => (LTerm_app (subst_in_term t k s) (subst_in_term u k s))
    | (LTerm_pair t u) => (LTerm_pair (subst_in_term t k s) (subst_in_term u k s))
    | (LTerm_fst t)    => (LTerm_fst (subst_in_term t k s))
    | (LTerm_snd t)    => (LTerm_snd (subst_in_term t k s))
  end
.

(** Typing context *)
Definition LEnv: Set := (list LType).

(** Relation: type assignment *)
Inductive WellTyped: LEnv -> LTerm -> LType -> Prop :=

  | WT_const:
    forall (G: LEnv) (c: LConst) (T: LType),
      (type_of_const c T) ->
      (WellTyped G (LTerm_const c) T)

  | WT_var:
    forall (G: LEnv) (i: nat) (T: LType),
      (List_from G i T) ->
      (WellTyped G (LTerm_var i) T)

  | WT_fun:
    forall (G: LEnv) (t: LTerm) (T U: LType),
      (WellTyped (snoc G T) t U) ->
      (WellTyped G (LTerm_fun T t) (LType_fun T U))

  | WT_app:
    forall (G: LEnv) (t u: LTerm) (T U: LType),
      (WellTyped G t (LType_fun T U)) ->
      (WellTyped G u T) ->
      (WellTyped G (LTerm_app t u) U)

  | WT_pair:
    forall (G: LEnv) (t u: LTerm) (T U: LType),
      (WellTyped G t T) ->
      (WellTyped G u U) ->
      (WellTyped G (LTerm_pair t u) (LType_prod T U))

  | WT_fst:
    forall (G: LEnv) (t: LTerm) (T U: LType),
      (WellTyped G t (LType_prod T U)) ->
      (WellTyped G (LTerm_fst t) T)

  | WT_snd:
    forall (G: LEnv) (t: LTerm) (T U: LType),
      (WellTyped G t (LType_prod T U)) ->
      (WellTyped G (LTerm_snd t) U)
.

(** Relation: reduction of terms **)
Inductive RedTerm: LTerm -> LTerm -> Prop :=

  | R_fun:
    forall (T: LType) (t t': LTerm),
      (RedTerm t t') ->
      (RedTerm (LTerm_fun T t) (LTerm_fun T t'))

  | R_app_left:
    forall (t t' u: LTerm),
      (RedTerm t t') ->
      (RedTerm (LTerm_app t u) (LTerm_app t' u))

  | R_app_right:
    forall (t u u': LTerm),
      (RedTerm u u') ->
      (RedTerm (LTerm_app t u) (LTerm_app t u'))

  | R_app_redex:
    forall (T: LType) (t u: LTerm),
      (RedTerm (LTerm_app (LTerm_fun T t) u) (subst_in_term t 0 u))

  | R_pair_left:
    forall (t t' u: LTerm),
      (RedTerm t t') ->
      (RedTerm (LTerm_pair t u) (LTerm_pair t' u))

  | R_pair_right:
    forall (t u u': LTerm),
      (RedTerm u u') ->
      (RedTerm (LTerm_pair t u) (LTerm_pair t u'))

  | R_fst_cong:
    forall (t t': LTerm),
      (RedTerm t t') ->
      (RedTerm (LTerm_fst t) (LTerm_fst t'))

  | R_fst_redex:
    forall (t u: LTerm),
      (RedTerm (LTerm_fst (LTerm_pair t u)) t)

  | R_snd_cong:
    forall (t t': LTerm),
      (RedTerm t t') ->
      (RedTerm (LTerm_snd t) (LTerm_snd t'))

  | R_snd_redex:
    forall (t u: LTerm),
      (RedTerm (LTerm_snd (LTerm_pair t u)) u)
.

(*****************************************************************)

(** Successive variable liftings *)
Lemma Successive_variable_liftings:
  forall (n1 n2 m: nat) (i: nat) (k: nat),
    (lift_at_depth_in_var (lift_at_depth_in_var i k (n1 + n2)) (k + n2) m) =
    (lift_at_depth_in_var i k (n1 + n2 + m))
.

Proof.
intros n1 n2 m i k.
unfold lift_at_depth_in_var.
induction (match_comp_nat2_aux_0 i k).
(* i < k *)
pattern (match_comp_nat2 i k nat i (i + (n1 + n2))).
rewrite match_comp_nat2_lt.
rewrite match_comp_nat2_lt.
rewrite match_comp_nat2_lt. trivial. trivial. auto with arith. trivial.
(* i >= k *)
pattern (match_comp_nat2 i k nat i (i + (n1 + n2))).
rewrite match_comp_nat2_ge.
pattern (match_comp_nat2 i k nat i (i + (n1 + n2 + m))).
rewrite match_comp_nat2_ge.
rewrite match_comp_nat2_ge. auto with arith. omega. trivial. trivial.
Qed.

(** Successive type liftings (auxiliary) *)
Lemma Successive_type_liftings_aux:
  forall (n1 n2 m: nat) (t: LTerm) (k: nat),
    (lift_at_depth_in_term (lift_at_depth_in_term t k (n1 + n2)) (k + n2) m) =
    (lift_at_depth_in_term t k (n1 + n2 + m))
.

Proof.
intros n1 n2 m t. induction t.
(* const *)
intro. simpl. trivial.
(* var *)
intro. simpl. rewrite Successive_variable_liftings. trivial.
(* fun *)
intro. simpl. replace (S (k + n2)) with ((S k) + n2). rewrite IHt. trivial. trivial.
(* apply *)
intro. simpl. rewrite IHt1. rewrite IHt2. trivial.
(* pair *)
intro. simpl. rewrite IHt1. rewrite IHt2. trivial.
(* fst *)
intro. simpl. rewrite IHt. trivial.
(* snd *)
intro. simpl. rewrite IHt. trivial.
Qed.

(*****************************************************************)

(** Permuting type lifting and substitution (auxiliary) *)
Lemma Permuting_type_lifting_and_substitution_aux:
  forall (s: LTerm) (k1 n: nat) (t: LTerm) (k2: nat),
    (lift_at_depth_in_term (subst_in_term t (k1 + k2) s) k2 n) =
    (subst_in_term (lift_at_depth_in_term t k2 n) (k1 + k2 + n) s)
.

Proof.
intros s k1 n t. induction t.
(* const *)
intro. simpl. trivial.
(* var *)
rename n0 into i.
intro. simpl.
induction (lt_eq_lt_dec i (k1 + k2)). induction a.
(* i < k1 + k2 *)
unfold subst_in_var. rewrite match_comp_nat3_lt.
simpl.
induction (match_comp_nat2_aux_0 i k2).
unfold lift_at_depth_in_var. rewrite match_comp_nat2_lt.
rewrite match_comp_nat3_lt. trivial. omega. trivial.
unfold lift_at_depth_in_var. rewrite match_comp_nat2_ge.
rewrite match_comp_nat3_lt. trivial. omega. trivial. trivial.
(* i = k1 + k2 *)
unfold subst_in_var. rewrite match_comp_nat3_eq.
unfold lift_at_depth_in_var. rewrite match_comp_nat2_ge.
rewrite match_comp_nat3_eq.
unfold lift_in_term.
pattern k2 at 2.
replace (k2) with (0 + k2).
rewrite Successive_type_liftings_aux. trivial. trivial.
omega. omega. trivial.
(* i > k1 + k2 *)
unfold subst_in_var. rewrite match_comp_nat3_gt.
simpl.
unfold lift_at_depth_in_var. rewrite match_comp_nat2_ge.
rewrite match_comp_nat2_ge.
rewrite match_comp_nat3_gt.
replace (pred i + n) with (pred (i + n)). trivial.
omega. omega. omega. omega. trivial.
(* fun *)
intro. simpl.
replace (S (k1 + k2)) with (k1 + (S k2)).
replace (S (k1 + k2 + n)) with (k1 + (S k2) + n). rewrite IHt. trivial.
omega. auto with arith.
(* app *)
intro. simpl. rewrite IHt1. rewrite IHt2. trivial.
(* pair *)
intro. simpl. rewrite IHt1. rewrite IHt2. trivial.
(* fst *)
intro. simpl. rewrite IHt. trivial.
(* snd *)
intro. simpl. rewrite IHt. trivial.
Qed.

Lemma Permuting_type_lifting_and_substitution:
  forall (t: LTerm) (i: nat) (s: LTerm) (n: nat),
    (lift_in_term (subst_in_term t i s) n) =
    (subst_in_term (lift_in_term t n) (i + n) s)
.

Proof.
intros.
unfold lift_in_term.
replace (i) with (i + 0).
apply Permuting_type_lifting_and_substitution_aux.
auto with arith.
Qed.

Lemma Permuting_type_liftings_aux:
  forall (k1 n1 n2: nat) (t: LTerm) (k2: nat),
    (lift_at_depth_in_term (lift_at_depth_in_term t k2 n2) (k1 + n2 + k2) n1) =
    (lift_at_depth_in_term (lift_at_depth_in_term t (k1 + k2) n1) k2 n2)
.

Proof.
intros k1 n1 n2 t. induction t.
(* const *)
intro. simpl. trivial.
(* var *)
rename n into i.
intro. simpl.
unfold lift_at_depth_in_var.
induction (match_comp_nat2_aux_0 i k2).
(* i < k2 *)
rewrite (match_comp_nat2_lt i k2).
rewrite match_comp_nat2_lt.
rewrite (match_comp_nat2_lt i (k1 + k2)).
rewrite match_comp_nat2_lt. trivial.
trivial. omega. omega. trivial.
(* i >= k2 *)
rewrite (match_comp_nat2_ge i k2).
induction (match_comp_nat2_aux_0 i (k1 + k2)).
(* i < k1 + k2 *)
rewrite match_comp_nat2_lt.
rewrite (match_comp_nat2_lt i (k1 + k2)).
rewrite match_comp_nat2_ge. trivial.
trivial. trivial. omega.
(* i >= k1 + k2 *)
rewrite match_comp_nat2_ge. 
rewrite (match_comp_nat2_ge i (k1 + k2)).
rewrite match_comp_nat2_ge. replace (i + n2 + n1) with (i + n1 + n2). trivial.
omega. omega. trivial. omega. trivial.
(* fun *)
intro. simpl.
replace (S (k1 + n2 + k2)) with (k1 + n2 + (S k2)).
replace (S (k1 + k2)) with (k1 + (S k2)). rewrite IHt. trivial.
auto with arith. auto with arith.
(* app *)
intro. simpl. rewrite IHt1. rewrite IHt2. trivial.
(* pair *)
intro. simpl. rewrite IHt1. rewrite IHt2. trivial.
(* fst *)
intro. simpl. rewrite IHt. trivial.
(* snd *)
intro. simpl. rewrite IHt. trivial.
Qed.

(** Permuting type lifting and substitution 2 (auxiliary) *)
Lemma Permuting_type_lifting_and_substitution_2_aux:
  forall (s: LTerm) (k1 n: nat) (t: LTerm) (k2: nat),
    (lift_at_depth_in_term (subst_in_term t k2 s) (k1 + k2) n) =
    (subst_in_term (lift_at_depth_in_term t (S (k1 + k2)) n) k2 
      (lift_at_depth_in_term s k1 n))
.

Proof.
intros s k1 n t. induction t.
(* const *)
intro. simpl. trivial.
(* var *)
rename n0 into i.
intro. simpl.
induction (lt_eq_lt_dec i k2). induction a.
(* i < k2 *)
unfold subst_in_var. rewrite match_comp_nat3_lt.
simpl. unfold lift_at_depth_in_var. rewrite match_comp_nat2_lt.
rewrite match_comp_nat2_lt. rewrite match_comp_nat3_lt. trivial.
trivial. omega. omega. trivial.
(* i = k2 *)
unfold subst_in_var. rewrite match_comp_nat3_eq.
unfold lift_at_depth_in_var. rewrite match_comp_nat2_lt.
rewrite match_comp_nat3_eq. unfold lift_in_term.
replace (k1 + k2) with (k1 + k2 + 0).
pattern k1 at 2. replace (k1) with (k1 + 0).
rewrite Permuting_type_liftings_aux. trivial.
auto with arith. auto with arith.
trivial. omega. trivial.
(* i > k2 *)
unfold subst_in_var. rewrite match_comp_nat3_gt.
induction (match_comp_nat2_aux_0 i (S (k1 + k2))).
(* i < k1 + k2 + 1 *)
unfold lift_at_depth_in_var. rewrite match_comp_nat2_lt.
rewrite match_comp_nat3_gt.
simpl. unfold lift_at_depth_in_var. rewrite match_comp_nat2_lt. trivial.
omega. trivial. trivial.
(* i >= k1 + k2 + 1 *)
unfold lift_at_depth_in_var. rewrite match_comp_nat2_ge.
simpl. unfold lift_at_depth_in_var. rewrite match_comp_nat2_ge.
rewrite match_comp_nat3_gt. replace (pred i + n) with (pred (i + n)). trivial.
omega. omega. omega. trivial. trivial.
(* fun *)
intro. simpl.
replace (S (k1 + k2)) with (k1 + (S k2)).
replace (S (S (k1 + k2))) with (S (k1 + (S k2))).
rewrite IHt. trivial. auto with arith. auto with arith.
(* app *)
intro. simpl. rewrite IHt1. rewrite IHt2. trivial.
(* pair *)
intro. simpl. rewrite IHt1. rewrite IHt2. trivial.
(* fst *)
intro. simpl. rewrite IHt. trivial.
(* snd *)
intro. simpl. rewrite IHt. trivial.
Qed.

Lemma Type_lifting_stronger_than_substitution_aux:
  forall (n d: nat) (s t: LTerm) (k: nat),
    (subst_in_term (lift_at_depth_in_term t k (S (n + d))) (n + k) s) =
    (lift_at_depth_in_term t k (n + d))
.

Proof.
intros n d s t. induction t.
(* const *)
intro. simpl. trivial.
(* var *)
rename n0 into i.
intro. simpl.
induction (match_comp_nat2_aux_0 i k).
(* i < k *)
unfold lift_at_depth_in_var. rewrite match_comp_nat2_lt.
rewrite match_comp_nat2_lt.
unfold subst_in_var. rewrite match_comp_nat3_lt. trivial.
omega. trivial. trivial.
(* i >= k *)
unfold lift_at_depth_in_var. rewrite match_comp_nat2_ge.
rewrite match_comp_nat2_ge.
unfold subst_in_var. rewrite match_comp_nat3_gt.
replace (pred (i + S (n + d))) with (i + (n + d)). trivial.
omega. omega. trivial. trivial.
(* fun *)
intro. simpl.
replace (S (n + k)) with (n + (S k)). rewrite IHt. trivial.
auto with arith.
(* app *)
intro. simpl. rewrite IHt1. rewrite IHt2. trivial.
(* pair *)
intro. simpl. rewrite IHt1. rewrite IHt2. trivial.
(* fst *)
intro. simpl. rewrite IHt. trivial.
(* snd *)
intro. simpl. rewrite IHt. trivial.
Qed.

Lemma Type_lifting_stronger_than_substitution:
  forall (n d: nat) (s t: LTerm),
    (subst_in_term (lift_in_term t (S (n + d))) n s) =
    (lift_in_term t (n + d))
.

Proof.
intros.
unfold lift_in_term.
pattern n at 2. replace (n) with (n + 0).
apply Type_lifting_stronger_than_substitution_aux.
auto with arith.
Qed.

(** Permuting type substitutions *)
Lemma Permuting_type_substitutions:
  forall (s s': LTerm) (k: nat) (t: LTerm) (n: nat),
    (subst_in_term (subst_in_term t n s) (n + k) s') =
    (subst_in_term (subst_in_term t (S (n + k)) s') n (subst_in_term s k s'))
.

Proof.
intros s s' k t.
induction t.
(* const *)
intro. simpl. trivial.
(* var *)
rename n into i.
intro. simpl.
induction (lt_eq_lt_dec i n). induction a.
(* i < n *)
unfold subst_in_var. rewrite match_comp_nat3_lt.
simpl. unfold subst_in_var. rewrite match_comp_nat3_lt.
rewrite match_comp_nat3_lt. simpl.  unfold subst_in_var. rewrite match_comp_nat3_lt.
trivial.
trivial. auto with arith. auto with arith. trivial.
(* i = n *)
unfold subst_in_var. rewrite match_comp_nat3_eq.
rewrite match_comp_nat3_lt.
simpl. unfold subst_in_var. rewrite match_comp_nat3_eq.
rewrite Permuting_type_lifting_and_substitution.
replace (n + k) with (k + n). trivial. auto with arith.
trivial. omega. trivial.
(* i > n *)
unfold subst_in_var. rewrite match_comp_nat3_gt.
simpl.
induction (lt_eq_lt_dec i (S (n + k))). induction a.
(* i < n + k + 1 *)
unfold subst_in_var. rewrite match_comp_nat3_lt.
rewrite match_comp_nat3_lt.
simpl. unfold subst_in_var. rewrite match_comp_nat3_gt.
trivial.
trivial. trivial. omega.
(* i = n + k + 1 *)
unfold subst_in_var. rewrite match_comp_nat3_eq.
rewrite match_comp_nat3_eq.
rewrite Type_lifting_stronger_than_substitution. trivial.
trivial. omega.
(* i > n + k + 1 *)
unfold subst_in_var. rewrite match_comp_nat3_gt.
rewrite match_comp_nat3_gt.
simpl. unfold subst_in_var. rewrite match_comp_nat3_gt.
trivial. omega. trivial. omega. trivial.
(* fun *)
intro. simpl.
replace (S (n + k)) with ((S n) + k).
rewrite IHt. trivial. trivial.
(* app *)
intro. simpl.
rewrite IHt1. rewrite IHt2. trivial.
(* pair *)
intro. simpl.
rewrite IHt1. rewrite IHt2. trivial.
(* fst *)
intro. simpl.
rewrite IHt. trivial.
(* snd *)
intro. simpl.
rewrite IHt. trivial.
Qed.

(************************************************************************)

Lemma Substitution_in_reduction:
  forall (s t t': LTerm),
    (RedTerm t t') ->
    forall (k: nat),
      (RedTerm (subst_in_term t k s) (subst_in_term t' k s))
.

Proof.
intros s t t' H. induction H.
(* R-fun *)
intro. simpl. apply R_fun. apply IHRedTerm.
(* R-app-left *)
intro. simpl. apply R_app_left. apply IHRedTerm.
(* R-app-right *)
intro. simpl. apply R_app_right. apply IHRedTerm.
(* R-app-redex *)
intro. simpl.
replace  (subst_in_term (subst_in_term t 0 u) k s) with
  (subst_in_term (subst_in_term t (S k) s) 0 (subst_in_term u k s)).
apply R_app_redex.
pattern k at 3. replace (k) with (0 + k).
rewrite Permuting_type_substitutions. trivial. trivial.
(* R-pair-left *)
intro. simpl. apply R_pair_left. apply IHRedTerm.
(* R-pair-right *)
intro. simpl. apply R_pair_right. apply IHRedTerm.
(* R-fst-cong *)
intro. simpl. apply R_fst_cong. apply IHRedTerm.
(* RT-fst-redex *)
intro. simpl. apply R_fst_redex.
(* R-snd-cong *)
intro. simpl. apply R_snd_cong. apply IHRedTerm.
(* RT-snd-redex *)
intro. simpl. apply R_snd_redex.
Qed.

Inductive RedTermStar: LTerm -> LTerm -> Prop :=
  | RedTerm_step: forall (t t': LTerm), (RedTerm t t') -> (RedTermStar t t')
  | RedTerm_refl: forall (t: LTerm), (RedTermStar t t)
  | RedTerm_trans: forall (t u t': LTerm), (RedTermStar t u) -> (RedTermStar u t') ->
    (RedTermStar t t')
.

(* Definition RedTermStar: LTerm -> LTerm -> Prop := clos_refl_trans LTerm RedTerm. *)

Lemma RS_fun:
  forall (t t': LTerm) (T: LType),
    (RedTermStar t t') ->
    (RedTermStar (LTerm_fun T t) (LTerm_fun T t'))
.

Proof.
intros t t' T H. induction H.
(* step *)
apply RedTerm_step. apply R_fun. trivial.
(* refl *)
apply RedTerm_refl.
(* trans *)
eapply RedTerm_trans. apply IHRedTermStar1. apply IHRedTermStar2.
Qed.

Lemma RS_app_left:
  forall (t t' u: LTerm),
    (RedTermStar t t') ->
    (RedTermStar (LTerm_app t u) (LTerm_app t' u))
.

Proof.
intros t t' u H. induction H.
(* step *)
apply RedTerm_step. apply R_app_left. trivial.
(* refl *)
apply RedTerm_refl.
(* trans *)
eapply RedTerm_trans. apply IHRedTermStar1. apply IHRedTermStar2.
Qed.

Lemma RS_app_right:
  forall (t u u': LTerm),
    (RedTermStar u u') ->
    (RedTermStar (LTerm_app t u) (LTerm_app t u'))
.

Proof.
intros t u u' H. induction H.
(* step *)
apply RedTerm_step. apply R_app_right. trivial.
(* refl *)
apply RedTerm_refl.
(* trans *)
eapply RedTerm_trans. apply IHRedTermStar1. apply IHRedTermStar2.
Qed.

Lemma RS_pair_left:
  forall (t t' u: LTerm),
    (RedTermStar t t') ->
    (RedTermStar (LTerm_pair t u) (LTerm_pair t' u))
.

Proof.
intros t t' u H. induction H.
(* step *)
apply RedTerm_step. apply R_pair_left. trivial.
(* refl *)
apply RedTerm_refl.
(* trans *)
eapply RedTerm_trans. apply IHRedTermStar1. apply IHRedTermStar2.
Qed.

Lemma RS_pair_right:
  forall (t u u': LTerm),
    (RedTermStar u u') ->
    (RedTermStar (LTerm_pair t u) (LTerm_pair t u'))
.

Proof.
intros t u u' H. induction H.
(* step *)
apply RedTerm_step. apply R_pair_right. trivial.
(* refl *)
apply RedTerm_refl.
(* trans *)
eapply RedTerm_trans. apply IHRedTermStar1. apply IHRedTermStar2.
Qed.

Lemma RS_fst_cong:
  forall (t t': LTerm),
    (RedTermStar t t') ->
    (RedTermStar (LTerm_fst t) (LTerm_fst t'))
.

Proof.
intros t t' H. induction H.
(* step *)
apply RedTerm_step. apply R_fst_cong. trivial.
(* refl *)
apply RedTerm_refl.
(* trans *)
eapply RedTerm_trans. apply IHRedTermStar1. apply IHRedTermStar2.
Qed.

Lemma RS_snd_cong:
  forall (t t': LTerm),
    (RedTermStar t t') ->
    (RedTermStar (LTerm_snd t) (LTerm_snd t'))
.

Proof.
intros t t' H. induction H.
(* step *)
apply RedTerm_step. apply R_snd_cong. trivial.
(* refl *)
apply RedTerm_refl.
(* trans *)
eapply RedTerm_trans. apply IHRedTermStar1. apply IHRedTermStar2.
Qed.

Lemma Reduction_in_lifting:
  forall (t t': LTerm) (n: nat),
    (RedTerm t t') ->
    forall (k: nat),
      (RedTerm (lift_at_depth_in_term t k n) (lift_at_depth_in_term t' k n))
.

Proof.
intros t t' n H. induction H.
(* R-fun *)
intro. simpl. apply R_fun. apply IHRedTerm.
(* R-app-left *)
intro. simpl. apply R_app_left. apply IHRedTerm.
(* R-app-right *)
intro. simpl. apply R_app_right. apply IHRedTerm.
(* R-app-redex *)
intro. simpl.
replace  (lift_at_depth_in_term (subst_in_term t 0 u) k n) with
(subst_in_term (lift_at_depth_in_term t (S k) n) 0 (lift_at_depth_in_term u k n)).
apply R_app_redex.
pattern k at 1. replace (k) with (k + 0).
pattern k at 3. replace (k) with (k + 0).
rewrite Permuting_type_lifting_and_substitution_2_aux. trivial.
auto with arith. auto with arith.
(* R-pair-left *)
intro. simpl. apply R_pair_left. apply IHRedTerm.
(* R-pair-right *)
intro. simpl. apply R_pair_right. apply IHRedTerm.
(* R-fst-cong *)
intro. simpl. apply R_fst_cong. apply IHRedTerm.
(* R-fst-redex *)
intro. simpl. apply R_fst_redex.
(* R-snd-cong *)
intro. simpl. apply R_snd_cong. apply IHRedTerm.
(* R-snd-redex *)
intro. simpl. apply R_snd_redex.
Qed.

Lemma Reduction_in_substitution:
  forall (s s': LTerm),
    (RedTerm s s') ->
    forall (t: LTerm) (k: nat),
      (RedTermStar (subst_in_term t k s) (subst_in_term t k s'))
.

Proof.
intros s s' H t. induction t.
(* const *)
intro. simpl. apply RedTerm_refl.
(* var *)
rename n into i.
intro. simpl.
unfold subst_in_var.
induction (lt_eq_lt_dec i k). induction a.
(* i < k *)
rewrite match_comp_nat3_lt.
rewrite match_comp_nat3_lt. apply RedTerm_refl. trivial. trivial.
(* i = K *)
rewrite match_comp_nat3_eq.
rewrite match_comp_nat3_eq.
apply RedTerm_step.
unfold lift_in_term.
apply Reduction_in_lifting.
trivial. trivial. trivial.
(* i > k *)
rewrite match_comp_nat3_gt.
rewrite match_comp_nat3_gt. apply RedTerm_refl. trivial. trivial.
(* fun *)
intro. simpl.
eapply RedTerm_trans. apply RS_fun. apply IHt.
apply RedTerm_refl.
(* app *)
intro. simpl.
eapply RedTerm_trans. apply RS_app_left. apply IHt1.
eapply RedTerm_trans. apply RS_app_right. apply IHt2.
apply RedTerm_refl.
(* pair *)
intro. simpl.
eapply RedTerm_trans. apply RS_pair_left. apply IHt1.
eapply RedTerm_trans. apply RS_pair_right. apply IHt2.
apply RedTerm_refl.
(* fst *)
intro. simpl. apply RS_fst_cong. apply IHt.
(* snd *)
intro. simpl. apply RS_snd_cong. apply IHt.
Qed.

Lemma Substitution_in_reduction_star:
  forall (s t t': LTerm) (k: nat),
    (RedTermStar t t') ->
    (RedTermStar (subst_in_term t k s) (subst_in_term t' k s))
.

Proof.
intros. induction H.
apply RedTerm_step. apply Substitution_in_reduction. trivial.
apply RedTerm_refl.
eapply RedTerm_trans. apply IHRedTermStar1. apply IHRedTermStar2.
Qed.

Lemma Reduction_in_substitution_star:
  forall (s s': LTerm) (t: LTerm) (k: nat),
    (RedTermStar s s') ->
    (RedTermStar (subst_in_term t k s) (subst_in_term t k s'))
.

Proof.
intros. induction H.
apply Reduction_in_substitution. trivial.
apply RedTerm_refl.
eapply RedTerm_trans. apply IHRedTermStar1. apply IHRedTermStar2.
Qed.

Lemma Reduction_and_substitution_star:
  forall (t t' s s': LTerm) (k: nat),
    (RedTermStar t t') ->
    (RedTermStar s s') ->
    (RedTermStar (subst_in_term t k s) (subst_in_term t' k s'))
.

Proof.
intros.
eapply RedTerm_trans.
apply Reduction_in_substitution_star. apply H0.
apply Substitution_in_reduction_star. trivial.
Qed.

(************************************************************************)
(************************************************************************)
(************************************************************************)
(************************************************************************)
(************************************************************************)

(** Weakening of declaration by the right **)
Lemma Weakening_of_declaration_by_the_right:
  forall (G1 G2 G3: LEnv) (i: nat) (T: LType),
    (List_from (G1 ++ G3) i T) ->
    (List_from
      (G1 ++ G2 ++ G3)
      (lift_at_depth_in_var i (length G3) (length G2))
      T)
.

Proof.
intros.
destruct (match_comp_nat2_aux_0 i (length G3)).
(* i < |G3| *)
unfold lift_at_depth_in_var. rewrite match_comp_nat2_lt.
apply List_from_lem3.
apply List_from_lem3.
eapply List_from_lem4. apply H. trivial. trivial.
(* i >= |G3| *)
unfold lift_at_depth_in_var. rewrite match_comp_nat2_ge.
replace (i + (length G2)) with
((length G3) + ((length G2) + (i - (length G3)))).
replace (G1 ++ G2 ++ G3) with ((G1 ++ G2) ++ G3).
apply List_from_lem2.
apply List_from_lem2.
eapply List_from_lem1.
replace i with ((length G3) + (i - (length G3))) in H. apply H.
omega. auto with datatypes.
omega. trivial.
Qed.

(** Weakening of well-typing by the right (auxiliary) **)
Lemma Weakening_of_well_typing_by_the_right_aux:
  forall (G1 G2 G: LEnv) (t: LTerm) (T: LType),
    (WellTyped G t T) ->
    forall (G3: LEnv),
      G = (G1 ++ G3) ->
      (WellTyped (G1 ++ G2 ++ G3)
         (lift_at_depth_in_term t (length G3) (length G2)) T)
.

Proof.
intros G1 G2 G t T H. induction H.

(* WT-const *)
intros.
simpl. eapply WT_const; trivial.

(* WT-var *)
intros.
simpl. eapply WT_var.
apply Weakening_of_declaration_by_the_right.
rewrite <- H0. trivial.

(* WT-fun *)
intros.
simpl. apply WT_fun.
rewrite snoc_of_append. rewrite snoc_of_append.
replace (S (length G3)) with (length (snoc G3 T)).
apply IHWellTyped.
rewrite H0. rewrite snoc_of_append. trivial.
apply length_of_snoc.

(* WT-app *)
intros. simpl. eapply WT_app.
apply IHWellTyped1. trivial.
apply IHWellTyped2. trivial.

(* WT-pair *)
intros. simpl. eapply WT_pair.
apply IHWellTyped1. trivial.
apply IHWellTyped2. trivial.

(* WT-fst *)
intros.
simpl. eapply WT_fst.
apply IHWellTyped. trivial.

(* WT-snd *)
intros.
simpl. eapply WT_snd.
apply IHWellTyped. trivial.
Qed.

(** Weakening_of_well-typing by the right **)
Lemma Weakening_of_well_typing_by_the_right:
  forall (G G': LEnv) (t: LTerm) (T: LType),
    (WellTyped G t T) ->
    (WellTyped (G ++ G') (lift_in_term t (length G')) T)
.

Proof.
intros.
unfold lift_in_term.
replace (0) with (length (A := LType) nil).
replace (G ++ G') with (G ++ G' ++ nil).
eapply Weakening_of_well_typing_by_the_right_aux.
apply H.
rewrite <- app_nil_end; trivial.
rewrite <- app_nil_end. trivial.
trivial.
Qed.

(** Substitution preserves typing **)
Lemma Substitution_preserves_typing:
  forall (G1 G2: LEnv) (T U: LType) (t u: LTerm),
    (WellTyped G1 u U) ->
    (WellTyped (G1 ++ (U :: G2)) t T) ->
    (WellTyped (G1 ++ G2) (subst_in_term t (length G2) u) T)
.

Proof.
cut (
  forall (G1: LEnv) (U: LType) (u: LTerm),
    (WellTyped G1 u U) ->
    forall (G: LEnv) (t: LTerm) (T: LType),
      (WellTyped G t T) ->
      forall (G2: LEnv),
        G = (G1 ++ (U :: G2)) ->
        (WellTyped (G1 ++ G2) (subst_in_term t (length G2) u) T)
).
intros. eapply H. apply H0. apply H1. trivial.

intros G1 U u H1 G t T H.
induction H.
(* WT-const *)
intros. simpl. apply WT_const; trivial.
(* WT-var *)
intros. simpl.
induction (lt_eq_lt_dec i (length G2)). induction a.
(* i < |G2| *)
unfold subst_in_var; rewrite match_comp_nat3_lt.
rewrite H0 in H.
apply WT_var.
eapply Declaration_of_variable_2. trivial. apply H. trivial.
(* i = |G2| *)
unfold subst_in_var; rewrite match_comp_nat3_eq.
rewrite H0 in H.
rewrite b in H.
assert (T = U).
eapply List_from_is_a_partial_function. apply H.
apply List_from_lem6.
rewrite H2.
apply Weakening_of_well_typing_by_the_right. trivial.
trivial.
(* i > |G2| *)
unfold subst_in_var; rewrite match_comp_nat3_gt.
rewrite H0 in H.
apply WT_var.
eapply Declaration_of_variable_1.
trivial. apply H. trivial.

(* WT-fun *)
intros. simpl. apply WT_fun.
rewrite snoc_of_append.
rewrite <- (length_of_snoc LType T).
apply IHWellTyped.
rewrite H0.
rewrite snoc_of_append.
rewrite snoc_of_cons. trivial.

(* WT-app *)
intros. simpl. eapply WT_app.
apply IHWellTyped1; trivial.
apply IHWellTyped2; trivial.

(* WT-pair *)
intros. simpl. eapply WT_pair.
apply IHWellTyped1; trivial.
apply IHWellTyped2; trivial.

(* WT-fst *)
intros. simpl. eapply WT_fst. apply IHWellTyped; trivial.

(* WT-snd *)
intros. simpl. eapply WT_snd. apply IHWellTyped; trivial.
Qed.

Lemma Substitution_preserves_typing_snoc:
  forall (G : LEnv) (T U : LType) (t u: LTerm),
    (WellTyped G u U) ->
    (WellTyped (snoc G U) t T) ->
    (WellTyped G (subst_in_term t 0 u) T)
.

Proof.
intros.
replace 0 with (length (nil (A := LType))).
replace G with (G ++ nil).
eapply Substitution_preserves_typing.
apply H. apply H0.  auto with datatypes. trivial.
Qed.

(******************************************************************************)

(** Weakening of well-typing by the left **)
Lemma Weakening_of_well_typing_by_the_left:
  forall (G1 G2: LEnv) (t: LTerm) (T: LType),
    (WellTyped G2 t T) ->
    (WellTyped (G1 ++ G2) t T)
.

Proof.
intros. induction H.
(* WT_const *)
apply WT_const; trivial.
(* WT_var *)
eapply WT_var. apply List_from_lem3. trivial.
(* WT_fun *)
apply WT_fun. rewrite snoc_of_append. trivial.
(* WT_apply *)
eapply WT_app. apply IHWellTyped1. trivial.
(* WT_pair *)
eapply WT_pair. apply IHWellTyped1. trivial.
(* WT-fst *)
eapply WT_fst. apply IHWellTyped.
(* WT-snd *)
eapply WT_snd. apply IHWellTyped.
Qed.

(******************************************************************************)
(******************************************************************************)

Theorem Subject_reduction:
  forall (G: LEnv) (t t': LTerm) (T: LType),
    (WellTyped G t T) ->
    (RedTerm t t') ->
    (WellTyped G t' T)
.

Proof.
intros. generalize G T H. clear G T H. induction H0.
(* R-fun *)
intros. inversion H. apply WT_fun. apply IHRedTerm; trivial.
(* R-app-left *)
intros. inversion H. eapply WT_app. apply IHRedTerm. apply H4. trivial.
(* R-app-right *)
intros. inversion H. eapply WT_app. apply H4. apply IHRedTerm. trivial.
(* RT-apply-redex *)
intros. inversion H. inversion H3.
eapply Substitution_preserves_typing_snoc. apply H5. trivial.
(* R-pair-left *)
intros. inversion H. eapply WT_pair. apply IHRedTerm. apply H4. trivial.
(* R-pair-right *)
intros. inversion H. eapply WT_pair. apply H4. apply IHRedTerm. trivial.
(* R-fst-cong *)
intros. inversion H. eapply WT_fst. apply IHRedTerm. apply H3.
(* R-fst-redex *)
intros. inversion H. inversion H2. trivial.
(* R-snd-cong *)
intros. inversion H. eapply WT_snd. apply IHRedTerm. apply H3.
(* R-snd-redex *)
intros. inversion H. inversion H2. trivial.
Qed.

Lemma Subject_reduction_star:
  forall (G: LEnv) (t t': LTerm) (T: LType),
    (WellTyped G t T) ->
    (RedTermStar t t') ->
    (WellTyped G t' T)
.

Proof.
intros. induction H0.
eapply Subject_reduction. apply H. trivial.
trivial. tauto.
Qed.

End LambdaCalculus.