(* Author: Vincent Cremet *)
(* Date: June 2007 *) 
(* File: SubjectReduction_proofs.v *)
(* Description: Subject reduction *)

(** * Subject reduction *)

Require Import List.
Require Import Arith.
(* Require Import Relations. *)

Require Import MatchCompNat.
Require Import Util.

Require Syntax.
Module SetProgram(My_Program: Syntax.Program).
Import My_Program.

(* Import definitions of SubjectReduction_preliminaries.v *)
Require SubjectReduction_preliminaries.
Module My_SubjectReduction_preliminaries := SubjectReduction_preliminaries.SetProgram(My_Program).
Import My_SubjectReduction_preliminaries.

Import My_TermSubstitutionInTyping.
Import My_TypeSubstitutionInTyping.
Import My_BoundCompliantSubstitutions_proofs.
Import My_Progress_proofs.
Import My_TransitivityElimination_proofs.
Import My_Normalization_proofs.
Import My_Subtyping_proofs.
Import My_Typing_proofs.
Import My_Confluence.
Import My_Auxiliary_proofs.
Import My_Substitutions_proofs.
Import My_Binders_proofs.
Import My_Semantics.
Import My_WellFormedness.
Import My_Typing.
Import My_Auxiliary.
Import My_Substitutions.
Import My_Binders.

(********************************************************************)
(********************************************************************)

Lemma Typed_term_list:
  forall (D: TypeEnv) (G: TermEnv) (ts: Term) (Ts: (list TypeS)),
    (Term_WT D G ts (TermListI Ts)) ->
    (Forall2 (fun t T => (Term_WT D G t (TermI T))) (list_of_termList ts) Ts)
.

Proof.
cut (
  forall (D: TypeEnv) (G: TermEnv) (ts: Term) (info: TermInfo),
    (Term_WT D G ts info) ->
    forall (Ts: (list TypeS)),
      info = (TermListI Ts) ->
      (Forall2 (fun t T => (Term_WT D G t (TermI T))) (list_of_termList ts) Ts)
).
intros. eapply H; eauto.

intros D G ts info H. induction H.
intros. discriminate H0.
intros. discriminate H2.
intros. discriminate H6.
intros. discriminate H3.
intros.  injection H. intros. rewrite <- H0. simpl. constructor.
intros. injection H1.  intros. rewrite <- H2. simpl. constructor. trivial. auto.
intros. discriminate H.
intros. discriminate H5.
Qed.

(********************************************************************)

Inductive FieldVal_WF (D: TypeEnv) (G: TermEnv) (N: TypeS)
                      (f: FieldSym) (t: Term): Prop :=

  | FieldVal_WF_intro:
    forall (S R: TypeS),
      let C := (fieldOwner f) in
      let T := (fieldType f) in
      (Type_WK D N Type_) ->
      (Subtyping D N (Class_type C R)) ->
      (Type_WK D (Class_type C R) Type_) ->
      (Term_WT D G t (TermI S)) ->
      (Subtyping D S (typeSubst_in_type T 0 R)) ->
      (FieldVal_WF D G N f t)
.


Lemma Well_formed_field_value:
  forall (D: TypeEnv) (G: TermEnv) (N: TypeS) (f: FieldSym) (t fts: Term),
    (Term_WT D G fts (FieldTermListI N)) ->
    (fieldTerm_lookup fts f t) ->
    (FieldVal_WF D G N f t)
.

Proof.
intros. generalize H. clear H. induction H0.
(* *)
intros. inversion H. eapply FieldVal_WF_intro; eauto.
(* *)
intros. inversion H. tauto.
Qed.

(********************************************************************)

Lemma Term_reduction_preserves_fields_of_fieldTermList:
  forall (D: TypeEnv) (G: TermEnv) (N: TypeS) (fts fts': Term),
    (Term_WT D G fts (FieldTermListI N)) ->
    (RedTerm fts fts') ->
    (fields_of_fieldTermList fts') =
    (fields_of_fieldTermList fts)
.

Proof.
intros. generalize fts' H H0. clear fts' H H0. induction fts.
intros. inversion H.
intros. inversion H.
intros. inversion H.
intros. inversion H.
intros. inversion H.
intros. inversion H.
intros. inversion H0.
intros. inversion H. inversion H0.
simpl. trivial. simpl. rewrite IHfts2. trivial. trivial. trivial.
Qed.

(********************************************************************)

Lemma Class_subtyping_implies_subtyping:
  forall (D: TypeEnv) (C C': ClassSym) (R R': TypeS),
    (Class_subtyping (Class_type C R) (Class_type C' R')) ->
    (Subtyping D (Class_type C R) (Class_type C' R'))
.

Proof.
cut (
  forall (D: TypeEnv) (N N': TypeS),
    (Class_subtyping N N') ->
    forall (C C': ClassSym) (R R': TypeS),
      N = (Class_type C R) ->
      N' = (Class_type C' R') ->
      (Subtyping D (Class_type C R) (Class_type C' R'))
).
intros. eapply H; eauto.

intros D N N' H. induction H.
(* refl *)
intros. rewrite H in H0. rewrite H0. apply S_Refl.
(* extends *)
intros. rewrite <- H1. clear C0 R0 H1.
destruct (classes_are_well_kinded C).
destruct (H3 _ H). clear H1 H3 H5. destruct H4. destruct H1.
rename x into B. rename x0 into RB.
eapply S_Class. apply H.
rewrite H1. rewrite typeSubst_of_class_type.
apply IHClass_subtyping.
rewrite H1. rewrite typeSubst_of_class_type. trivial. trivial.
Qed.

(*******************************************************************************)

Lemma Self_type_is_well_kinded:
  forall (C: ClassSym),
    let Pi := (classTypeParams C) in
    (Type_WK (Pi :: nil) (Class_type C (Type_var 0)) Type_)
.

Proof.
intros. unfold Class_type.
destruct (Class_kind_existence C). rename x into k.
eapply WK_apply. apply WK_class. apply H.
eapply WK_var.
apply decl_of_typeVar_intro.
replace (Pi :: nil) with (snoc nil Pi).
apply TypeSection_at_tail. trivial.
replace (lift_in_type Pi 1) with Pi. trivial.
symmetry. unfold lift_in_type.
destruct (classes_are_well_kinded C).
eapply Lifting_in_closed_type_cor1; eauto.
Qed.

(********************************************************************)

Lemma Type_env_reduction_preserves_typing:
  forall (D D': TypeEnv) (G: TermEnv) (t: Term) (info: TermInfo),
    (RedTypeEnvStar D D') ->
    (Term_WT D G t info) ->
    (Term_WT D' G t info)
.

Proof.
intros. induction H0.
(* T-Var *)
apply T_Var. trivial.
(* T-Select *)
unfold T. eapply T_Select. eauto.
eapply Reduction_preserves_subtyping. eauto. trivial.
apply RedType_refl. apply RedType_refl.
eapply Well_kindedness_is_preserved_by_type_env_reduction_star. eauto. trivial.
(* T-Call *)
unfold T. eapply T_Call. eauto.
eapply Well_kindedness_is_preserved_by_type_env_reduction_star; eauto.
eauto. 
eapply Reduction_preserves_subtyping. eauto. trivial.
apply RedType_refl. apply RedType_refl.
eapply Well_kindedness_is_preserved_by_type_env_reduction_star; eauto.
eapply Reduction_preserves_simultaneous_satisfaction with
  (D := D) (Pi := (typeSubst_in_type Pi 0 R')).
unfold Pi. eapply Substitution_in_method_type_params. eauto. eauto. trivial.
apply RedType_refl. apply RedType_refl.
unfold Pointwise_subtyping.
apply Forall2_implies with (P1 := fun T1 T2 => (Subtyping D T1 T2)).
intros. eapply Reduction_preserves_subtyping. apply H5. trivial.
apply RedType_refl. apply RedType_refl. trivial.
(* T-New *)
unfold N. eapply T_New.
eapply Well_kindedness_is_preserved_by_type_env_reduction_star; eauto.
eapply Reduction_preserves_simultaneous_satisfaction with
  (D := D) (Pi := (classTypeParams C)).
rewrite app_nil_end with (l := D).
apply Weakening_of_well_kindedness_by_the_left.
destruct (classes_are_well_kinded C). trivial.
eauto. trivial. apply RedType_refl. apply RedType_refl. trivial. trivial.
(* T-term-Nil *)
apply T_term_Nil.
(* T-term-Cons *)
apply T_term_Cons; trivial.
(* T-fieldTerm-Nil *)
apply T_fieldTerm_Nil.
(* T-fieldTerm-Cons *)
eapply T_fieldTerm_Cons.
eapply Well_kindedness_is_preserved_by_type_env_reduction_star; eauto.
eapply Reduction_preserves_subtyping. apply H1. trivial.
apply RedType_refl. apply RedType_refl.
eapply Well_kindedness_is_preserved_by_type_env_reduction_star; eauto.
eauto.
eapply Reduction_preserves_subtyping. apply H3. trivial.
apply RedType_refl. apply RedType_refl. trivial.
Qed.

(********************************************************************)

Lemma RTS_class_type:
  forall (C: ClassSym) (R R': TypeS),
    (RedTypeStar R R') ->
    (RedTypeStar (Class_type C R) (Class_type C R'))
.

Proof.
intros.
unfold Class_type. eapply RTS_apply_right. trivial.
Qed.

(********************************************************************)

Lemma RTE_head_star:
  forall (D: TypeEnv) (T T': TypeS),
    (RedTypeStar T T') ->
    (RedTypeEnvStar (T :: D) (T' :: D))
.

Proof.
intros. induction H.
apply RedTypeEnv_step. apply RTE_head; trivial.
apply RedTypeEnv_refl.
eapply RedTypeEnv_trans; eauto.
Qed.

Lemma RTE_tail_star:
  forall (D D': TypeEnv) (T: TypeS),
    (RedTypeEnvStar D D') ->
    (RedTypeEnvStar (T :: D) (T :: D'))
.

Proof.
intros. induction H.
apply RedTypeEnv_step. apply RTE_tail; trivial.
apply RedTypeEnv_refl.
eapply RedTypeEnv_trans; eauto.
Qed.

(********************************************************************)

Lemma Subject_reduction_case_T_call_redex_aux1:
  forall (m: MethodSym) (Ci: ClassSym) (R Ri Ri0 R' Rm S: TypeS),
    let C    := (methodOwner m) in
    let Pim  := (methodTypeParams m) in
    let T    := (methodType m) in
    let PiCi := (classTypeParams Ci) in

    (Type_WK nil (Class_type C R) Type_) ->
    (Type_WK (PiCi :: nil) (Class_type C Ri0) Type_) ->
    (Type_WK nil (Class_type C (typeSubst_in_type Ri0 0 Ri)) Type_) ->
    (Conform nil Ci Ri) ->

    (SectionSat nil Rm (typeSubst_in_type Pim 0 R)) -> 
    (SectionSat nil Rm (typeSubst_in_type Pim 0 R')) -> 

    (RedTypeStar R R') ->
    (RedTypeStar (typeSubst_in_type Ri0 0 Ri) R') ->

    let D    := (PiCi :: (typeSubst_in_type Pim 0 Ri0) :: nil) in
    (Type_WK D S Type_) ->
    (Subtyping D S (typeSubst_in_type T 1 Ri0)) ->

    (Subtyping nil
               (typeSubst_in_type (typeSubst_in_type S 1 Ri) 0 Rm)
               (typeSubst_in_type (typeSubst_in_type T 1 R) 0 Rm))
.

Proof.
intros until PiCi.
intros Hy1 Hy2 Hy3 Hy4 Hy5 H Hy6 Hy7.
intros D Hy8 Hy9.

(* *)
inversion H.
eapply Type_substitution_preserves_subtyping_cor2. eauto. apply H.
change (typeSubst_in_type Pim 0 R' :: nil) with
  (snoc nil (typeSubst_in_type Pim 0 R')). apply WK_typeEnv_snoc.
apply WK_typeEnv_nil.
unfold Pim. eapply Substitution_in_method_type_params.
eapply Well_kindedness_is_preserved_by_type_reduction_star.
apply Hy1.
apply RTS_class_type; trivial.
(* *)
apply Well_kindedness_is_preserved_by_type_env_reduction_star with
  (D := (typeSubst_in_type Pim 0 (typeSubst_in_type Ri0 0 Ri)) :: nil).
replace Pim with (typeSubst_in_type Pim 1 Ri).
Focus 2. unfold Pim. apply methodTypeParams_closed.
rewrite <- Permuting_type_substitutions_cor1.
inversion Hy4. inversion H4. inversion H9.
eapply Type_substitution_preserves_well_kindedness_cor1; eauto.
apply RTE_head_star. apply Reduction_in_substitution_star; trivial.
(* *)
apply Well_kindedness_is_preserved_by_type_env_reduction_star with
  (D := (typeSubst_in_type Pim 0 R :: nil)).
change (typeSubst_in_type Pim 0 R :: nil) with
  (nil ++ ((typeSubst_in_type Pim 0 R :: nil))).
unfold Pim. unfold T. apply Substitution_in_method_type_2. trivial.
apply RTE_head_star. apply Reduction_in_substitution_star; trivial.

(* *)
apply S_beta_right_star with (U' := (typeSubst_in_type T 1 R')).
Focus 2.
apply Reduction_in_substitution_star; trivial.

(* *)
apply Reduction_preserves_subtyping with 
  (D := (typeSubst_in_type Pim 0 (typeSubst_in_type Ri0 0 Ri)) :: nil)
  (T := (typeSubst_in_type S 1 Ri))
  (U := (typeSubst_in_type T 1 (typeSubst_in_type Ri0 0 Ri))).
Focus 2.
apply RTE_head_star.
apply Reduction_in_substitution_star. trivial.
Focus 2. apply RedType_refl.
Focus 2. apply Reduction_in_substitution_star. trivial.

(* *)
replace Pim with (typeSubst_in_type Pim 1 Ri).
Focus 2.
unfold Pim. apply methodTypeParams_closed.

(* *)
replace T with (typeSubst_in_type T 2 Ri).
Focus 2.
unfold T. apply methodType_closed.

(* *)
rewrite <- Permuting_type_substitutions_cor1.
rewrite <- Permuting_type_substitutions_cor2.

(* *)
inversion Hy4. inversion H4. clear D0 K R0 H8 H6 H7.
eapply Type_substitution_preserves_subtyping_cor1. eauto. eauto.
Focus 4. trivial.
(* *)
fold PiCi.
replace (PiCi :: (typeSubst_in_type Pim 0 Ri0) :: nil) with
        (snoc (snoc nil PiCi) (typeSubst_in_type Pim 0 Ri0)).
apply WK_typeEnv_snoc. apply WK_typeEnv_snoc. apply WK_typeEnv_nil.
destruct (classes_are_well_kinded Ci). trivial.
unfold Pim. eapply Substitution_in_method_type_params. eauto. trivial.
(* *)
trivial.
(* *)
fold PiCi.
change (PiCi :: typeSubst_in_type Pim 0 Ri0 :: nil) with
       ((PiCi :: nil) ++ typeSubst_in_type Pim 0 Ri0 :: nil).
unfold Pim. unfold T.
apply Substitution_in_method_type_2. trivial.
Qed.

(********************************************************************)

Lemma Subject_reduction_case_T_call_redex_aux2:
  forall (m: MethodSym) (Ci: ClassSym) (R Ri Ri0 R' Rm S: TypeS) (tb: Term),
    let C    := (methodOwner m) in
    let Pim  := (methodTypeParams m) in
    let Ts   := (methodParams m) in
    let PiCi := (classTypeParams Ci) in

    (Type_WK nil (Class_type C R) Type_) ->
    (Type_WK (PiCi :: nil) (Class_type C Ri0) Type_) ->
    (Type_WK nil (Class_type C (typeSubst_in_type Ri0 0 Ri)) Type_) ->
    (Conform nil Ci Ri) ->

    (SectionSat nil Rm (typeSubst_in_type Pim 0 R')) -> 

    (RedTypeStar R R') ->
    (RedTypeStar (typeSubst_in_type Ri0 0 Ri) R') ->

    let D    := (PiCi :: (typeSubst_in_type Pim 0 Ri0) :: nil) in
    let G    := ((Class_type Ci (Type_var 1)) :: (typeSubst_in_termEnv Ts 1 Ri0)) in
    (TermEnv_WK D G) ->
    (Term_WT D G tb (TermI S)) ->

    let G0 := ((Class_type Ci Ri) ::
               (typeSubst_in_termEnv
               (typeSubst_in_termEnv Ts 1 (typeSubst_in_type Ri0 0 Ri)) 0 Rm)) in
    (Term_WT nil G0
                 (typeSubst_in_term (typeSubst_in_term tb 1 Ri) 0 Rm)
                 (TermI (typeSubst_in_type (typeSubst_in_type S 1 Ri) 0 Rm)))
.

Proof.
intros until PiCi.
intros Hy1 Hy2 Hy3 Hy4 Hy5 Hy6 Hy7.
intros D G Hy42 Hy8 G0.

(* *)
set (D' := (typeSubst_in_type (typeSubst_in_type Pim 0 Ri0) 0 Ri) :: nil).
set (G' := (typeSubst_in_type (Class_type Ci (Type_var 1)) 1 Ri) ::
           (typeSubst_in_termEnv (typeSubst_in_termEnv Ts 1 Ri0) 1 Ri)).
assert (Term_WT D' G' (typeSubst_in_term tb 1 Ri) (TermI (typeSubst_in_type S 1 Ri))).
unfold D'. unfold G'.
inversion Hy4. inversion H. clear D0 K R0 H3 H1 H2.
eapply Type_substitution_preserves_typing_cor4; eauto.
 (* TypeEnv_WK *)
 fold PiCi.
 replace (PiCi :: (typeSubst_in_type Pim 0 Ri0) :: nil) with
         (snoc (snoc nil PiCi) (typeSubst_in_type Pim 0 Ri0)).
 apply WK_typeEnv_snoc. apply WK_typeEnv_snoc. apply WK_typeEnv_nil.
 destruct (classes_are_well_kinded Ci). trivial.
 unfold Pim. eapply Substitution_in_method_type_params. apply Hy2. trivial.
(* *)
clear D G Hy8 Hy42. rename D' into D. rename G' into G. rename H into Hy8. 
(* *)
assert ((typeSubst_in_type (Class_type Ci (Type_var 1)) 1 Ri) = (Class_type Ci Ri)).
simpl.
unfold typeSubst_in_typeVar. rewrite match_comp_nat3_eq.
replace (lift_in_type Ri 1) with Ri. trivial.
symmetry. unfold lift_in_type.
inversion Hy4. inversion H0.
eapply Lifting_in_closed_type_cor1; eauto. trivial.
rename H into Hy9.
(* *)
unfold G in Hy8. rewrite Hy9 in Hy8. clear Hy9. unfold D in Hy8. clear D G.
(* *)
rewrite Permuting_type_substitutions_cor1 in Hy8.
unfold Pim in Hy8. rewrite methodTypeParams_closed in Hy8. fold Pim in Hy8.

(* *)
unfold G0.
(* *)
replace (Class_type Ci Ri) with (typeSubst_in_type (Class_type Ci Ri) 0 Rm).
Focus 2.
inversion Hy4.
change 0 with ((length (nil (A := TypeS))) + 0).
eapply Substitution_in_closed_type_aux; eauto.
(* *)
change (typeSubst_in_type (Class_type Ci Ri) 0 Rm
        :: typeSubst_in_termEnv
           (typeSubst_in_termEnv Ts 1 (typeSubst_in_type Ri0 0 Ri)) 0 Rm) with
 (typeSubst_in_termEnv ((Class_type Ci Ri) ::
      (typeSubst_in_termEnv Ts 1 (typeSubst_in_type Ri0 0 Ri))) 0 Rm).
(* *)
inversion Hy5.
eapply Type_substitution_preserves_typing_cor5; eauto.
(* *)
change (typeSubst_in_type Pim 0 R' :: nil) with
       (snoc nil (typeSubst_in_type Pim 0 R')).
constructor. constructor.
unfold Pim. eapply Substitution_in_method_type_params.
eapply Well_kindedness_is_preserved_by_type_reduction_star.
apply Hy1.
apply RTS_class_type; trivial.

(* *)
constructor.
rewrite app_nil_end with (l := (typeSubst_in_type Pim 0 R' :: nil)).
apply Weakening_of_well_kindedness_by_the_left. inversion Hy4. trivial.
(* *)
unfold typeSubst_in_termEnv. apply List_forall_of_map.
apply List_forall_implies_forall with (P1 := fun U =>
  (Type_WK ((typeSubst_in_type Pim 0 (typeSubst_in_type Ri0 0 Ri)) :: nil)
        (typeSubst_in_type U 1 (typeSubst_in_type Ri0 0 Ri)) Type_)).
intros. rename x into U.
eapply Well_kindedness_is_preserved_by_type_env_reduction_star. eauto.
apply RTE_head_star. apply Reduction_in_substitution_star; trivial.
change (typeSubst_in_type Pim 0 (typeSubst_in_type Ri0 0 Ri) :: nil) with
       (nil ++ (typeSubst_in_type Pim 0 (typeSubst_in_type Ri0 0 Ri) :: nil)).
unfold Pim. unfold Ts. 
apply Substitution_in_method_param_types_2. trivial.

(* *)
replace Ts with (typeSubst_in_termEnv Ts 2 Ri).
Focus 2.
unfold Ts. apply methodParams_closed.

(* *)
rewrite <- Permuting_type_substitutions_in_lists_cor2.
change typeSubst_in_typeList with typeSubst_in_termEnv.

(* *)
apply Type_env_reduction_preserves_typing with
  (D := (typeSubst_in_type Pim 0 (typeSubst_in_type Ri0 0 Ri) :: nil)).
apply RTE_head_star. apply Reduction_in_substitution_star; trivial.

(* *)
trivial.
Qed.

(*************************************************************************)

Lemma Subject_reduction_case_T_call_redex_aux3:
  forall (m: MethodSym) (Ci: ClassSym) (R Ri Ri0 R' Rm S N: TypeS) (tb t ts: Term)
         (Ss: (list TypeS)),
    let C    := (methodOwner m) in
    let Pim  := (methodTypeParams m) in
    let Ts   := (methodParams m) in
    let PiCi := (classTypeParams Ci) in

    (Type_WK nil (Class_type C R) Type_) ->
    (Type_WK (PiCi :: nil) (Class_type C Ri0) Type_) ->
    (Type_WK nil (Class_type C (typeSubst_in_type Ri0 0 Ri)) Type_) ->
    (Conform nil Ci Ri) ->

    (SectionSat nil Rm (typeSubst_in_type Pim 0 R')) -> 

    (RedTypeStar R R') ->
    (RedTypeStar (typeSubst_in_type Ri0 0 Ri) R') ->

    (Term_WT nil nil t (TermI N)) ->
    (Subtyping nil N (Class_type Ci Ri)) ->

    let Ts1 := (typeSubst_in_termEnv (typeSubst_in_termEnv Ts 1 R) 0 Rm) in
    (Term_WT nil nil ts (TermListI Ss)) ->
    (Pointwise_subtyping nil Ss Ts1) ->

    let Ts' := (typeSubst_in_termEnv (typeSubst_in_termEnv
               Ts 1 (typeSubst_in_type Ri0 0 Ri)) 0 Rm) in
    let Us := ((Class_type Ci Ri) :: Ts') in
    let tb1 := (typeSubst_in_term (typeSubst_in_term tb 1 Ri) 0 Rm) in
    let S1 := (typeSubst_in_type (typeSubst_in_type S 1 Ri) 0 Rm) in
    (Term_WT nil Us tb1 (TermI S1)) ->

    exists S': TypeS,
      (Term_WT nil nil (polyTermSubst_in_term tb1 0 (t :: (list_of_termList ts)))
                       (TermI S')) /\
      (Subtyping nil S' S1)
.

Proof.
intros until PiCi.
intros Hy1 Hy2 Hy3 Hy4 Hy5 Hy6 Hy7 Hy8 Hy9.
intros Ts1 Hy10 Hy11 Ts' Us tb1 S1 Hy12.

(* *)
set (us := t :: (list_of_termList ts)).
(* *)
set (Us' := N :: Ss).
(* *)
assert (Hy13: Forall2 (fun (u : Term) (U' : TypeS) => Term_WT nil nil u (TermI U'))
         us Us').
unfold us. unfold Us'. constructor. trivial. apply Typed_term_list. trivial.
(* *)
assert (Hy14: Pointwise_subtyping nil Us' Us).
unfold Us. unfold Us'. constructor. trivial.
unfold Ts'.
unfold typeSubst_in_termEnv.
rewrite <- (map_id_is_id _ Ss). apply Forall2_of_map_cor1.
rewrite <- (map_id_is_id _ Ss). apply Forall2_of_map_cor1.
unfold Pointwise_subtyping in Hy11. unfold Ts1 in Hy11.
move Hy11 after Hy13.
apply Forall2_implies with (P1 := (fun T U : TypeS =>
      (Subtyping nil T (typeSubst_in_type (typeSubst_in_type U 1 R) 0 Rm)))).
(* *)
intros. rename xA into T. rename xB into U.
apply S_beta_right_star with
(U' := (typeSubst_in_type (typeSubst_in_type U 1 R') 0 Rm)).
eapply Reduction_preserves_subtyping. apply H. apply RedTypeEnv_refl.
apply RedType_refl.
apply Substitution_in_reduction_star. apply Reduction_in_substitution_star. trivial.
apply Substitution_in_reduction_star. apply Reduction_in_substitution_star. trivial.
(* *)
apply Forall2_of_map_cor2 with
  (P := fun T U => (Subtyping nil T (typeSubst_in_type U 0 Rm)))
  (f := fun T: TypeS => T)
  (g := fun U => (typeSubst_in_type U 1 R)) 
.
apply Forall2_of_map_cor2 with
  (P := fun T U => (Subtyping nil T U))
  (f := fun T: TypeS => T)
  (g := fun U => (typeSubst_in_type U 0 Rm)) 
.
rewrite map_id_is_id. rewrite map_id_is_id.
trivial.

(* *)
assert (Hy15: TermEnv_WK nil Us).
unfold Us. constructor. inversion Hy4; trivial.
unfold Ts'.
unfold typeSubst_in_termEnv. apply List_forall_of_map.
apply List_forall_implies_forall with 
  (P1 := fun T => (Type_WK ((typeSubst_in_type Pim 0 R') :: nil) T Type_)).
intros. rename x into T.
inversion Hy5.
eapply Substitution_preserves_well_kindedness_snoc. eauto. 
eapply Satisfaction_implies_same_kind; eauto. trivial.
(* *)
apply List_forall_implies_forall with 
  (P1 := fun T =>
  (Type_WK ((typeSubst_in_type Pim 0 (typeSubst_in_type Ri0 0 Ri)) :: nil) T Type_)).
intros. rename x into T.
eapply Well_kindedness_is_preserved_by_type_env_reduction_star. eauto.
apply RTE_head_star.
apply Reduction_in_substitution_star. trivial.
(* *)
apply List_forall_of_map.
change (typeSubst_in_type Pim 0 (typeSubst_in_type Ri0 0 Ri) :: nil) with
       (nil ++ (typeSubst_in_type Pim 0 (typeSubst_in_type Ri0 0 Ri) :: nil)).
unfold Pim. unfold Ts. apply Substitution_in_method_param_types_2. trivial.

(* *)
eapply Poly_term_substitution_preserves_typing_cor2; eauto.
Qed.

(*************************************************************************)

Lemma Subject_reduction_case_T_call_redex_aux4:
  forall (m: MethodSym) (Ci: ClassSym) (R Ri Ri0 R' Rm S N: TypeS) (tb t ts: Term)
         (Ss: (list TypeS)),
    let C    := (methodOwner m) in
    let Pim  := (methodTypeParams m) in
    let Ts   := (methodParams m) in
    let T    := (methodType m) in
    let PiCi := (classTypeParams Ci) in

    let D    := (PiCi :: (typeSubst_in_type Pim 0 Ri0) :: nil) in
    let G    := ((Class_type Ci (Type_var 1)) :: (typeSubst_in_termEnv Ts 1 Ri0)) in

    (Type_WK nil (Class_type C R) Type_) ->
    (Type_WK (PiCi :: nil) (Class_type C Ri0) Type_) ->
    (Type_WK nil (Class_type C (typeSubst_in_type Ri0 0 Ri)) Type_) ->
    (Conform nil Ci Ri) ->

    (SectionSat nil Rm (typeSubst_in_type Pim 0 R)) -> 
    (SectionSat nil Rm (typeSubst_in_type Pim 0 R')) -> 

    (RedTypeStar R R') ->
    (RedTypeStar (typeSubst_in_type Ri0 0 Ri) R') ->

    (Type_WK D S Type_) ->
    (Subtyping D S (typeSubst_in_type T 1 Ri0)) ->

    (Term_WT nil nil t (TermI N)) ->
    (Subtyping nil N (Class_type Ci Ri)) ->

    let Ts1 := (typeSubst_in_termEnv (typeSubst_in_termEnv Ts 1 R) 0 Rm) in
    (Term_WT nil nil ts (TermListI Ss)) ->
    (Pointwise_subtyping nil Ss Ts1) ->

    (Term_WT D G tb (TermI S)) ->
    (TermEnv_WK D G) ->

    let tb1 := (typeSubst_in_term (typeSubst_in_term tb 1 Ri) 0 Rm) in

    exists S': TypeS,
      (Term_WT nil nil (polyTermSubst_in_term tb1 0 (t :: (list_of_termList ts)))
                       (TermI S')) /\
      (Subtyping nil S' (typeSubst_in_type (typeSubst_in_type T 1 R) 0 Rm))
.

Proof.
intros until G.
intros Hy1 Hy2 Hy3 Hy4 Hy5 Hy6 Hy7 Hy8 Hy9 Hy10 Hy11 Hy12.
intros Ts1 Hy13 Hy14 Hy15 Hy42 tb1.

(* *)
assert (Hy16: (Type_WK nil (typeSubst_in_type (typeSubst_in_type S 1 Ri) 0 Rm) Type_)).
inversion Hy6.
(* *)
eapply Type_substitution_preserves_well_kindedness_cor2 with
  (Pi := (typeSubst_in_type Pim 0 R')). eauto.
rewrite <- H. constructor; trivial.
(* *)
eapply Well_kindedness_is_preserved_by_type_env_reduction_star with
  (D := (typeSubst_in_type Pim 0 (typeSubst_in_type Ri0 0 Ri)) :: nil).
Focus 2. apply RTE_head_star.
apply Reduction_in_substitution_star. trivial.
(* *)
replace Pim with (typeSubst_in_type Pim 1 Ri).
Focus 2. unfold Pim. apply methodTypeParams_closed.
rewrite <- Permuting_type_substitutions_cor1.
(* *)
inversion Hy4. inversion H3.
eapply Type_substitution_preserves_well_kindedness_cor1. eauto.
eapply Satisfaction_implies_same_kind. apply H9. apply H4.
(* *)
trivial.

(* *)
cut (
    exists S': TypeS,
      (Term_WT nil nil (polyTermSubst_in_term tb1 0 (t :: (list_of_termList ts)))
                       (TermI S')) /\
      (Subtyping nil S' (typeSubst_in_type (typeSubst_in_type S 1 Ri) 0 Rm))
).
intros. destruct H. destruct H. rename x into S'. rename H into Hy17.
rename H0 into Hy18.
exists S'. split. trivial.
apply S_Trans with (S := (typeSubst_in_type (typeSubst_in_type S 1 Ri) 0 Rm)).
trivial. trivial.
(* *)
unfold T.
eapply Subject_reduction_case_T_call_redex_aux1; eauto.

(* *)
unfold tb1.
eapply Subject_reduction_case_T_call_redex_aux3.
unfold C in Hy1. apply Hy1. unfold PiCi in Hy2. apply Hy2.
trivial. trivial. apply Hy6. trivial. trivial. apply Hy11. trivial.
eauto. trivial.

(* *)
eapply Subject_reduction_case_T_call_redex_aux2; eauto.
Qed.

(********************************************************************)

Theorem Subject_reduction:
  forall (t u: Term) (info: TermInfo),
    (Term_WT nil nil t info) ->
    (RedTerm t u) ->
    exists info': TermInfo,
      (Term_WT nil nil u info') /\
      (SubInfo nil info' info)
.

Proof.
intros. generalize u H0. clear u H0. induction H.

(* T-var *)
intros. inversion H0.

(* T-select *)
intros. inversion H2.

(* - R-select-redex *)
clear IHTerm_WT. rewrite <- H4 in H6. rewrite <- H4. clear u H2 f0 H5 H4.
(* *)
rewrite <- H3 in H. clear t H3. inversion H. clear fts H3.
rewrite <- H4 in H0. rewrite <- H4 in H. clear S H4.
rewrite <- H2 in H. clear N H2.
(* *)
destruct (Well_formed_field_value _ _ _ _ _ _ H8 H6).
(* *)
unfold N0 in H0. unfold N0 in H3. unfold N0 in H2.
assert (BetaType R R1).
eapply Uniqueness_of_subtyping. apply H2. apply H1. trivial. trivial. trivial.
(* *)
exists (TermI S). split. trivial.
(* *)
apply SubInfo_term.
apply S_Trans with (S := (typeSubst_in_type T0 0 R1)).
(* *)
unfold T0. eapply Substitution_in_field_type. apply H4.
(* *)
trivial.
(* *)
apply BetaType_implies_subtyping_lem1. apply BetaType_sym; trivial.

(* - R-select *)
destruct (IHTerm_WT _ H6). destruct H7. rename x into info'.
inversion H8.
(* *)
exists (TermI (typeSubst_in_type T 0 R)). split.
(* *)
assert (Subtyping nil T0 (Class_type C R)).
eapply S_Trans.
eapply Typing_returns_well_kinded_types_cor1. apply H. apply List_forall_nil.
trivial. trivial.
(* *)
rewrite <- H10 in H7.
unfold T. eapply T_Select; eauto.
(* *)
apply SubInfo_term. apply S_Refl.

(* T-call *)
intros. inversion H6.

(* - R-call-redex *)
clear IHTerm_WT1. clear IHTerm_WT2.
(* *)
clear u H6 H9. clear m0 H8. subst t. clear vs H11.
rename C0 into Ci. clear u1. rename R'0 into Ri.
rename t0 into tb. rename u0 into u. rename R into Rm. rename R' into R.
rename Pi into Pim.
(* *)
unfold v in H. inversion H. clear fts H7. rewrite <- H6 in H12.
rename C0 into Cv. rename N0 into Nv. rename R1 into Rv.
clear H15. subst S.
(* *)
assert (Conform nil Ci Ri).
assert (Conform nil Cv Rv). apply Conform_intro; eauto.
eapply Class_subtyping_subject_reduction. apply WK_typeEnv_nil. apply H7. trivial.
destruct H7.
(* *)
destruct (method_implementations_are_well_formed m Ci tb H13).
rename Rm0 into Ri0.
change Cm with C in H15. change Cm with C in H16. clear Cm.
(* *)
assert (Subtyping nil (Class_type Cv Rv) (Class_type Ci Ri)).
apply Class_subtyping_implies_subtyping; trivial.
(* *)
assert (Subtyping nil (Class_type Ci Ri) (Class_type C (typeSubst_in_type Ri0 0 Ri))).
rewrite <- typeSubst_of_class_type.
replace (Class_type Ci Ri) with (typeSubst_in_type (Class_type Ci (Type_var 0)) 0 Ri).
inversion H7.
eapply Type_substitution_preserves_subtyping_snoc; eauto.
apply WK_typeEnv_snoc. apply WK_typeEnv_nil.
destruct (classes_are_well_kinded Ci). trivial.
unfold Class_type.
inversion H23.
eapply WK_apply. apply WK_class. apply H28.
eapply WK_var. apply decl_of_typeVar_intro.
apply TypeSection_at_tail. unfold lift_in_type.
apply Lifting_preserves_kind_of. trivial.
simpl. unfold typeSubst_in_typeVar. rewrite match_comp_nat3_eq.
unfold lift_in_type. rewrite lift_in_type_by_zero. trivial. trivial.
(* *)
assert (Subtyping nil (Class_type Cv Rv) (Class_type C (typeSubst_in_type Ri0 0 Ri))).
apply S_Trans with (S := (Class_type Ci Ri)); trivial.
(* *)
assert (Type_WK nil (Class_type C (typeSubst_in_type Ri0 0 Ri)) Type_).
rewrite <- typeSubst_of_class_type.
inversion H8.
eapply Substitution_preserves_well_kindedness_snoc. eauto.
eapply Satisfaction_implies_same_kind; eauto.
replace (snoc nil (classTypeParams Ci)) with (PiC :: nil). trivial.
trivial.
(* *)
assert (BetaType R (typeSubst_in_type Ri0 0 Ri)).
eapply Uniqueness_of_subtyping. unfold Nv in H9. apply H9. apply H3.
trivial. trivial. trivial.
(* *)
destruct (ChurchRosser _ _ H23). destruct H24. rename x into R'.
(* *)
assert (SectionSat nil Rm (typeSubst_in_type Pim 0 R')).
eapply Reduction_preserves_simultaneous_satisfaction with
  (Pi := (typeSubst_in_type Pim 0 R)).
unfold Pim. eapply Substitution_in_method_type_params.
apply H3. apply H4. apply RedTypeEnv_refl. apply RedType_refl.
apply Reduction_in_substitution_star. trivial.
(* *)
assert (Conform nil Ci Ri). constructor; trivial. clear H7 H8.
clear H20 H21 H23.
(* *)
assert (Hy0: (TermEnv_WK D G)).
unfold TermEnv_WK. unfold G.
apply List_forall_cons. fold PiC.
replace (Class_type Ci (Type_var 1)) with
        (lift_in_type (Class_type Ci (Type_var 0))
                      (length ((typeSubst_in_type Pim 0 Ri0) :: nil))).
unfold D. change Pim0 with Pim.
change (PiC :: (typeSubst_in_type Pim 0 Ri0) :: nil) with
       ((PiC :: nil) ++ ((typeSubst_in_type Pim 0 Ri0) :: nil)).
apply Weakening_of_well_kindedness_by_the_right.
unfold PiC. apply Self_type_is_well_kinded.
simpl. unfold lift_in_type. simpl.
unfold lift_at_depth_in_typeVar. rewrite match_comp_nat2_ge. trivial. omega.
unfold typeSubst_in_termEnv. apply List_forall_of_map.
destruct (method_declarations_are_well_formed m).
fold PiC. unfold D. change Pim0 with Pim.
change (PiC :: typeSubst_in_type Pim 0 Ri0 :: nil) with
       ((PiC :: nil) ++ (typeSubst_in_type Pim 0 Ri0 :: nil)).
unfold Pim. unfold Ts0. apply Substitution_in_method_param_types_2. trivial.
(* *)
assert (Hy: (Type_WK D S Type_)).
eapply Typing_returns_well_kinded_types_cor1 with (G := G). apply H17.
apply Hy0.

(* *)
cut (
 exists S' : TypeS,
     (Term_WT nil nil u' (TermI S')) /\
     (Subtyping nil S' (typeSubst_in_type (typeSubst_in_type T 1 R) 0 Rm))
).
intros. destruct H7. rename x into S'. destruct H7. exists (TermI S'). split.
tauto. constructor; trivial.
unfold T. unfold u'. unfold u. rewrite H10.
eapply  Subject_reduction_case_T_call_redex_aux4. eauto.
unfold PiC in H16. apply H16. apply H22. trivial. trivial. apply H26.
trivial. trivial. apply Hy.
apply H18. eauto. apply H19. eauto. trivial. trivial. trivial.

(* - R-call-left *)
clear IHTerm_WT2.
(* *)
exists (TermI (typeSubst_in_type (typeSubst_in_type T 1 R') 0 R)).
(* *)
destruct (IHTerm_WT1 _ H12). destruct H13. rename x into info'.
(* *)
inversion H14. rewrite <- H16 in H13. rename T0 into S'.
(* *)
assert (Subtyping nil S' (Class_type C R')).
apply S_Trans with (S := S).
eapply Typing_returns_well_kinded_types_cor1. apply H. apply List_forall_nil.
trivial. trivial.
(* *)
split. unfold T. eapply T_Call; eauto.
(* *)
apply SubInfo_term. apply S_Refl.
(* - R-call-right *)
clear IHTerm_WT1. clear u H6 t0 ts0 m0 R0 H7 H9 H10 H11 H8.
(* *)
exists (TermI (typeSubst_in_type (typeSubst_in_type T 1 R') 0 R)).
(* *)
destruct (IHTerm_WT2 _ H12). destruct H6. rename x into info'.
(* *)
inversion H7. rename Ts0 into Ss'. rewrite <- H9 in H6. clear info' H7 Us H9 H8.
(* *)
assert (Pointwise_subtyping nil Ss'
        (typeSubst_in_typeList (typeSubst_in_typeList Ts 1 R') 0 R)).
apply SubInfo_trans_cor1 with (Ts2 := Ss).
eapply Typing_returns_well_kinded_types. apply H1. apply List_forall_nil.
trivial. trivial.
(* *)
split. unfold T. eapply T_Call; eauto.
(* *)
apply SubInfo_term. apply S_Refl.

(* T-new *)
intros. inversion H3.
(* *)
destruct (IHTerm_WT _ H7). destruct H8. rename x into info'.
(* *)
inversion H9. rewrite <- H10 in H8. rewrite H12 in H8.
(* *)
exists (TermI N). split.
(* *)
unfold N. apply T_New; trivial.
rewrite (Term_reduction_preserves_fields_of_fieldTermList _ _ _ _ _ H1 H7). trivial.
(* *)
apply SubInfo_term. apply S_Refl.

(* T-term-nil *)
intros. inversion H0.

(* T-term-cons *)
intros. inversion H1.
(* - R-term-cons-left *)
destruct (IHTerm_WT1 _ H5). rename x into info'. destruct H6.
(* *)
inversion H7. rename T0 into T'. rewrite <- H9 in H6.
(* *)
exists (TermListI (T' :: Ts)). split.
(* *)
apply T_term_Cons. trivial. trivial.
(* *)
apply SubInfo_termList.
unfold Pointwise_subtyping. apply Forall2_cons. trivial.
apply Forall2_refl. intros. apply S_Refl.
(* R-term-cons-right *)
destruct (IHTerm_WT2 _ H5). rename x into info'. destruct H6.
(* *)
inversion H7. rename Ts0 into Ts'. rewrite <- H9 in H6.
(* *)
exists (TermListI (T :: Ts')). split.
(* *)
apply T_term_Cons. trivial. trivial.
(* *)
apply SubInfo_termList.
unfold Pointwise_subtyping. apply Forall2_cons.
(* *)
apply S_Refl.
(* *)
apply H10.

(* T-fieldTerm-nil *)
intros. inversion H0.

(* T-fieldTerm-cons *)
intros. exists (FieldTermListI N). inversion H5.
(* - R-fieldTerm-cons-left *)
clear IHTerm_WT2.
(* *)
destruct (IHTerm_WT1 _ H10). destruct H11. rename x into info'.
(* *)
inversion H12. rename T0 into S'. rewrite <- H14 in H11.
(* *)
split.
(* *)
eapply T_fieldTerm_Cons; eauto.
(* *)
apply S_Trans with (S := S).
eapply Typing_returns_well_kinded_types_cor1. apply H2. apply List_forall_nil.
trivial. trivial.
(* *)
apply SubInfo_fieldTermList.
(* - R-fieldTerm-cons-right *)
clear IHTerm_WT1.
(* *)
destruct (IHTerm_WT2 _ H10). destruct H11. rename x into info'.
(* *)
inversion H12. rewrite <- H13 in H11. rewrite H15 in H11.
(* *)
split.
(* *)
eapply T_fieldTerm_Cons; eauto.
(* *)
apply SubInfo_fieldTermList.

Qed.

End SetProgram.
