(* Author: Vincent Cremet *)
(* Date: June 2007 *) 
(* File: Typing_proofs.v *)
(* Description: Typing of FGJ-omega (Proofs) *)

(** * Definition of Typing in FGJ-omega (Proofs) *)

Require Import List.
Require Import Arith.

Require Import MatchCompNat.
Require Import Util.

Require Syntax.
Module SetProgram(My_Program: Syntax.Program).
Import My_Program.

(* Import definitions of Confluence.v *)
Require Confluence.
Module My_Confluence := Confluence.SetProgram(My_Program).
Import My_Confluence.

Import My_Auxiliary_proofs.
Import My_Substitutions_proofs.
Import My_Binders_proofs.
Import My_Semantics.
Import My_WellFormedness.
Import My_Typing.
Import My_Auxiliary.
Import My_Substitutions.
Import My_Binders.

(********************************************************************)

(***************************************************************************)
(** Well-kindedness is preserved by subtitution of types that match kinds **)
(***************************************************************************)

Lemma Substitution_preserves_class_types:
  forall (X: TypeVar) (R N: TypeS),
    (is_classType N) ->
    (is_classType (typeSubst_in_type N X R))
.

Proof.
intros. destruct H. destruct H. rewrite H.
simpl. exists x. exists (typeSubst_in_type x0 X R). trivial.
Qed.

Lemma Lifting_preserves_class_types:
  forall (n k: nat) (N: TypeS),
    (is_classType N) ->
    (is_classType (lift_at_depth_in_type N n k))
.

Proof.
intros. destruct H. destruct H. rewrite H.
simpl. exists x. exists (lift_at_depth_in_type x0 n k). trivial.
Qed.

Lemma Substitution_preserves_kind_of:
  forall (t: TypeS) (R: TypeS) (k: KindS),
    (kind_of t k) ->
    forall (X: TypeVar),
      (kind_of (typeSubst_in_type t X R) k)
.

Proof.
intros t R k H.
induction H.
intro. simpl. apply Kind_of_section. apply IHkind_of.
intro. simpl. apply Kind_of_param. apply IHkind_of.
intro. simpl. apply Kind_of_paramList_nil.
intro. simpl. apply Kind_of_paramList_cons. apply IHkind_of1. apply IHkind_of2.
Qed.

Lemma Dropping_preserves_kind_of:
  forall (t: TypeS) (k: KindS) (d: nat),
    (kind_of t k) ->
    (kind_of (drop_at_depth_in_type t d) k)
.

Proof.
intros. generalize d. clear d. induction H.
intro. simpl. apply Kind_of_section. apply IHkind_of.
intro. simpl. apply Kind_of_param. apply IHkind_of.
intro. simpl. apply Kind_of_paramList_nil.
intro. simpl. apply Kind_of_paramList_cons. apply IHkind_of1. apply IHkind_of2.
Qed.

Lemma typeSubst_of_snoc:
  forall (Pi: TypeS) (R: TypeS) (D: TypeEnv),
    forall (X: TypeVar),
      (typeSubst_in_typeEnv (snoc D Pi) X R) =
      (snoc (typeSubst_in_typeEnv D X R) (typeSubst_in_type Pi (X + length D) R))
.

Proof.
intros Pi R D.
induction D.
intro.
simpl.
rewrite plus_0_r; trivial.
intro.
rewrite snoc_of_cons.
simpl.
rewrite snoc_of_cons.
rewrite (IHD (S X)).
rewrite plus_Snm_nSm.
trivial.
Qed.

Lemma lift_of_snoc_aux:
  forall (Pi: TypeS) (n: nat) (D: TypeEnv) (k: nat),
      (lift_at_depth_in_typeEnv (snoc D Pi) n k) =
      (snoc (lift_at_depth_in_typeEnv D n k) (lift_at_depth_in_type Pi n (k + length D)))
.

Proof.
intros Pi n D.
induction D.
intro. simpl.
rewrite plus_0_r; trivial.
intro.
rewrite snoc_of_cons. simpl.
rewrite snoc_of_cons.
rewrite (IHD (S k)).
rewrite plus_Snm_nSm.
trivial.
Qed.

Lemma lift_of_snoc:
  forall (D: TypeEnv) (Pi: TypeS) (n: nat),
      (lift_in_typeEnv (snoc D Pi) n) =
      (snoc (lift_in_typeEnv D n) (lift_at_depth_in_type Pi n (length D)))
.

Proof.
intros.
unfold lift_in_typeEnv.
replace (length D) with (0 + length D).
apply lift_of_snoc_aux. trivial.
Qed.

Lemma Lifting_preserves_kind_of:
  forall (t: TypeS) (k: KindS) (n: nat),
    (kind_of t k) ->
    forall (d: nat),
      (kind_of (lift_at_depth_in_type t n d) k)
.

Proof.
intros t k n H.
induction H.
intro. simpl. apply Kind_of_section. apply IHkind_of.
intro. simpl. apply Kind_of_param. apply IHkind_of.
intro. simpl. apply Kind_of_paramList_nil.
intro. simpl. apply Kind_of_paramList_cons. apply IHkind_of1. apply IHkind_of2.
Qed.

Lemma TypeSection_at_is_a_partial_function:
  forall (ts: TypeEnv) (k: nat) (t: TypeS),
    (TypeSection_at ts k t) ->
    forall (t': TypeS),
      (TypeSection_at ts k t') ->
      (t = t')
.

Proof.
intros ts k t H. induction H.
intros. inversion H. replace Pi with Pi0. trivial.
destruct (snoc_inject _ _ _ _ _ H1). trivial.
intros. inversion H0. apply IHTypeSection_at. replace D with D0. trivial.
destruct (snoc_inject _ _ _ _ _ H1). trivial.
Qed.

Lemma decl_of_typeVar_is_a_partial_function:
  forall (D: TypeEnv) (X: TypeVar) (Pi Pi': TypeS),
    (decl_of_typeVar D X Pi) ->
    (decl_of_typeVar D X Pi') ->
    (Pi = Pi')
.

Proof.
intros.
inversion H. inversion H0.
replace Pi0 with Pi1. trivial.
eapply TypeSection_at_is_a_partial_function. apply H5. trivial.
Qed.

Lemma kind_of_is_a_partial_function:
  forall (t: TypeS) (k: KindS),
    (kind_of t k) ->
    forall (k': KindS),
      (kind_of t k') ->
      (k = k')
.

Proof.
intros t k H.
induction H.
intros. inversion H0. replace (k0) with (k). trivial. apply IHkind_of. trivial.
intros. inversion H0. replace (k0) with (k). trivial. apply IHkind_of. trivial.
intros. inversion H. trivial.
intros. inversion H1. replace (k0) with (k). replace (ks0) with (ks). trivial.
apply IHkind_of2; trivial. apply IHkind_of1; trivial.
Qed.

Lemma length_of_typeSubst:
  forall (R: TypeS) (D: TypeEnv) (X: TypeVar),
    (length (typeSubst_in_typeEnv D X R)) = (length D)
.

Proof.
intros R D.
induction D.
intro. simpl. trivial.
intro. simpl. rewrite IHD. trivial.
Qed.

Lemma length_of_typeLift:
  forall (n: nat) (D: TypeEnv) (k: nat),
    (length (lift_at_depth_in_typeEnv D n k)) = (length D)
.

Proof.
intros n D. induction D.
intro. simpl. trivial.
intro. simpl. rewrite IHD. trivial.
Qed.

Lemma TypeSection_at_lem1:
  forall (D1 D2: TypeEnv) (Pi: TypeS) (k: nat),
    (TypeSection_at (D1 ++ D2) ((length D2) + k) Pi) ->
    (TypeSection_at D1 k Pi)
.

Proof.
intros D1 D2 Pi k. 
pattern D2. apply (snoc_ind TypeS).
simpl. rewrite <- (app_nil_end D1). tauto.
intros. apply H.
rewrite <- snoc_of_append in H0.
rewrite length_of_snoc in H0. simpl in H0. inversion H0.
replace (D1 ++ xs) with D. trivial.
destruct (snoc_inject _ _ _ _ _ H1). trivial.
Qed.

Lemma TypeSection_at_lem2:
  forall (D1 D2: TypeEnv) (Pi: TypeS) (k: nat),
    (TypeSection_at D1 k Pi) ->
    (TypeSection_at (D1 ++ D2) ((length D2) + k) Pi)
.

Proof.
intros. 
pattern D2. apply (snoc_ind TypeS).
rewrite <- (app_nil_end D1). simpl. trivial.
intros. rewrite <- snoc_of_append. rewrite length_of_snoc. simpl.
apply TypeSection_at_head. trivial.
Qed.

Lemma TypeSection_at_lem3:
  forall (D1 D2: TypeEnv) (Pi: TypeS) (k: nat),
    (TypeSection_at D2 k Pi) ->
    (TypeSection_at (D1 ++ D2) k Pi)
.

Proof.
intros D1 D2 Pi k H.
induction H.
rewrite <- snoc_of_append. apply TypeSection_at_tail.
rewrite <- snoc_of_append. apply TypeSection_at_head. trivial.
Qed.

Lemma TypeSection_at_lem4:
  forall (D1 D2: TypeEnv) (Pi: TypeS) (k: nat),
    (TypeSection_at (D1 ++ D2) k Pi) ->
    (k < (length D2)) ->
    (TypeSection_at D2 k Pi)
.

Proof.
cut (
  forall (D1 D: TypeEnv) (k: nat) (Pi: TypeS),
    (TypeSection_at D k Pi) ->
    forall (D2: TypeEnv),
      D = D1 ++ D2 ->
      (k < (length D2)) ->
      (TypeSection_at D2 k Pi)
).
intros.
eapply H. apply H0. reflexivity. trivial.

intros D1 D k Pi H. induction H.

intros.
assert ((length D2) > 0). trivial. induction (snoc_ind_aux_5 TypeS D2 H1).
induction H2. rewrite H2 in H. rewrite <- snoc_of_append in H.
rewrite H2. replace x0 with Pi. apply TypeSection_at_tail.
destruct (snoc_inject _ _ _ _ _ H). trivial.

intros.
assert ((length D2) > 0). omega. induction (snoc_ind_aux_5 TypeS D2 H2).
destruct H3. rewrite H3. apply TypeSection_at_head.
apply IHTypeSection_at.
rewrite H3 in H0. rewrite <- snoc_of_append in H0.
destruct (snoc_inject _ _ _ _ _ H0). trivial.
rewrite H3 in H1.
rewrite length_of_snoc in H1. omega.
Qed.

Lemma TypeSection_at_lem5:
  forall (Pi R: TypeS) (D: TypeEnv) (X: TypeVar),
    (TypeSection_at D X Pi) ->
    (TypeSection_at (typeSubst_in_typeEnv D 0 R) X
                    (typeSubst_in_type Pi ((length D) - (S X)) R))
.

Proof.
intros Pi R D X H.
induction H.
rewrite typeSubst_of_snoc. simpl. rewrite length_of_snoc.
replace (S (length D) - 1) with (length D).
apply TypeSection_at_tail. omega.
rewrite typeSubst_of_snoc. simpl. rewrite length_of_snoc.
replace (S (length D) - (S (S n))) with (length D - S n).
apply TypeSection_at_head. trivial. omega.
Qed.

Lemma by_omega1: forall (x: nat), (x > 0) -> (S (pred x)) = x.
Proof. intros. omega. Qed.

Lemma Declaration_of_type_variable_1:
  forall (D1 D2: TypeEnv) (Pi Pi' R: TypeS) (X: TypeVar),
    X > (length D2) ->
    (decl_of_typeVar (D1 ++ (Pi :: D2)) X Pi') ->
    (decl_of_typeVar (D1 ++ (typeSubst_in_typeEnv D2 0 R)) (pred X)
      (drop_in_type Pi'))
.

Proof.
intros.
inversion H0. clear D X0 H2 H3.
replace  (drop_in_type (lift_in_type Pi0 (S X))) with (lift_in_type Pi0 X).
pattern X at 2. replace X with (S (pred X)).
apply decl_of_typeVar_intro.
replace (pred X) with ((length (typeSubst_in_typeEnv D2 0 R)) + (
  (pred X) - (length (typeSubst_in_typeEnv D2 0 R))
)).
apply TypeSection_at_lem2.
apply TypeSection_at_lem1 with (D2 := Pi :: D2).
rewrite length_of_typeSubst. simpl.
replace (S (length D2 + (pred X - length D2))) with X. trivial.
replace (length D2 + (pred X - length D2)) with (pred X).
rewrite by_omega1. trivial. omega.
omega.
rewrite length_of_typeSubst. omega.
omega.
unfold lift_in_type. unfold drop_in_type.
rewrite Lifting_then_dropping. trivial.
Qed.

Lemma Declaration_of_type_variable_1_bis:
  forall (D1 D2: TypeEnv) (Pi Pi' R: TypeS) (X: TypeVar),
    X > (length D2) ->
    (decl_of_typeVar (D1 ++ (Pi :: D2)) X Pi') ->
    (decl_of_typeVar (D1 ++ (typeSubst_in_typeEnv D2 0 R)) (pred X)
      (typeSubst_in_type Pi' (length D2) R))
.

Proof.
intros.
inversion H0. clear D X0 H2 H3.
replace (typeSubst_in_type (lift_in_type Pi0 (S X)) (length D2) R) with
        (lift_in_type Pi0 X).
pattern X at 2. replace X with (S (pred X)).
apply decl_of_typeVar_intro.
replace (pred X) with ((length (typeSubst_in_typeEnv D2 0 R)) + (
  (pred X) - (length (typeSubst_in_typeEnv D2 0 R))
)).
apply TypeSection_at_lem2.
apply TypeSection_at_lem1 with (D2 := Pi :: D2).
rewrite length_of_typeSubst. simpl.
replace (S (length D2 + (pred X - length D2))) with X. trivial.
replace (length D2 + (pred X - length D2)) with (pred X).
rewrite by_omega1. trivial. omega.
omega.
rewrite length_of_typeSubst. omega.
omega.
replace X with ((length D2) + (X - (length D2))).
rewrite Type_lifting_stronger_than_substitution. trivial.
omega.
Qed.

Lemma Declaration_of_type_variable_2:
  forall (D1 D2: TypeEnv) (Pi Pi' R: TypeS) (X: TypeVar),
    X < (length D2) ->
    (decl_of_typeVar (D1 ++ (Pi :: D2)) X Pi') ->
    (decl_of_typeVar (D1 ++ (typeSubst_in_typeEnv D2 0 R)) X (typeSubst_in_type Pi' (length D2) R))
.

Proof.
intros.
inversion H0. clear D X0 H2 H3.
replace (typeSubst_in_type (lift_in_type Pi0 (S X)) (length D2) R) with
  (lift_in_type (typeSubst_in_type Pi0 ((length D2) - (S X)) R) (S X)).
apply decl_of_typeVar_intro.
apply TypeSection_at_lem3.
apply TypeSection_at_lem5.
apply TypeSection_at_lem4 with (D1 := (snoc D1 Pi)). trivial.
replace (snoc D1 Pi ++ D2) with (D1 ++ Pi :: D2). trivial.
rewrite snoc_then_append. trivial. trivial.
rewrite Permuting_type_lifting_and_substitution.
replace  (length D2 - S X + S X) with (length D2). trivial.
omega.
Qed.

Lemma TypeSection_at_lem6:
  forall (D1 D2: TypeEnv) (Pi: TypeS),
    (TypeSection_at (D1 ++ (Pi :: D2)) (length D2) Pi)
.

Proof.
intros.
pattern D2. apply (snoc_ind TypeS).
simpl.
replace (D1 ++ Pi :: nil) with (snoc D1 Pi). apply TypeSection_at_tail. trivial.
intros. rewrite length_of_snoc. rewrite <- snoc_of_cons. rewrite <- snoc_of_append. apply TypeSection_at_head. trivial.
Qed.

Lemma Declaration_of_type_variable_3:
  forall (D1 D2: TypeEnv) (Pi: TypeS),
    (decl_of_typeVar (D1 ++ (Pi :: D2)) (length D2) (lift_in_type Pi (S (length D2))))
.

Proof.
intros.
apply decl_of_typeVar_intro.
apply TypeSection_at_lem6.
Qed.

Lemma TypeSection_at_lem7:
  forall (Pi: TypeS) (D: TypeEnv) (X: TypeVar) (n: nat),
    (TypeSection_at D X Pi) ->
    (TypeSection_at (lift_at_depth_in_typeEnv D n 0) X
                    (lift_at_depth_in_type Pi n ((length D) - (S X))))
.

Proof.
intros. induction H.
rewrite lift_of_snoc_aux. simpl. rewrite length_of_snoc.
replace (S (length D) - 1) with (length D).
apply TypeSection_at_tail. omega.
rewrite lift_of_snoc_aux. simpl. rewrite length_of_snoc.
replace (S (length D) - (S (S n0))) with (length D - S n0).
apply TypeSection_at_head. trivial. omega.
Qed.

Lemma TypeSection_at_lem8:
  forall (D1 D2 D3: TypeEnv) (X: TypeVar) (Pi: TypeS),
    (TypeSection_at (D1 ++ D3) X Pi) ->
    (TypeSection_at (D1 ++ D2 ++ D3)
                    (lift_at_depth_in_typeVar X (length D2) (length D3)) 
                     Pi)
.

Proof.
intros.
destruct (match_comp_nat2_aux_0 X (length D3)).
(* X < |D3| *)
unfold lift_at_depth_in_typeVar. rewrite match_comp_nat2_lt.
apply TypeSection_at_lem3. apply TypeSection_at_lem3.
eapply TypeSection_at_lem4. apply H. trivial. trivial.
(* X >= |D3| *)
unfold lift_at_depth_in_typeVar. rewrite match_comp_nat2_ge.
replace (X + (length D2)) with ((length D3) + ((length D2) + (X - (length D3)))).
replace (D1 ++ D2 ++ D3) with ((D1 ++ D2) ++ D3).
apply TypeSection_at_lem2.
apply TypeSection_at_lem2.
eapply TypeSection_at_lem1.
replace X with ((length D3) + (X - (length D3))) in H. apply H.
omega. auto with datatypes. omega. trivial.
Qed.

(** Weakening of type declaration by the right **)
Lemma Weakening_of_type_declaration_by_the_right:
  forall (D1 D2 D3: TypeEnv) (X: TypeVar) (Pi: TypeS),
    (decl_of_typeVar (D1 ++ D3) X Pi) ->
    (decl_of_typeVar
      (D1 ++ D2 ++ (lift_in_typeEnv D3 (length D2)))
      (lift_at_depth_in_typeVar X (length D2) (length D3)) 
      (lift_at_depth_in_type Pi (length D2) (length D3)))
.

Proof.
intros. inversion H. clear D X0 H1 H2.
destruct (match_comp_nat2_aux_0 X (length D3)).
(* X < |D3| *)
unfold lift_at_depth_in_typeVar. rewrite match_comp_nat2_lt.
pattern (length D3).
replace (length D3) with (((length D3) - (S X)) + (S X) + 0).
unfold lift_in_type. rewrite Permuting_type_liftings_aux.
replace (length D3 - S X + 0) with (length D3 - S X).
fold (lift_in_type
        (lift_at_depth_in_type Pi0 (length D2) (length D3 - S X))
        (S X)).
apply decl_of_typeVar_intro.
apply TypeSection_at_lem3. apply TypeSection_at_lem3.
unfold lift_in_typeEnv.
apply TypeSection_at_lem7. eapply TypeSection_at_lem4. apply H0.
trivial. omega. omega. trivial.
(* X >= |D3| *)
unfold lift_at_depth_in_typeVar. rewrite match_comp_nat2_ge.
replace (lift_at_depth_in_type (lift_in_type Pi0 (S X)) (length D2) (length D3)) with
        (lift_in_type Pi0 (S (X + length D2))).
apply decl_of_typeVar_intro.
replace (X + (length D2)) with
((length (lift_in_typeEnv D3 (length D2))) + ((length D2) + (X - (length D3)))).
replace (D1 ++ D2 ++ (lift_in_typeEnv D3 (length D2))) with
        ((D1 ++ D2) ++ (lift_in_typeEnv D3 (length D2))).
apply TypeSection_at_lem2.
apply TypeSection_at_lem2.
eapply TypeSection_at_lem1.
replace X with ((length D3) + (X - (length D3))) in H0. apply H0.
omega. auto with datatypes.
unfold lift_in_typeEnv. rewrite length_of_typeLift. omega.
replace (length D3) with (0 + (length D3)).
replace (S X) with (((S X) - (length D3)) + (length D3)). 
unfold lift_in_type. rewrite Successive_type_liftings_aux.
replace (S (X + length D2)) with (S X - length D3 + length D3 + length D2).
trivial. omega. omega. omega. trivial.
Qed.

(** Weakening of well-kindedness by the right (auxiliary) **)
Lemma Weakening_of_well_kindedness_by_the_right_aux:
  forall (D1 D2 D: TypeEnv) (s: TypeS) (c: TypeC),
    (Type_WK D s c) ->
    forall (D3: TypeEnv),
      D = (D1 ++ D3) ->
      (Type_WK (D1 ++ D2 ++ (lift_in_typeEnv D3 (length D2))) 
         (lift_at_depth_in_type s (length D2) (length D3)) c)
.

Proof.
intros D1 D2 D s c H.
induction H.

(* Case (WK_proj) *)
intros.
simpl. eapply WK_proj. apply IHType_WK. trivial. trivial.

(* Case (WK_class) *)
intros.
simpl. apply WK_class. trivial.

(* Case (WK_fun) *)
intros.
simpl. apply WK_fun.
apply IHType_WK1. trivial. clear IHType_WK1.
rewrite snoc_of_append. rewrite snoc_of_append.
rewrite <- lift_of_snoc.
replace (S (length D3)) with (length (snoc D3 Pi)).
apply IHType_WK2.
rewrite H3. rewrite snoc_of_append. trivial.
apply length_of_snoc.
trivial. (* apply Lifting_preserves_class_types; trivial. *)
apply Lifting_preserves_kind_of; trivial.

(* Case (WK_apply) *)
intros.
simpl. eapply WK_apply.
apply IHType_WK1. trivial.
apply IHType_WK2. trivial.

(* Case (WK_var) *)
intros.
simpl. eapply WK_var.
apply Weakening_of_type_declaration_by_the_right.
rewrite <- H1. apply H.
apply Lifting_preserves_kind_of. trivial.

(* Case (WK_tuple) *)
intros.
simpl. apply WK_tuple.
apply IHType_WK. trivial.

(* Case (WK_param) *)
intros.
simpl. apply WK_param.
apply IHType_WK1. trivial. clear IHType_WK1.
rewrite snoc_of_append. rewrite snoc_of_append.
rewrite <- lift_of_snoc.
replace (S (length D3)) with (length (snoc D3 Pi)).
apply IHType_WK2.
rewrite H1. rewrite snoc_of_append. trivial.
apply length_of_snoc.

(* Case (WK_section) *)
intros.
simpl. apply WK_section.
replace (Type_section (lift_at_depth_in_type Ps (length D2) (S (length D3))))
  with (lift_at_depth_in_type (Type_section Ps) (length D2) (length D3)).
rewrite snoc_of_append. rewrite snoc_of_append.
rewrite <- lift_of_snoc.
replace (S (length D3)) with (length (snoc D3 (Type_section Ps))).
apply IHType_WK.
rewrite H0. rewrite snoc_of_append. trivial.
apply length_of_snoc. trivial.

(* Case (WK_none) *)
intros.
simpl. apply WK_none.

(* Case (WK_some) *)
intros.
simpl. apply WK_some.
apply Lifting_preserves_class_types. trivial.
apply IHType_WK. trivial.

(* Case (WK_param_nil) *)
intros.
simpl. apply WK_param_nil.

(* Case (WK_param_cons) *)
intros.
simpl. apply WK_param_cons.
apply IHType_WK1; trivial.
apply IHType_WK2; trivial.

(* Case (WK_constr_nil) *)
intros.
simpl. apply WK_constr_nil.

(* Case (WK_constr_cons) *)
intros.
simpl. apply WK_constr_cons.
apply IHType_WK1; trivial.
apply IHType_WK2; trivial.
Qed.

(** Weakening_of_well-kindedness by the right **)
Lemma Weakening_of_well_kindedness_by_the_right:
  forall (D: TypeEnv) (s: TypeS) (c: TypeC),
    (Type_WK D s c) ->
    forall (D': TypeEnv),
      (Type_WK (D ++ D') (lift_in_type s (length D')) c)
.

Proof.
intros.
unfold lift_in_type.
replace (0) with (length (A := TypeS) nil).
replace (D ++ D') with (D ++ D' ++ (lift_in_typeEnv nil (length D'))).
eapply Weakening_of_well_kindedness_by_the_right_aux.
apply H.
rewrite <- app_nil_end; trivial.
unfold lift_in_typeEnv. simpl.
rewrite <- app_nil_end. trivial.
trivial.
Qed.

(** Weakening_of_well-kindedness by the right **)
Lemma Weakening_of_well_kindedness_by_the_right_snoc:
  forall (D: TypeEnv) (s Pi: TypeS) (c: TypeC),
    (Type_WK D s c) ->
    (Type_WK (snoc D Pi) (lift_in_type s 1) c)
.

Proof.
intros.
unfold snoc.
change 1 with (length (Pi :: nil)).
apply Weakening_of_well_kindedness_by_the_right. trivial.
Qed.

(*******************************************************************)

Lemma TypeSection_at_lem1_cor1:
  forall (D1 D2 : TypeEnv) (Pi : TypeS) (k : nat),
    k >= (length D2) ->
    (TypeSection_at (D1 ++ D2) k Pi) ->
    (TypeSection_at D1 (k - (length D2)) Pi)
.

Proof.
intros.
cut (
  (TypeSection_at (D1 ++ D2) (length D2 + (k - length D2)) Pi) -> 
  (TypeSection_at D1 (k - length D2) Pi)
).
replace (length D2 + (k - length D2)) with k. tauto. omega.
apply TypeSection_at_lem1.
Qed.

(******************************************************************)

Theorem Weakening_of_type_declaration:
  forall (D1 D2 D3: TypeEnv) (X: TypeVar) (Pi: TypeS),
    (decl_of_typeVar (D1 ++ D3) X Pi) ->
    (decl_of_typeVar (D1 ++ D2 ++ (lift_in_typeEnv D3 (length D2)))
                     (lift_at_depth_in_typeVar X (length D2) (length D3))
                     (lift_at_depth_in_type Pi (length D2) (length D3)))
.

Proof.
intros.
destruct (match_comp_nat2_aux_0 X (length D3)).

(* X < |D3| *)
unfold lift_at_depth_in_typeVar. rewrite match_comp_nat2_lt.
Focus 2. trivial.
inversion H. clear D X0 H1 H2 H3 H.
rewrite Permuting_type_liftings_cor1. Focus 2. omega.
apply decl_of_typeVar_intro.
rewrite <- app_ass.
apply TypeSection_at_lem3.
unfold lift_in_typeEnv. apply TypeSection_at_lem7.
eapply TypeSection_at_lem4; eauto.

(* X >= |D3| *)
unfold lift_at_depth_in_typeVar. rewrite match_comp_nat2_ge.
Focus 2. trivial.
inversion H. clear D X0 H1 H2 H3 H.
rewrite Successive_type_liftings_cor1. Focus 2. omega.
change (S X + length D2) with (S (X + length D2)).
apply decl_of_typeVar_intro.
replace (X + length D2) with
  ((length (lift_in_typeEnv D3 (length D2))) + ((length D2) + (X - (length D3)))).
Focus 2. unfold lift_in_typeEnv. rewrite length_of_typeLift. omega.
rewrite <- app_ass. apply TypeSection_at_lem2.
apply TypeSection_at_lem2.
apply TypeSection_at_lem1_cor1; trivial.
Qed.

Theorem Weakening_of_well_kindedness:
  forall (D1 D2 D3: TypeEnv) (t: TypeS) (c: TypeC),
    (Type_WK (D1 ++ D3) t c) ->
    (Type_WK (D1 ++ D2 ++ (lift_in_typeEnv D3 (length D2)))
             (lift_at_depth_in_type t (length D2) (length D3)) c)
.

Proof.
cut (
  forall (D1 D2 D: TypeEnv) (t: TypeS) (c: TypeC),
    (Type_WK D t c) ->
    forall (D3: TypeEnv),
      D = (D1 ++ D3) ->
      (Type_WK (D1 ++ D2 ++ (lift_in_typeEnv D3 (length D2)))
               (lift_at_depth_in_type t (length D2) (length D3)) c)
).
intros. eapply H; eauto.

intros D1 D2 D t c H. induction H.
(* *)
intros; subst D; simpl; econstructor; eauto.
(* *)
intros; subst D; simpl; econstructor; eauto.
(* *)
intros; subst D; simpl; econstructor; eauto.
rewrite <- length_of_snoc with (x := Pi).
rewrite snoc_of_append. rewrite snoc_of_append. rewrite <- lift_of_snoc.
apply IHType_WK2. rewrite snoc_of_append. trivial.
apply Lifting_preserves_kind_of; trivial.
(* *)
intros; subst D; simpl; econstructor; eauto.
(* *)
intros. subst D. simpl. econstructor. apply Weakening_of_type_declaration. eauto.
apply Lifting_preserves_kind_of; trivial.
(* *)
intros; subst D; simpl; econstructor; eauto.
(* *)
intros; subst D; simpl; econstructor; eauto.
rewrite <- length_of_snoc with (x := Pi).
rewrite snoc_of_append. rewrite snoc_of_append. rewrite <- lift_of_snoc.
apply IHType_WK2. rewrite snoc_of_append. trivial.
(* *)
intros; subst D; simpl; econstructor; eauto.
rewrite snoc_of_append. rewrite snoc_of_append.
change  (Type_section (lift_at_depth_in_type Ps (length D2) (S (length D3)))) with
        (lift_at_depth_in_type (Type_section Ps) (length D2) (length D3)).
rewrite <- lift_of_snoc.
rewrite <- length_of_snoc with (x := (Type_section Ps)).
apply IHType_WK. rewrite snoc_of_append. trivial.
(* *)
intros; subst D; simpl; econstructor; eauto.
(* *)
intros; subst D; simpl; econstructor; eauto.
apply Lifting_preserves_class_types; trivial.
(* *)
intros; subst D; simpl; econstructor; eauto.
(* *)
intros; subst D; simpl; econstructor; eauto.
(* *)
intros; subst D; simpl; econstructor; eauto.
(* *)
intros; subst D; simpl; econstructor; eauto.
Qed.

(***********************************************************************)

(** Well-kindedness is preserved by subtitution of types that match kinds **)
Lemma Type_substitution_preserves_well_kindedness:
  forall (D1: TypeEnv) (Pi R: TypeS) (k: KindS),
    (Type_WK D1 R (TypeTuple k)) ->
    (kind_of Pi k) ->
    forall (D: TypeEnv) (s: TypeS) (c: TypeC),
      (Type_WK D s c) ->
      forall (D2: TypeEnv),
        D = (D1 ++ (Pi :: D2)) ->
        (Type_WK (D1 ++ (typeSubst_in_typeEnv D2 0 R))
                 (typeSubst_in_type s (length D2) R) c)
.

Proof.
intros D1 Pi R k H1 H2 D s c H.
induction H.

(* Case (WK_proj) *)
intros.
simpl.
eapply WK_proj.
apply IHType_WK; trivial. trivial.

(* Case (WK_class) *)
intros.
simpl.
apply WK_class. trivial.

(* Case (WK_fun) *)
intros.
simpl.
apply WK_fun. apply IHType_WK1; trivial.
clear IHType_WK1.
rewrite H5 in IHType_WK2.
rewrite snoc_of_append in IHType_WK2.
rewrite snoc_of_cons in IHType_WK2.
rewrite snoc_of_append.
replace (length D2) with (0 + length D2).
rewrite <- typeSubst_of_snoc.
simpl.
rewrite <- (length_of_snoc TypeS Pi0).
apply (IHType_WK2 (snoc D2 Pi0)). trivial.
trivial.
trivial. (* apply Substitution_preserves_class_types; trivial. *)
apply Substitution_preserves_kind_of; trivial.

(* Case (WK_apply). *)
intros.
simpl.
eapply WK_apply.
apply IHType_WK1; trivial.
apply IHType_WK2; trivial.

(* Case (WK_var) *)
intros.
simpl.
induction (lt_eq_lt_dec X (length D2)). induction a.
(* X < |D2| *)
unfold typeSubst_in_typeVar; rewrite match_comp_nat3_lt.
rewrite H3 in H.
apply WK_var with (Pi := (typeSubst_in_type Pi0 (length D2) R)).
eapply Declaration_of_type_variable_2.
trivial. apply H.
apply Substitution_preserves_kind_of. trivial. trivial.
(* X = |D2| *)
unfold typeSubst_in_typeVar; rewrite match_comp_nat3_eq.
rewrite H3 in H.
rewrite b in H.
assert (Pi0 = (lift_in_type Pi (S (length D2)))).
eapply decl_of_typeVar_is_a_partial_function. apply H.
apply Declaration_of_type_variable_3.
assert (k0 = k).
eapply kind_of_is_a_partial_function. apply H0.
rewrite H4.
unfold lift_in_type. apply Lifting_preserves_kind_of. trivial.
rewrite H5.
replace (length D2) with (length (typeSubst_in_typeEnv D2 0 R)).
apply Weakening_of_well_kindedness_by_the_right. trivial.
apply length_of_typeSubst.
trivial.

(* X > |D2| *)
unfold typeSubst_in_typeVar; rewrite match_comp_nat3_gt.
rewrite H3 in H.
apply WK_var with (Pi := (drop_in_type Pi0)).
eapply Declaration_of_type_variable_1.
trivial. apply H.
unfold drop_in_type. apply Dropping_preserves_kind_of. trivial.
trivial.

(* Case (WK_tuple) *)
intros.
simpl.
apply WK_tuple.
apply IHType_WK; trivial.

(* Case (WK_param) *)
intros.
simpl.
apply WK_param. apply IHType_WK1; trivial.
clear IHType_WK1.
rewrite H3 in IHType_WK2.
rewrite snoc_of_append in IHType_WK2.
rewrite snoc_of_cons in IHType_WK2.
rewrite snoc_of_append.
replace (length D2) with (0 + length D2).
rewrite <- typeSubst_of_snoc.
simpl.
rewrite <- (length_of_snoc TypeS Pi0).
apply (IHType_WK2 (snoc D2 Pi0)). trivial.
trivial.

(* Case (WK_section) *)
intros.
simpl.
apply WK_section.
replace (Type_section (typeSubst_in_type Ps (S (length D2)) R))
  with (typeSubst_in_type (Type_section Ps) (length D2) R).
rewrite snoc_of_append.
replace (length D2) with (0 + (length D2)).
rewrite <- typeSubst_of_snoc.
simpl.
rewrite <- (length_of_snoc TypeS (Type_section Ps)).
apply (IHType_WK (snoc D2 (Type_section Ps))).
rewrite H0.
rewrite snoc_of_append. rewrite snoc_of_cons. trivial.
trivial. trivial.

(* Case (WK_none) *)
intros.
simpl.
apply WK_none.

(* Case (WK_some) *)
intros.
simpl.
apply WK_some.
apply Substitution_preserves_class_types; trivial.
apply IHType_WK; trivial.

(* Case (WK_param_nil) *)
intros.
simpl.
apply WK_param_nil.

(* Case (WK_param_cons) *)
intros.
simpl.
apply WK_param_cons.
apply IHType_WK1; trivial.
apply IHType_WK2; trivial.

(* Case (WK_constr_nil) *)
intros.
simpl.
apply WK_constr_nil.

(* Case (WK_param_cons) *)
intros.
simpl.
apply WK_constr_cons.
apply IHType_WK1; trivial.
apply IHType_WK2; trivial.
Qed.

Lemma Substitution_preserves_well_kindedness_snoc:
  forall (D : TypeEnv) (Pi R : TypeS) (k : KindS) (s : TypeS) (c : TypeC),
    (Type_WK D R (TypeTuple k)) ->
    (kind_of Pi k) ->
    (Type_WK (snoc D Pi) s c) ->
    (Type_WK D (typeSubst_in_type s 0 R) c)
.

Proof.
intros.
replace 0 with (length (nil (A := TypeS))).
replace D with (D ++ (typeSubst_in_typeEnv (nil (A := TypeS)) 0 R)).
eapply Type_substitution_preserves_well_kindedness.
apply H. apply H0. apply H1. trivial. simpl. auto with datatypes. trivial.
Qed.

Lemma Type_substitution_preserves_well_kindedness_cor1:
  forall (Pi1 Pi2 R: TypeS) (k: KindS) (t: TypeS) (c: TypeC),
    (Type_WK nil R (TypeTuple k)) ->
    (kind_of Pi1 k) ->
    (Type_WK (Pi1 :: Pi2 :: nil) t c) ->
    (Type_WK ((typeSubst_in_type Pi2 0 R) :: nil)
             (typeSubst_in_type t 1 R) c)
.

Proof.
intros.
cut (
  (Type_WK (nil ++ (typeSubst_in_typeEnv (Pi2 :: nil) 0 R))
           (typeSubst_in_type t (length (Pi2 :: nil)) R) c)
).
simpl. tauto.
eapply Type_substitution_preserves_well_kindedness; eauto.
Qed.

Lemma Type_substitution_preserves_well_kindedness_cor2:
  forall (Pi R : TypeS) (k : KindS) (t : TypeS) (c : TypeC),
    (Type_WK nil R (TypeTuple k)) ->
    (kind_of Pi k) ->
    (Type_WK (Pi :: nil) t c) ->
    (Type_WK nil (typeSubst_in_type t 0 R) c)
.

Proof.
intros.
cut (
  (Type_WK (nil ++ (typeSubst_in_typeEnv nil 0 R))
           (typeSubst_in_type t (length (nil (A := TypeS))) R) c)
).
simpl. tauto.
eapply Type_substitution_preserves_well_kindedness; eauto.
Qed.

(******************************************************************************)

(** Weakening of type declaration by the left **)
Lemma Weakening_of_type_declaration_by_the_left:
  forall (D1 D2: TypeEnv) (X: TypeVar) (Pi: TypeS),
    (decl_of_typeVar D2 X Pi) ->
    (decl_of_typeVar (D1 ++ D2) X Pi)
.

Proof.
intros.
inversion H.
apply decl_of_typeVar_intro.
apply TypeSection_at_lem3. trivial.
Qed.

(** Weakening of well-kindedness by the left **)
Lemma Weakening_of_well_kindedness_by_the_left:
  forall (D1 D2: TypeEnv) (t: TypeS) (c: TypeC),
    (Type_WK D2 t c) ->
    (Type_WK (D1 ++ D2) t c)
.

Proof.
intros.
induction H.
(* WK-proj *)
eapply WK_proj. apply IHType_WK. trivial.
(* WK_class *)
apply WK_class. trivial.
(* WK_fun *)
apply WK_fun. trivial. rewrite snoc_of_append. trivial. trivial. trivial.
(* WK_apply *)
eapply WK_apply. apply IHType_WK1. trivial.
(* WK_var *)
eapply WK_var. apply Weakening_of_type_declaration_by_the_left. apply H. trivial.
(* WK_tuple *)
apply WK_tuple. trivial.
(* WK_param *)
apply WK_param. trivial. rewrite snoc_of_append. trivial.
(* WK_section *)
apply WK_section. rewrite snoc_of_append. trivial.
(* WK_class *)
apply WK_none.
(* WK_some *)
apply WK_some. trivial. trivial.
(* WK_param_nil *)
apply WK_param_nil.
(* WK_param_cons *)
apply WK_param_cons. trivial. trivial.
(* WK_constr_nil *)
apply WK_constr_nil.
(* WK_constr_cons *)
apply WK_constr_cons. trivial. trivial.
Qed.

(*********************************************************************)

Hint Rewrite snoc_of_append snoc_of_cons: snoc_rewrite.

(*********************************************************************)

Lemma Lifting_preserves_kind_of_reverse:
  forall (t: TypeS) (k: KindS) (n: nat) (d: nat),
    (kind_of (lift_at_depth_in_type t n d) k) ->
    (kind_of t k)
.

Proof.
intros. generalize k d H. clear k d H. induction t.
(* proj *)
simpl. intros. inversion H.
(* class *)
simpl. intros. inversion H.
(* fun *)
simpl. intros. inversion H.
(* apply *)
simpl. intros. inversion H.
(* var *)
simpl. intros. inversion H.
(* tuple *)
simpl. intros. inversion H.
(* param *)
simpl. intros. inversion H. apply Kind_of_param. eauto.
(* section *)
simpl. intros. inversion H. apply Kind_of_section. eauto.
(* none *)
simpl. intros. inversion H.
(* some *)
simpl. intros. inversion H.
(* nil *)
simpl. intros. inversion H. apply Kind_of_paramList_nil.
(* cons *)
simpl. intros. inversion H. apply Kind_of_paramList_cons; eauto. 
Qed.

(******************************************************************)

Lemma Forall2_app:
  forall (A B: Set) (P: A -> B -> Prop) (xs1 xs2: (list A)) (ys1 ys2: (list B)),
    (Forall2 P xs1 ys1) ->
    (Forall2 P xs2 ys2) ->
    (Forall2 P (xs1 ++ xs2) (ys1 ++ ys2))
.

Proof.
intros. generalize xs2 ys2 H0. clear xs2 ys2 H0. induction H.
(* *)
intros. simpl. trivial.
(* *)
intros. simpl. constructor; auto.
Qed.

(******************************************************************)

Lemma TypeSection_at_implies_List_from:
  forall (D: TypeEnv) (i: nat) (Pi: TypeS),
    (TypeSection_at D i Pi) ->
    (List_from D i Pi)
.

Proof.
intros. induction H; constructor; trivial.
Qed.

Lemma List_from_implies_TypeSection_at:
  forall (D: TypeEnv) (i: nat) (Pi: TypeS),
    (List_from D i Pi) ->
    (TypeSection_at D i Pi)
.

Proof.
intros. induction H; constructor; trivial.
Qed.

Lemma TypeSection_at_and_kind_of:
  forall (D1 D2: TypeEnv) (X: TypeVar) (Pi1: TypeS) (k: KindS),
  (Forall2 (fun Pi1 Pi2 => forall (k: KindS),
    (kind_of Pi1 k) ->(kind_of Pi2 k)) D1 D2) ->
  (TypeSection_at D1 X Pi1) ->
  (kind_of Pi1 k) ->
  exists Pi2: TypeS,
    (TypeSection_at D2 X Pi2) /\
    (kind_of Pi2 k)
.

Proof.
intros.
assert (List_from D1 X Pi1). apply TypeSection_at_implies_List_from; trivial.
destruct (Forall2_and_ListFrom _ _ _ _ _ _ _ H H2). destruct H3. rename x into Pi2.
exists Pi2. split.
apply List_from_implies_TypeSection_at; trivial.
auto.
Qed.

Lemma decl_of_typeVar_and_kind_of:
  forall (D1 D2: TypeEnv) (X: TypeVar) (Pi1: TypeS) (k: KindS),
    (Forall2 (fun Pi1 Pi2 => forall (k: KindS),
      (kind_of Pi1 k) ->(kind_of Pi2 k)) D1 D2) ->
    (decl_of_typeVar D1 X Pi1) ->
    (kind_of Pi1 k) ->
    exists Pi2: TypeS,
      (decl_of_typeVar D2 X Pi2) /\
      (kind_of Pi2 k)
.

Proof.
intros. destruct H0.
assert (kind_of Pi k).
eapply Lifting_preserves_kind_of_reverse. unfold lift_in_type in H1. eauto.
destruct (TypeSection_at_and_kind_of _ _ _ _ _ H H0 H2). destruct H3.
rename x into Pi2.
exists (lift_in_type Pi2 (S X)). split.
constructor. trivial.
unfold lift_in_type. apply Lifting_preserves_kind_of. trivial.
Qed.

Lemma decl_of_typeVar_and_kind_of_cor1:
  forall (D1 D2: TypeEnv) (X: TypeVar) (Pi Pi' Pi1: TypeS) (k k1: KindS),    
    (kind_of Pi k) ->
    (kind_of Pi' k) ->
    (decl_of_typeVar (D1 ++ Pi :: D2) X Pi1) ->
    (kind_of Pi1 k1) ->
    exists Pi2: TypeS,
      (decl_of_typeVar (D1 ++ Pi' :: D2) X Pi2) /\
      (kind_of Pi2 k1)
.

Proof.
intros.
eapply decl_of_typeVar_and_kind_of with (D1 := (D1 ++ Pi :: D2)).
(* *)
apply Forall2_app.
apply Forall2_refl. auto.
constructor.
intros.
replace k0 with k. trivial.
eapply kind_of_is_a_partial_function. apply H. trivial.
apply Forall2_refl. auto.
eauto. trivial.
Qed.

(*********************************************************************)

Theorem Well_kindedness_does_not_depend_on_bounds:
  forall (D1 D2: TypeEnv) (t Pi Pi': TypeS) (c: TypeC) (k: KindS),
    (kind_of Pi k) ->
    (kind_of Pi' k) ->
    (Type_WK (D1 ++ Pi :: D2) t c) ->
    (Type_WK (D1 ++ Pi' :: D2) t c)
.

Proof.
cut (
  forall (D1 D: TypeEnv) (t Pi Pi': TypeS) (c: TypeC) (k: KindS),
    (kind_of Pi k) ->
    (kind_of Pi' k) ->
    (Type_WK D t c) ->
    forall (D2: TypeEnv),
      D = (D1 ++ Pi :: D2) ->
      (Type_WK (D1 ++ Pi' :: D2) t c)
).
intros. eapply H. apply H0. trivial. eauto. trivial.
intros D1 D2 t Pi Pi' c k H0 H1 H. induction H.
(* *)
intros; subst D; econstructor; eauto.
(* *)
intros; subst D; econstructor; eauto.
(* *)
intros; subst D; econstructor; eauto.
autorewrite with snoc_rewrite. apply IHType_WK2.
autorewrite with snoc_rewrite. trivial.
(* *)
intros; subst D; econstructor; eauto.
(* *)
intros. subst D.
destruct (decl_of_typeVar_and_kind_of_cor1 _ _ _ _ _ _ _ _ H0 H1 H H2).
destruct H3. rename x into Pi0'.
econstructor. apply H3. trivial.
(* *)
intros; subst D; econstructor; eauto.
(* *)
intros; subst D; econstructor; eauto.
autorewrite with snoc_rewrite. apply IHType_WK2.
autorewrite with snoc_rewrite. trivial.
(* *)
intros; subst D; econstructor; eauto.
autorewrite with snoc_rewrite. apply IHType_WK.
autorewrite with snoc_rewrite. trivial.
(* *)
intros; subst D; econstructor; eauto.
(* *)
intros; subst D; econstructor; eauto.
(* *)
intros; subst D; econstructor; eauto.
(* *)
intros; subst D; econstructor; eauto.
(* *)
intros; subst D; econstructor; eauto.
(* *)
intros; subst D; econstructor; eauto.
Qed.

End SetProgram.