(* Author: Vincent Cremet *)
(* Date: July 2006 *) 
(* File: Util.v *)
(* Description: Definitions and theorems of general use *)

(** * Useful Properties of Basic Datatypes *)

Require Export SetIsType.

Require Import Arith.
Require Import List.
Require Import ListSet.
Require Import Omega.

(*****************************************************************)

(* Axiom AdmitIt: forall (P: Prop), P. *)

(* Ltac admit := apply AdmitIt. *)

(*****************************************************************)

(* Facts about appending of lists (++) *)

Lemma length_of_append:
  forall (A: Set) (xs ys: (list A)),
    (length (xs ++ ys)) = (length xs) + (length ys)
.

Proof.
intros A.
induction xs.
intros. simpl. trivial.
intros. simpl. rewrite IHxs. trivial.
Qed.

Lemma app_inject:
  forall (A: Set) (xs2 ys2 xs1 ys1: (list A)),
    (xs1 ++ xs2) = (ys1 ++ ys2) ->
    (length xs1) = (length ys1) ->
    (xs1 = ys1) /\ (xs2 = ys2)
.

Proof.
intros A xs2 ys2 xs1. induction xs1. intro ys1. induction ys1.
(* nil/nil *)
intros. tauto.
(* nil/cons *)
intros. simpl in H0. discriminate H0.
intro ys1. induction ys1.
(* cons/nil *)
intros. simpl in H0. discriminate H0.
(* cons/cons *)
intros. simpl in H. injection H. intros.
simpl in H0. injection H0. intro.
pose (H4 := (IHxs1 ys1 H1 H3)).
split. rewrite H2. replace ys1 with xs1. trivial. tauto. tauto.
Qed.

Lemma app_length_right:
  forall (A: Set) (xs1 xs2 ys1 ys2: (list A)),
    (length xs2) = (length ys2) ->
    (length (xs1 ++ xs2)) = (length (ys1 ++ ys2)) ->
    (length xs1) = (length ys1)
.

Proof.
intros.
rewrite length_of_append in H0.
rewrite length_of_append in H0. omega.
Qed.

(*****************************************************************)
(* Elimination of list map *)

Lemma map_elim:
  forall (A B: Set) (f: A -> B) (xs: (list A)) (y: B),
    (In y (map f xs)) ->
    exists x: A, (In x xs) /\ y = (f x)
.

Proof.
intros.
generalize xs H. clear xs H.
induction xs.
simpl. tauto.
simpl. intros.
destruct H.
exists a.
split. left. trivial.
symmetry. trivial.
destruct (IHxs H). destruct H0.
exists x.
split. right. trivial. trivial.
Qed.

(*****************************************************************)
(* Introduction of list map *)

Lemma map_intro:
  forall (A B: Set) (f: A -> B) (x: A) (xs: (list A)),
    (In x xs) ->
    (In (f x) (map f xs))
.

Proof.
intros.
generalize xs H. clear xs H.
induction xs.
simpl. tauto.
simpl. intros.
destruct H.
left. rewrite H. trivial.
right. apply IHxs. trivial.
Qed.

(*****************************************************************)
(* Length of map *)

Lemma length_of_map:
  forall (A B: Set) (f: A -> B) (xs: (list A)),
    (length (map f xs)) = (length xs)
.

Proof.
intros.
induction xs.
simpl. trivial.
simpl. rewrite IHxs. trivial.
Qed.

(*****************************************************************)
(* Length of map *)

Lemma map_of_append:
  forall (A B: Set) (f: A -> B) (xs ys: (list A)),
    (map f (xs ++ ys)) = (map f xs) ++ (map f ys)
.

Proof.
intros A B f xs.
induction xs.
simpl. trivial.
simpl. intros.
rewrite IHxs. trivial.
Qed.

(*****************************************************************)

Lemma cons_destruct:
    forall (A: Set) (x y: A) (xs ys: (list A)),
      (x = y) -> (xs = ys) -> (x :: xs) = (y :: ys)
.

Proof.
intros.
rewrite H. rewrite H0. trivial.
Qed.

(*****************************************************************)
(* Arithmetic properties *)

Lemma plus_minus_assoc:
  forall (n m p: nat),
    (p <= m) ->
    n + m - p = n + (m - p)
.

Proof.
intros. generalize n p H. clear n p H.
induction m.
intros.
simpl.
destruct n.
simpl. trivial.
rewrite plus_0_r.
assert (p = 0). auto with arith. rewrite H0.
rewrite <- minus_n_O. trivial.
intros.
destruct p.
simpl.
rewrite <- minus_n_O. trivial.
simpl.
rewrite <- plus_Snm_nSm. simpl.
apply IHm. auto with arith.
Qed.

Lemma myarith_1:
  forall (n m: nat),
    (m <= n) ->
    (n - m + m) = n
.

Proof.
intros.
rewrite plus_comm.
rewrite <- plus_minus_assoc.
apply minus_plus.
trivial.
Qed.

Lemma myarith_2:
  forall (n m p: nat),
    ((m + p) <= n) ->
    (n - m - p) = (n - (m + p))
.

Proof.
intros. generalize n p H. clear n p H.
induction m.
intros.
simpl. rewrite <- minus_n_O. trivial.
simpl. intros.
destruct n.
simpl. trivial.
simpl. apply IHm.
omega.
Qed.

(*******************************************************************)
(* Properties of listsets *)

Lemma empty_is_empty_1:
  forall (A: Set) (eqdec: forall (x y: A), { x = y } + { x <> y }),
    forall (x: A),
      ~(set_In x (empty_set A))
.

Proof.
intros.
unfold set_In. simpl. tauto.
Qed.

Lemma empty_is_empty_2:
  forall (A: Set) (eqdec: forall (x y: A), { x = y } + { x <> y }),
    forall (s: (set A)),
      (forall (x: A), ~(set_In x s)) ->
      (s = (empty_set A))
.

Proof.
intros A eqdec s.
destruct s.
trivial.
intros.
cut False. tauto.
unfold not in H.
apply H with (x := a).
simpl. tauto.
Qed.

Lemma empty_union_left:
  forall (A: Set) (eqdec: forall (x y: A), { x = y } + { x <> y }),
    forall (s1 s2: (set A)),
    (set_union eqdec s1 s2) = (empty_set A) ->
    s1 = (empty_set A)
.

Proof.
intros.
apply empty_is_empty_2. trivial.
unfold not.
intros.
assert (set_In x (set_union eqdec s1 s2)).
apply set_union_intro1. trivial.
rewrite H in H1.
apply empty_is_empty_1 with (x := x). trivial. trivial.
Qed.

Lemma empty_union_right:
  forall (A: Set) (eqdec: forall (x y: A), { x = y } + { x <> y }),
    forall (s1 s2: (set A)),
    (set_union eqdec s1 s2) = (empty_set A) ->
    s2 = (empty_set A)
.

Proof.
intros.
apply empty_is_empty_2. trivial.
unfold not.
intros.
assert (set_In x (set_union eqdec s1 s2)).
apply set_union_intro2. trivial.
rewrite H in H1.
apply empty_is_empty_1 with (x := x). trivial. trivial.
Qed.

(*****************************************************************)

(** Function: append the element [x] at the end of the list [xs] **)
Definition snoc (A: Set) (xs: (list A)) (x: A): (list A) :=
   (xs ++ (x :: nil))
.
Implicit Arguments snoc [A].

Lemma snoc_of_append:
  forall (A: Set) (xs ys: (list A)) (z: A),
    (snoc (xs ++ ys) z) = (xs ++ (snoc ys z))
.

Proof.
intros. unfold snoc. rewrite app_ass. trivial.
Qed.

Lemma snoc_of_cons:
  forall (A: Set) (x y: A) (xs: (list A)),
    (snoc (x :: xs) y) = (x :: (snoc xs y))
.

Proof.
intros. unfold snoc. simpl. trivial.
Qed.

Lemma length_of_snoc:
  forall (A: Set) (x: A) (xs: list A),
    (length (snoc xs x)) = (S (length xs))
.

Proof.
intros.
induction xs.
simpl. trivial.
simpl. unfold snoc in IHxs. rewrite IHxs. trivial.
Qed.

Section SnocInd.

Variable A: Set.
Variable P: (list A) -> Prop.
Variable H_nil: (P nil).
Variable H_snoc: forall (x: A) (xs: (list A)), (P xs) -> (P (snoc xs x)).

Lemma snoc_ind_aux_1:
  forall (xs: (list A)),
    (length xs) = 0 ->
    xs = nil
.

Proof.
intro xs. induction xs.
intro. trivial.
intro. simpl in H.
absurd (S (length xs) = 0). auto with arith. trivial.
Qed.

Lemma snoc_ind_aux_2:
  forall (xs: (list A)),
    (length xs) > 0 ->
    exists ys: (list A), exists y: A, xs = y :: ys
.

Proof.
intro xs. induction xs.
intro. absurd (length (nil (A := A)) > 0). simpl. omega. trivial.
intro. exists xs. exists a. trivial.
Qed.

Lemma snoc_ind_aux_3:
  forall (xs: (list A)) (x: A),
    (snoc xs x) <> nil
.

Proof.
intros. induction xs. unfold snoc. simpl. intro. discriminate H.
rewrite snoc_of_cons. intro. discriminate H.
Qed.

Lemma snoc_ind_aux_4:
  forall (xs: (list A)) (x: A),
    exists ys: (list A), exists y: A, x :: xs = (snoc ys y)
.

Proof.
intros xs. induction xs.
intro. exists (nil (A := A)). exists x. unfold snoc. simpl. trivial.
intro. induction (IHxs a). induction H.
exists (x :: x0). exists x1. rewrite H. unfold snoc. simpl. trivial.
Qed.

Lemma snoc_ind_aux_5:
  forall (xs: (list A)),
    (length xs) > 0 ->
    exists ys: (list A), exists y: A, xs = (snoc ys y)
.

Proof.
intros.
induction (snoc_ind_aux_2 xs H). induction H0.
rewrite H0. apply snoc_ind_aux_4.
Qed.

Lemma snoc_ind_aux_6:
  forall (n: nat) (xs: (list A)),
    (length xs) = n ->
    (P xs)
.

Proof.
intro n. induction n.
(* nil *)
intros.
replace xs with (nil (A := A)).
apply H_nil. symmetry. apply snoc_ind_aux_1. trivial.
(* snoc *)
intros.
assert ((length xs) > 0). rewrite H. omega.
induction (snoc_ind_aux_5 xs H0). induction H1. rewrite H1.
apply H_snoc. apply IHn. rewrite H1 in H. rewrite length_of_snoc in H.
injection H. trivial.
Qed.

Lemma snoc_ind:
  forall (xs: (list A)),
    (P xs)
.

Proof.
intros. apply (snoc_ind_aux_6 (length xs)). trivial.
Qed.

End SnocInd.

Lemma snoc_inject:
  forall (A: Set) (x y: A) (xs ys: (list A)),
    (snoc xs x) = (snoc ys y) ->
    (xs = ys) /\ (x = y)
.

Proof.
intros. unfold snoc in H.
assert (xs = ys /\ x :: nil = y :: nil).
apply app_inject. trivial.
apply app_length_right with (xs2 := x :: nil) (ys2 := y :: nil).
trivial. rewrite H. trivial.
destruct H0. injection H1. tauto.
Qed.

Lemma snoc_then_append:
  forall (A: Set) (x: A) (ys xs: (list A)),
    (snoc xs x) ++ ys = xs ++ (x :: ys)
.

Proof.
intros A x ys xs. induction xs.
simpl. trivial.
simpl. unfold snoc in IHxs. rewrite IHxs. trivial.
Qed.

Lemma snoc_nil:
  forall (A: Set) (x: A),
    (snoc nil x) = (x :: nil)
.

Proof.
intros. trivial.
Qed.

(********************************************************************)
(********************************************************************)

(* Module NatSet. *)

Require ListSet.
Require Peano_dec.

Definition natSet: Set := (ListSet.set nat).
Definition natSet_empty: natSet := (ListSet.empty_set nat).
Definition natSet_union: natSet -> natSet -> natSet :=
  fun x y => (ListSet.set_union Peano_dec.eq_nat_dec x y)
.
Definition natSet_add: nat -> natSet -> natSet :=
  fun x y => (ListSet.set_add Peano_dec.eq_nat_dec x y)
.
Definition natSet_in: nat -> natSet -> Prop :=
  fun x y => (ListSet.set_In x y)
.

(* End NatSet. *)

(********************************************************************)
(********************************************************************)

Section Forall2_section.

Variable A: Set.
Variable B: Set.

Inductive Forall2 (P: A -> B -> Prop): (list A) -> (list B) -> Prop :=
  
  | Forall2_nil:
    (Forall2 P nil nil)

  | Forall2_cons:
    forall (xA: A) (xB: B) (xsA: (list A)) (xsB: (list B)),
      (P xA xB) ->
      (Forall2 P xsA xsB) ->
      (Forall2 P (xA :: xsA) (xB:: xsB))
.

Lemma Forall2_implies_same_length:
  forall (P: A -> B -> Prop) (xsA: (list A)) (xsB: (list B)),
    (Forall2 P xsA xsB) ->
    (length xsA) = (length xsB)
.

Proof.
intros. induction H.
trivial. simpl. rewrite IHForall2. trivial.
Qed.

Lemma Forall2_implies:
  forall (P1: A -> B -> Prop) (P2: A -> B -> Prop)
         (xsA: (list A)) (xsB: (list B)),
         (forall (xA: A) (xB: B), (P1 xA xB) -> (P2 xA xB)) -> 
         (Forall2 P1 xsA xsB) ->
         (Forall2 P2 xsA xsB)
.

Proof.
intros. induction H0.
apply Forall2_nil.
apply Forall2_cons. apply H. trivial. trivial.
Qed.

End Forall2_section.

Implicit Arguments Forall2 [A B].

(********************************************************************)

Inductive List_from (A: Set): (list A) -> nat -> A -> Prop :=

  | List_from_tail:
    forall (xs: (list A)) (x: A),
      (List_from A (snoc xs x) 0 x)

  | List_from_head:
    forall (xs: (list A)) (x y: A) (n : nat),
      (List_from A xs n y) ->
      (List_from A (snoc xs x) (S n) y)
.

Implicit Arguments List_from [A].

Lemma List_from_domain:
  forall (A: Set) (xs: (list A)) (i: nat),
    i < (length xs) ->
    exists x: A,
      (List_from xs i x)
.

Proof.
intros. generalize i H. clear i H.
pattern xs. apply snoc_ind.
(* nil *)
intros. simpl in H. absurd (i < 0). omega. trivial.
(* snoc *)
intros. rewrite length_of_snoc in H0.
destruct i.
exists x. apply List_from_tail.
assert (i < length xs0). omega. destruct (H _ H1).
exists x0. apply List_from_head. trivial.
Qed.

Lemma List_from_is_a_partial_function:
  forall (A: Set) (xs: (list A)) (i: nat) (x: A),
    (List_from xs i x) ->
    forall (x': A),
      (List_from xs i x') ->
      (x = x')
.

Proof.
intros A xs i x H. induction H.
intros. inversion H. replace x with x0. trivial.
destruct (snoc_inject _ _ _ _ _ H1). trivial.
intros. inversion H0. apply IHList_from. replace xs with xs0. trivial.
destruct (snoc_inject _ _ _ _ _ H1). trivial.
Qed.

Lemma List_from_lem1:
  forall (A: Set) (xs1 xs2: (list A)) (x: A) (k: nat),
    (List_from (xs1 ++ xs2) ((length xs2) + k) x) ->
    (List_from xs1 k x)
.

Proof.
intros A xs1 xs2 x k. 
pattern xs2. apply (snoc_ind A).
simpl. rewrite <- (app_nil_end xs1). tauto.
intros. apply H.
rewrite <- snoc_of_append in H0.
rewrite length_of_snoc in H0. simpl in H0. inversion H0.
replace (xs1 ++ xs) with xs0. trivial.
destruct (snoc_inject _ _ _ _ _ H1). trivial.
Qed.

Lemma List_from_lem1_cor1:
  forall (A: Set) (xs1 xs2: (list A)) (x: A) (i: nat),
    i >= (length xs2) ->
    (List_from (xs1 ++ xs2) i x) ->
    (List_from xs1 (i - (length xs2)) x)
.

Proof.
intros.
eapply List_from_lem1 with (xs2 := xs2).
replace (length xs2 + (i - length xs2)) with i. trivial.
omega.
Qed.

Lemma List_from_lem2:
  forall (A: Set) (xs1 xs2: (list A)) (x: A) (k: nat),
    (List_from xs1 k x) ->
    (List_from (xs1 ++ xs2) ((length xs2) + k) x)
.

Proof.
intros. 
pattern xs2. apply (snoc_ind A).
rewrite <- (app_nil_end xs1). simpl. trivial.
intros. rewrite <- snoc_of_append. rewrite length_of_snoc. simpl.
apply List_from_head. trivial.
Qed.

Lemma List_from_lem3:
  forall (A: Set) (xs1 xs2: (list A)) (x: A) (k: nat),
    (List_from xs2 k x) ->
    (List_from (xs1 ++ xs2) k x)
.

Proof.
intros A xs1 xs2 x k H.
induction H.
rewrite <- snoc_of_append. apply List_from_tail.
rewrite <- snoc_of_append. apply List_from_head. trivial.
Qed.

Lemma List_from_lem4:
  forall (A: Set) (xs1 xs2: (list A)) (x: A) (k: nat),
    (List_from (xs1 ++ xs2) k x) ->
    (k < (length xs2)) ->
    (List_from xs2 k x)
.

Proof.
cut (
  forall (A: Set) (xs1 xs: (list A)) (k: nat) (x: A),
    (List_from xs k x) ->
    forall (xs2: (list A)),
      xs = xs1 ++ xs2 ->
      (k < (length xs2)) ->
      (List_from xs2 k x)
).
intros.
eapply H. apply H0. reflexivity. trivial.

intros A xs1 xs k x H. induction H.

intros.
assert ((length xs2) > 0). trivial. induction (snoc_ind_aux_5 A xs2 H1).
induction H2. rewrite H2 in H. rewrite <- snoc_of_append in H.
rewrite H2. replace x1 with x. apply List_from_tail.
destruct (snoc_inject _ _ _ _ _ H). trivial.

intros.
assert ((length xs2) > 0). omega. induction (snoc_ind_aux_5 A xs2 H2).
destruct H3. rewrite H3. apply List_from_head.
apply IHList_from.
rewrite H3 in H0. rewrite <- snoc_of_append in H0.
destruct (snoc_inject _ _ _ _ _ H0). trivial.
rewrite H3 in H1.
rewrite length_of_snoc in H1. omega.
Qed.

Lemma by_omega1: forall (x: nat), (x > 0) -> (S (pred x)) = x.
Proof. intros. omega. Qed.

Lemma Declaration_of_variable_1:
  forall (A: Set) (xs1 xs2: (list A)) (x x': A) (i: nat),
    i > (length xs2) ->
    (List_from (xs1 ++ (x :: xs2)) i x') ->
    (List_from (xs1 ++ xs2) (pred i) x')
.

Proof.
intros.
replace (pred i) with ((length xs2) + ((pred i) - (length xs2))).
apply List_from_lem2.
apply List_from_lem1 with (xs2 := x :: xs2).
simpl.
replace (S (length xs2 + (pred i - length xs2))) with i. trivial.
omega. omega.
Qed.

Lemma Declaration_of_variable_2:
  forall (A: Set) (xs1 xs2: (list A)) (x x': A) (i: nat),
    i < (length xs2) ->
    (List_from (xs1 ++ (x :: xs2)) i x') ->
    (List_from (xs1 ++ xs2) i x')
.

Proof.
intros.
apply List_from_lem3.
apply List_from_lem4 with (xs1 := (snoc xs1 x)). 
rewrite snoc_then_append. trivial. trivial.
Qed.

Lemma List_from_lem6:
  forall (A: Set) (xs1 xs2: (list A)) (x: A),
    (List_from (xs1 ++ (x :: xs2)) (length xs2) x)
.

Proof.
intros.
pattern xs2. apply (snoc_ind A).
simpl.
replace (xs1 ++ x :: nil) with (snoc xs1 x). apply List_from_tail. trivial.
intros. rewrite length_of_snoc. rewrite <- snoc_of_cons.
rewrite <- snoc_of_append. apply List_from_head. trivial.
Qed.

(*********************************************************************)

Inductive List_at (A: Set): (list A) -> nat -> A -> Prop :=

  | List_at_head:
    forall (xs: (list A)) (x: A),
      (List_at A (x :: xs) 0 x)

  | List_at_tail:
    forall (xs: (list A)) (x y: A) (n : nat),
      (List_at A xs n y) ->
      (List_at A (x :: xs) (S n) y)
.

Implicit Arguments List_at [A].

Lemma Forall2_and_ListAt:
  forall (A B: Set) (P: A -> B -> Prop) (xs: (list A)) (ys: (list B)) (i: nat) (x: A),
    (Forall2 P xs ys) ->
    (List_at xs i x) ->
    exists y: B,
      (List_at ys i y) /\
      (P x y)
.

Proof.
intros. generalize i x H0. clear i x H0. induction H.
(* nil *)
intros. inversion H0.
(* cons *)
intros. inversion H1.
(* - List_at_head *)
exists xB. split. apply List_at_head. rewrite <- H5. trivial.
(* List_at_tail *)
rewrite <- H3 in H1. inversion H1.
destruct (IHForall2 _ _ H11). destruct H12.
exists x2. split. apply List_at_tail. trivial. trivial.
Qed.

Lemma List_at_snoc_tail:
  forall (A: Set) (xs: (list A)) (x: A),
    (List_at (snoc xs x) (length xs) x)
.

Proof.
intros. induction xs.
simpl. unfold snoc. simpl. apply List_at_head.
simpl. rewrite snoc_of_cons. apply List_at_tail. trivial.
Qed.

Lemma List_at_right_weakening:
  forall (A: Set) (xs ys: (list A)) (i: nat) (x: A),
    (List_at xs i x) ->
    (List_at (xs ++ ys) i x)
.

Proof.
intros. induction H.
simpl. apply List_at_head.
simpl. apply List_at_tail. trivial.
Qed.

Lemma ListFrom_implies_ListAt:
  forall (A: Set) (xs: (list A)) (i: nat) (x: A),
    (List_from xs i x) ->
    (List_at xs ((length xs) - (S i)) x)
.

Proof.
intros. induction H.
(* List_from_tail *)
rewrite length_of_snoc.
replace (S (length xs) - 1) with (length xs).
apply List_at_snoc_tail. omega.
(* List_from_head *)
rewrite length_of_snoc. simpl.
unfold snoc. apply List_at_right_weakening. trivial.
Qed.

Lemma ListAt_implies_ListFrom:
  forall (A: Set) (xs: (list A)) (i: nat) (x: A),
    (List_at xs i x) ->
    (List_from xs ((length xs) - (S i)) x)
.

Proof.
intros. induction H.
(* List_at_head *)
simpl. replace (length xs - 0) with (length xs).
replace (x :: xs) with (nil ++ x :: xs).
apply List_from_lem6. trivial. omega.
(* List_at_tail *)
simpl.
replace (x :: xs) with ((x :: nil) ++ xs).
apply List_from_lem3. trivial.
auto with datatypes.
Qed.

Lemma List_from_inversion_on_length:
  forall (A: Set) (xs: (list A)) (i: nat) (x: A),
    (List_from xs i x) ->
    i < (length xs)
.

Proof.
intros A xs. pattern xs. apply snoc_ind.
intros. inversion H.
destruct (snoc_ind_aux_3 _ _ _ H1).
destruct (snoc_ind_aux_3 _ _ _ H0).
intros. inversion H0. rewrite length_of_snoc. omega.
rewrite length_of_snoc.
cut (n < (length xs1)). omega.
destruct (snoc_inject _ _ _ _ _ H1). rewrite H5.
eapply H. rewrite <- H5. apply H2.
Qed.

Lemma Forall2_and_ListFrom:
  forall (A B: Set) (P: A -> B -> Prop) (xs: (list A)) (ys: (list B)) (i: nat) (x: A),
    (Forall2 P xs ys) ->
    (List_from xs i x) ->
    exists y: B,
      (List_from ys i y) /\
      (P x y)
.

Proof.
intros.
assert ((length xs) = (length ys)). eapply Forall2_implies_same_length. apply H.
assert (List_at xs ((length xs) - (S i)) x). apply ListFrom_implies_ListAt. trivial.
destruct (Forall2_and_ListAt _ _ _ _ _ _ _ H H2). destruct H3.
exists x0. split.
assert (i < (length xs)). eapply List_from_inversion_on_length; eauto.
replace i with ((length ys) - (S ((length ys) - (S i)))).
apply ListAt_implies_ListFrom. 
rewrite <- H1. trivial. rewrite <- H1. omega.
trivial.
Qed.

(************************************************************************)

Lemma Forall2_snoc:
  forall (A B: Set) (P: A -> B -> Prop) (xAs: (list A)) (xBs: (list B)) (xA: A) (xB: B),
    (Forall2 P xAs xBs) ->
    (P xA xB) ->
    (Forall2 P (snoc xAs xA) (snoc xBs xB))
.

Proof.
intros A B P xAs. induction xAs.
intros. inversion H. unfold snoc. simpl. apply Forall2_cons. trivial. apply Forall2_nil.
intros. inversion H. rewrite snoc_of_cons. rewrite snoc_of_cons.
apply Forall2_cons. trivial. eauto.
Qed.

(********************************************************************)

Lemma Forall2_refl:
  forall (A: Set) (P: A -> A -> Prop),
    (forall (x: A), (P x x)) ->
    forall (xs: (list A)),
      (Forall2 P xs xs)
.

Proof.
intros. induction xs.
apply Forall2_nil.
apply Forall2_cons; trivial.
Qed.

Lemma Forall2_trans:
  forall (A B C: Set) (P1: A -> B -> Prop) (P2: B -> C -> Prop) (P3: A -> C -> Prop)
         (xs: (list A)) (ys: (list B)) (zs: (list C)),
    (forall (x: A) (y: B) (z: C),
      (P1 x y) -> (P2 y z) -> (P3 x z)) ->
    (Forall2 P1 xs ys) ->
    (Forall2 P2 ys zs) ->
    (Forall2 P3 xs zs)
.

Proof.
intros. generalize zs H1. clear zs H1. induction H0.
intros. inversion H1. apply Forall2_nil.
intros. inversion H2. apply Forall2_cons.
eapply H; eauto. eapply IHForall2; eauto.
Qed.   

(*******************************************************************)
(* Transitive closure of a binary relation *)
(*******************************************************************)

Definition symRel (A B: Set) (R: A -> B -> Prop) (y: B) (x: A) := (R x y).
Implicit Arguments symRel [A B].

(* Definition symRel (A: Set) (R: A -> A -> Prop): A -> A -> Prop := *)
(*   fun x y => (R y x) *)
(* . *)
(* Implicit Arguments symRel [A]. *)

Section Rplus_Section.

Variable A: Set.
Variable R: A -> A -> Prop.

(* Transitive closure *)
Inductive Rplus: A -> A -> Prop :=
  | tc_step: forall (x y: A), (R x y) -> (Rplus x y)
  | tc_trans: forall (x y z: A), (Rplus x y) -> (Rplus y z) -> (Rplus x z)
.

(* Transitive closure (other formulation) *)
Inductive Rplus1: A -> A -> Prop :=
  | tc_step1: forall (x y: A), (R x y) -> (Rplus1 x y)
  | tc_trans1: forall (x y z: A), (R x y) -> (Rplus1 y z) -> (Rplus1 x z)
.

Lemma Rplus1_is_transitive:
  forall (x y z: A), (Rplus1 x y) -> (Rplus1 y z) -> (Rplus1 x z)
.

Proof.
intros x y z H. generalize z. clear z. induction H.
intros. eapply tc_trans1; eauto.
intros. eapply tc_trans1. apply H. auto.
Qed.

Lemma Rplus_is_transitive:
  forall (x y z: A), (R x y) -> (Rplus y z) -> (Rplus x z)
.

Proof.
intros. eapply tc_trans. apply tc_step. eauto. trivial.
Qed.

Lemma Rplus_implies_Rplus1:
  forall (x y: A),
    (Rplus x y) ->
    (Rplus1 x y)
.

Proof.
intros. induction H.
apply tc_step1; trivial.
eapply Rplus1_is_transitive; eauto.
Qed.

Lemma Rplus1_implies_Rplus:
  forall (x y: A),
    (Rplus1 x y) ->
    (Rplus x y)
.

Proof.
intros. induction H.
apply tc_step; trivial.
eapply Rplus_is_transitive; eauto.
Qed.

End Rplus_Section.
Implicit Arguments Rplus [A].
Implicit Arguments Rplus1 [A].

Definition subrel (A B: Set) (R1 R2: A -> B -> Prop): Prop :=
  forall (x: A) (y: B),
    (R1 x y) ->
    (R2 x y)
.
Implicit Arguments subrel [A B].

Lemma Accessibility_is_preserved_by_inclusion:
  forall (A: Set) (R1 R2: A -> A -> Prop),
    (subrel R2 R1) ->
    forall (x: A),
      (Acc R1 x) ->
      (Acc R2 x)
.

Proof.
intros. induction H0. apply Acc_intro. intros.
apply H1. apply H. trivial.
Qed.

Lemma Rplus_of_symRel:
  forall (A: Set) (R: A -> A -> Prop),
  (subrel (Rplus (symRel R)) (symRel (Rplus R)))
.

Proof.
intros. unfold subrel. intros.
induction H.
unfold symRel. apply tc_step. trivial.
unfold symRel. eapply tc_trans. apply IHRplus2. apply IHRplus1.
Qed.

Lemma symRel_of_Rplus:
  forall (A: Set) (R: A -> A -> Prop),
  (subrel (symRel (Rplus R)) (Rplus (symRel R)))
.

Proof.
intros. unfold subrel. intros. unfold symRel in H. induction H.
apply tc_step. trivial.
eapply tc_trans. apply IHRplus2. apply IHRplus1.
Qed.

Lemma Rplus1_of_symRel:
  forall (A: Set) (R: A -> A -> Prop),
  (subrel (Rplus1 (symRel R)) (symRel (Rplus1 R)))
.

Proof.
cut (
  forall (A: Set) (R: A -> A -> Prop),
  (subrel (Rplus (symRel R)) (symRel (Rplus R)))
).
intros. unfold subrel. intros.
unfold symRel. apply Rplus_implies_Rplus1. apply H. apply Rplus1_implies_Rplus. trivial.

intros. unfold subrel. intros.
induction H.
unfold symRel. apply tc_step. trivial.
unfold symRel. eapply tc_trans. apply IHRplus2. apply IHRplus1.
Qed.

Lemma transitive_closure_well_founded:
  forall (A: Set) (R: A -> A -> Prop) (x: A),
    (Acc (symRel R) x) ->
    (Acc (symRel (Rplus R)) x)
.

Proof.
intros.
apply Accessibility_is_preserved_by_inclusion with (R1 := (Rplus (symRel R))).
apply symRel_of_Rplus.
apply Accessibility_is_preserved_by_inclusion with (R1 := (Rplus1 (symRel R))).
unfold subrel. apply Rplus_implies_Rplus1.
apply Accessibility_is_preserved_by_inclusion with (R1 := (symRel (Rplus1 R))).
apply Rplus1_of_symRel.
induction H. apply Acc_intro. intros.
unfold symRel in H1. destruct H1.
apply H0. trivial.
assert (Acc (symRel (Rplus1 R)) y). apply H0. unfold symRel. trivial.
inversion H3. apply H4. trivial.
Qed.

(*******************************************************************)

Section Exists_Section.

Variable A: Set.
Variable B: Set.
Variable P: A -> B -> Prop.

Hypothesis existence: forall (x: A), { y: B | (P x y) }.

Definition get_witness (x: A): B := let (y, _) := (existence x) in y.

Lemma correct_witness: forall (x: A), (P x (get_witness x)).
Proof. intros. unfold get_witness. destruct (existence x). trivial. Qed.

End Exists_Section.
Implicit Arguments get_witness [A B].

(*****************************************************************)

Inductive List_forall (A: Set) (P: A -> Prop): (list A) -> Prop :=
  | List_forall_nil:
    (List_forall A P nil)
  | List_forall_cons:
    forall (x: A) (xs: (list A)),
      (P x) -> (List_forall A P xs) ->
      (List_forall A P (x :: xs))
.
Implicit Arguments List_forall [A].

Lemma List_forall_and_List_at:
  forall (A: Set) (P: A -> Prop) (xs: (list A)) (i: nat) (x: A),
    (List_forall P xs) ->
    (List_at xs i x) ->
    (P x)
.

Proof.
intros. generalize i x H0. clear i x H0.
induction H.
intros. inversion H0.
intros. inversion H1.
replace x0 with x. trivial. eauto.
Qed.

Lemma List_forall_and_List_from:
  forall (A: Set) (P: A -> Prop) (xs: (list A)) (i: nat) (x: A),
    (List_forall P xs) ->
    (List_from xs i x) ->
    (P x)
.

Proof.
intros.
eapply List_forall_and_List_at. apply H.
apply ListFrom_implies_ListAt. apply H0.
Qed.

Lemma List_forall_implies_forall:
  forall (A: Set) (P1 P2: A -> Prop) (xs: (list A)),
    (forall (x: A), (P1 x) -> (P2 x)) ->
    (List_forall P1 xs) ->
    (List_forall P2 xs)
.

Proof.
intros. induction H0.
apply List_forall_nil.
apply List_forall_cons; auto.
Qed.

Lemma List_forall_equal_maps:
  forall (A B: Set) (f g: A -> B) (xs: (list A)),
    (List_forall (fun x => (f x) = (g x)) xs) ->
    (map f xs) = (map g xs)
.

Proof.
intros. induction H.
simpl. trivial.
simpl. rewrite H. rewrite IHList_forall. trivial.
Qed.

Lemma map_id_is_id:
  forall (A: Set) (xs: (list A)),
    (map (fun x => x) xs) = xs
.

Proof.
intros. induction xs. trivial. simpl. rewrite IHxs. trivial.
Qed.

Lemma List_forall_equal_maps_cor1:
  forall (A: Set) (f: A -> A) (xs: (list A)),
    (List_forall (fun x => (f x) = x) xs) ->
    (map f xs) = xs
.

Proof.
intros.
replace ((map f xs) = xs) with ((map f xs) = (map (fun x => x) xs)).
apply List_forall_equal_maps. trivial.
rewrite map_id_is_id. trivial.
Qed.

(**********************************************************************)

Lemma List_forall_of_map:
  forall (A B: Set) (P: B -> Prop) (f: A -> B) (xs: (list A)),
    (List_forall (fun x => (P (f x))) xs) ->
    (List_forall P (map f xs))
.

Proof.
intros. induction H.
simpl. apply List_forall_nil.
simpl. apply List_forall_cons; trivial. 
Qed.

(**********************************************************************)

Lemma List_forall_of_append_intro:
  forall (A: Set) (P: A -> Prop) (xs ys: (list A)),
    (List_forall P xs) ->
    (List_forall P ys) ->
    (List_forall P (xs ++ ys))
.

Proof.
intros. induction H.
simpl. trivial.
simpl. apply List_forall_cons; trivial. 
Qed.

Lemma List_forall_of_append_elim:
  forall (A: Set) (P: A -> Prop) (xs ys: (list A)),
    (List_forall P (xs ++ ys)) ->
    (List_forall P xs) /\
    (List_forall P ys)
.

Proof.
intros. induction xs.
split. apply List_forall_nil. trivial.
inversion H. destruct (IHxs H3).
split. apply List_forall_cons; trivial. trivial.
Qed.

Lemma List_forall_of_append_cor1:
  forall (A: Set) (P: A -> Prop) (xs ys: (list A)) (x: A),
    (List_forall P (xs ++ x :: ys)) ->
    (List_forall P (xs ++ ys))
.

Proof.
intros.
destruct (List_forall_of_append_elim _ _ _ _ H). inversion H1.
apply List_forall_of_append_intro; trivial.
Qed.

Lemma List_forall_implies_Forall2_right:
  forall (A B: Set) (P: B -> Prop) (xs: (list A)) (ys: (list B)),
    (List_forall P ys) ->
    (length xs) = (length ys) ->
    (Forall2 (fun x y => (P y)) xs ys)
.

Proof.
intros. generalize xs H0. clear xs H0. induction H.
intros. destruct xs. apply Forall2_nil. discriminate H0.
intros. destruct xs0. discriminate H1. apply Forall2_cons. trivial.
apply IHList_forall. simpl in H1. omega.
Qed.

Lemma List_forall_implies_map_id:
  forall (A: Set) (xs: (list A)) (f: A -> A),
    (List_forall (fun x => (f x) = x) xs) ->
    (map f xs) = xs
.

Proof.
intros. induction xs.
trivial.
simpl. inversion H. rewrite H2. rewrite IHxs. trivial. trivial.
Qed.

(******************************************************************)

Lemma List_at_of_map:
  forall (A B: Set) (f: A -> B) (xs: (list A)) (i: nat) (x: A),
    (List_at xs i x) ->
    (List_at (map f xs) i (f x))
.

Proof.
intros. induction H.
simpl. apply List_at_head.
simpl. apply List_at_tail. trivial.
Qed.

Lemma List_from_of_map:
  forall (A B: Set) (f: A -> B) (xs: (list A)) (i: nat) (x: A),
    (List_from xs i x) ->
    (List_from (map f xs) i (f x))
.

Proof.
intros.
assert (i < (length xs)). eapply List_from_inversion_on_length. eauto.
replace i with ((length xs) - (S ((length xs) - (S i)))).
rewrite <- length_of_map with (f := f).
apply ListAt_implies_ListFrom.
apply List_at_of_map. rewrite length_of_map.
apply ListFrom_implies_ListAt. trivial.
omega.
Qed.

(**********************************************************************)

Lemma Forall2_of_map:
  forall (A1 B1 A2 B2: Set) (P1: A1 -> B1 -> Prop) (P2: A2 -> B2 -> Prop)
         (fA: A1 -> A2) (fB: B1 -> B2) (xs: (list A1)) (ys: (list B1)),
    (forall (x: A1) (y: B1), (P1 x y) -> (P2 (fA x) (fB y))) ->
    (Forall2 P1 xs ys) ->
    (Forall2 P2 (map fA xs) (map fB ys))
.

Proof.
intros. induction H0.
simpl. apply Forall2_nil.
intros. simpl. apply Forall2_cons; auto.
Qed.

Lemma Forall2_of_map_cor1:
  forall (A : Set) (P : A -> A -> Prop) (f g: A -> A) (xs ys : list A),
    (Forall2 (fun x y => (P (f x) (g y))) xs ys) ->
    (Forall2 P (map f xs) (map g ys))
.

Proof.
intros.
apply Forall2_of_map with (P1 := fun x y => (P (f x) (g y))); auto.
Qed.

Lemma Forall2_of_map_cor2:
  forall (A : Set) (P : A -> A -> Prop) (f g: A -> A) (xs ys : list A),
    (Forall2 P (map f xs) (map g ys)) ->
    (Forall2 (fun x y => (P (f x) (g y))) xs ys)
.

Proof.
intros. generalize ys H. clear ys H. induction xs.
intros. inversion H. destruct ys. constructor. discriminate H0.
intros. inversion H. destruct ys. discriminate H2. simpl in H2. injection H2.
intros. rewrite H5 in H4. rewrite H6 in H3. constructor. trivial.
apply IHxs. trivial.
Qed.

(*************************************************************************)

Lemma Forall2_conj:
  forall (A B: Set) (P1: A -> B -> Prop) (P2: A -> B -> Prop)
         (xs: (list A)) (ys: (list B)),
    (Forall2 P1 xs ys) ->
    (Forall2 P2 xs ys) ->
    (Forall2 (fun x y => (P1 x y) /\ (P2 x y)) xs ys)
.

Proof.
intros. generalize H0. clear H0. induction H.
intros. apply Forall2_nil.
intros. inversion H1. apply Forall2_cons; tauto.
Qed.

Lemma List_forall_implies_Forall2:
  forall (A B: Set) (P1: A -> Prop) (P2: B -> Prop)
         (xs: (list A)) (ys: (list B)),
    (List_forall P1 xs) ->
    (length xs) = (length ys) ->
    (List_forall P2 ys) ->
    (Forall2 (fun x y => (P1 x) /\ (P2 y)) xs ys)
.

Proof.
intros. generalize ys H0 H1. clear ys H0 H1. induction H.
intros. inversion H1. apply Forall2_nil.
rewrite <- H3 in H0. discriminate H0.
intros. inversion H2. rewrite <- H3 in H1. discriminate H1.
apply Forall2_cons. tauto. apply IHList_forall. rewrite <- H5 in H1.
simpl in H1. omega. trivial.
Qed.
